#
# chipset.tcl - chips and chipsets
# 
# Copyright (C) 2005  Toni Ronkko
#
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Aug 23, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: chipset.tcl,v 1.8 2005/04/03 16:39:48 tronkko Exp $
#
# $Log: chipset.tcl,v $
# Revision 1.8  2005/04/03 16:39:48  tronkko
# highlight terrains
#
# Revision 1.7  2005/01/19 20:06:25  tronkko
# string compare -nocase does not work on tcl/tk 8.0.5
#
# Revision 1.6  2005/01/16 23:37:04  tronkko
# only prepare images for current zoom level to spare memory,
# zoom_levels is a list, shrink images
#
# Revision 1.5  2005/01/09 20:57:07  tronkko
# bug fix: do not reset accessibility when re-reading chipset
#
# Revision 1.4  2004/12/09 22:09:40  tronkko
# destroy_chipset, chip coordinates are now based on the chip width and
# height so that chip dimensions can be changed without destroying
# existing maps
#
# Revision 1.3  2004/12/01 20:11:04  tronkko
# clarified the difference between palette and workspace
#
# Revision 1.2  2004/11/30 20:31:21  tronkko
# using stipple pattern as a void image
#
# Revision 1.1  2004/11/28 11:08:32  tronkko
# Initial revision
#
#
#


#
# Initialize chipset from image with given chip size.  Beware: sets number
# of chips to zero.
#
# The function destroys chip properties such accessibility and depth
# coordinate.  Do not use the function unless you want to destroy
# the properties.
#
proc init_chipset { img w h } {
  global nchips
  global chip_depth
  global chip_bg
  global chip_terrain

  # initialize chip images
  set nchips 0

  # read chipset image into memory
  read_chipset_img $img $w $h

  # reset chip properties
  for {set i 0} {$i < $nchips} {incr i} {
    # reset access
    for {set j 1} {$j < 10} {incr j} {
      set chip_terrain($i,$j) "G"
    }

    # reset transparency
    set chip_bg($i) 0
    set chip_depth($i) 0
  }
}


#
# Re-read chipset image into memory.  The function is used for updating
# current image.  Chip properties are not be destroyed.
#
# The function can be used for replacing an existing chipset without
# changing the defined workspaces.  However, note that new chips will
# not automatically show up in existing workspaces - you need to
# explicitly insert chips.  Moreover, remember to refresh chips.
#
proc read_chipset_img { img w h } {
  global chipset_present
  global nchips
  global chipw
  global chiph
  global chip_image
  global chip_x1
  global chip_y1
  global chip_depth
  global chip_bg
  global chip_terrain

  # read image dimensions
  set imgw [image width $img]
  set imgh [image height $img]

  # destroy images of the existing chipset
  for {set i 0} {$i < $nchips} {incr i} {
    for {set z 0} {$z < 2} {incr z} {
      if { $chip_image($i,$z) != "" } {
        image delete $chip_image($i,$z)
      }
      set chip_image($i,$z) ""
    }
  }

  # reset flags indicating which chips have been updated
  for {set i 0} {$i < $nchips} {incr i} {
    set update($i) 0
  }

  # generate chip 0 for filling empty spaces
  set win [image create photo -width $w -height $h]
  $win copy chip0 -from 0 0 16 16 -to 0 0 $w $h
  set chip_image(0,0) $win
  set chip_image(0,1) ""
  set chip_x1(0) -1
  set chip_y1(0) -1
  if { $nchips == 0 } {
    set nchips 1
  }
  set chip_bg(0) 0
  set chip_depth(0) -1
  for {set j 1} {$j < 10} {incr j} {
    set chip_terrain(0,$j) "G"
  }
  set update(0) 1

  # divide image into chips
  for {set y1 0} {$y1 < $imgh} {incr y1 $h} {
    for {set x1 0} {$x1 < $imgw} {incr x1 $w} {

      # compute index of chip
      set a [expr int($x1 / $w)]
      set b [expr int($y1 / $h)]

      # Generete identifier for the chip at hand.  If we are overwriting
      # an existing chipset, then it makes sense to try to locate the old
      # chip first.  This way the existing workspaces do not need to be
      # destroyed.
      set id -1
      for {set i 1} {$i < $nchips} {incr i} {
        # does the chip exist in the previous set?
        if { $chip_x1($i) == $a  &&  $chip_y1($i) == $b } {
          # found a chip with the very same coordinates
          set id $i
          break
        }

        # any invalid chip will do unless a better alternative is found
        if { $chip_x1($i) < 0 } {
          set id $i
        }
      }

      # allocate new chip if no suitable chip was found
      if { $id < 0 } {
        set id $nchips
        incr nchips

        # reset access of new chip
        for {set i 1} {$i < 10} {incr i} {
          set chip_terrain($id,$i) "G"
        }

        # reset transparency
        set chip_bg($id) 0
        set chip_depth($id) 0
      }

      # coordinate of lower right corner
      set x2 [expr $x1 + $w]
      set y2 [expr $y1 + $h]

      # create image of chip
      set win [image create photo]

      # copy checkerboard pattern to background
      $win copy $chip_image(0,0)

      # Overwrite checkerboard pattern with front layer.  Pixels marked
      # as transparent will continue to show the checkerboard pattern.
      # The catch phrase lets us handle chips that are partially clipped
      # by boundaries of the image.
      catch {$win copy $img -from $x1 $y1 $x2 $y2}
      set chip_image($id,0) $win
      set chip_image($id,1) ""

      # remember location of the chip within the original image
      set chip_x1($id) $a
      set chip_y1($id) $b

      # mark the chip updated
      set update($id) 1
    }
  }

  # Invalidate chips of the existing chipset that fall outside the new 
  # image.  This is important when the new chipset is smaller than the
  # old one.  (Should be quite a rare event.)
  for {set i 0} {$i < $nchips} {incr i} {
    if { $update($i) == 0 } {
      # invalidate chip
      set chip_x1($i) -1
      set chip_y1($i) -1

      # create void image
      set win [image create photo -width $w -height $h]
      $win blank
      set chip_image($i,0) $win
    }
  }

  # remember chip size
  set chipw $w
  set chiph $h
 
  # we now have a chipset loaded
  set chipset_present 1
}

#
# Destroy chipset.  The function is used before loading a new chipset
# into memory.
#
proc destroy_chipset {} {
  global chipset_present
  global nchips
  global chipw
  global chiph
  global chip_image
  global chip_x1
  global chip_y1
  
  # destroy images of the existing chipset
  for {set i 0} {$i < $nchips} {incr i} {
    for {set z 0} {$z < 2} {incr z} {
      if { $chip_image($i,$z) != "" } {
        image delete $chip_image($i,$z)
      }
      set chip_image($i,$z) ""
    }
  }

  # erase chip database
  for {set i 0} {$i < $nchips} {incr i} {
    set chip_x1($i) -1
    set chip_y1($i) -1
  }

  # erase variables
  set chipw -1
  set chiph -1
  set nchips 0
  set chipset_present 0
}


#
# Re-create all chip images.  This is required after changing zoom level
# or reloading chipset image.
#
# Note: be sure to refresh all views after refreshing chipset.  Some
# images may be destroyed and re-created as a result of refresh, and
# graphics must be updated.
#
proc refresh_chipset { {renew 1} } {
  global nchips
  for {set i 0} {$i < $nchips} {incr i} {
    refresh_chip $i $renew
  }
}


#
# Combine chips to create an image for chip i.  The image will be saved to
# chip_image array with level 1.  Level 0 image must have been created
# before using this very function.  Argument renew specifies whether to
# re-create image or just redraw it.  This is important for speeding up
# certain operations.
#
# Note: be sure to refresh all views after remaking a chip.  That is,
# when argument renew is true, be sure to call refresh_view for all
# open views.  Some images may be destroyed and re-created as a result
# of refresh, and graphics must be updated.
#
proc refresh_chip { i renew } {
  global chipw
  global chiph
  global chip_image
  global zoom
  global chip_bg
  global chip_terrain
  global zoom_levels
  global transparent_color
  global view_terrains

  # compute subsample and zoom
  set z [lindex $zoom_levels $zoom]
  set ss [expr int(1/$z)]

  # get background chip
  set bg $chip_bg($i)

  # erase previous image
  if { $renew != 0 } {
    if { $chip_image($i,1) != "" } {
      image delete $chip_image($i,1)
      set chip_image($i,1) ""
    }
  }
  
  # create fresh image for chip
  if { $chip_image($i,1) != "" } {
    # modify existing image
    set win $chip_image($i,1)
    $win configure \
        -width [expr int($chipw*$z)] \
        -height [expr int($chiph*$z)]
    $win blank
  } else {
    # create fresh image
    set win [image create photo \
        -width [expr int($chipw*$z)] \
        -height [expr int($chiph*$z)] ]
  }

  # paint a chip
  if { $bg == 0 } {
    # no background, just copy chip image
    if { $z >= 1 } {
      $win copy $chip_image($i,0) -zoom [expr int($z)] [expr int($z)]
    } else {
      $win copy $chip_image($i,0) -subsample $ss $ss
    }
  } else {
    # copy background image
    if { $z >= 1 } {
      # enlarge
      $win copy $chip_image($bg,0) -zoom [expr int($z)] [expr int($z)]
    } else {
      # shrink
      $win copy $chip_image($bg,0) -subsample $ss $ss
    }

    # convert transparent pixel to list to ease comparison
    if { [scan $transparent_color "#%02x%02x%02x" tr tg tb] == 3 } {
      set bgcolor [list $tr $tg $tb]
    } else {
      error "cannot convert $transparent_color to rgb color value"
    }

    # compute increments and output pixel size
    if { $z >= 1 } {
      # enlarge
      set yi 1
      set xi 1
      set ps [expr int($z)]
    } else {
      # shrink
      set yi $ss
      set xi $ss
      set ps 1
    }

    #
    # Overlay foreground image on the top of the background.  If
    # foreground image has transparent pixels, those pixels will not
    # overwrite background image.
    #
    # Note that we try to avoid using $win put command, as it is quite
    # slow.
    #
    for {set y 0} {$y < $chiph} {incr y $yi} {
      for {set x 0} {$x < $chipw} {incr x $xi} {

        # get color value of source pixel as a list of r, g and b
        set px [$chip_image($i,0) get $x $y]
        if { [string compare $px $bgcolor] != 0 } {
          set x1 $x

          # process multiple pixels at once
          set row ""
          while { 1 } {
            # format source pixel to color value such as "#0accfe"
            set color [format "#%02x%02x%02x" \
                [lindex $px 0] [lindex $px 1] [lindex $px 2]]

            # duplicate pixel value to fill pixel size
            for {set k 0} {$k < $ps} {incr k} {
              lappend row $color
            }

            # check next pixel
            incr x $xi
            if { $x < $chipw } {
              set px [$chip_image($i,0) get $x $y]
              if { [string compare $px $bgcolor] == 0 } {
                # next pixel must not be touched
                break
              }
            } else {
              # end of row
              break
            }
          }

          # write pixels to destination image
          for {set k 0} {$k < $ps} {incr k} {
            $win put [list $row] \
              -to [expr int($x1*$z)] [expr int($y*$z+$k)]
          }
        }
      }
    }
  }
  
  # paint marked area directly to image
  if { [llength $view_terrains] > 0 } {
    # divide chip into 9 fragments  
    set w1 [expr int($chipw/3*$z)]
    set w2 [expr int($chipw*$z - 2*$w1)]
    set w3 $w1
    set h1 [expr int($chiph/3*$z)]
    set h2 [expr int($chiph*$z - 2*$h1)]
    set h3 $h1
    
    # compute fragment coordinates for splicing chip
    set x1 $w1
    set x2 [expr $x1 + $w2]
    set x3 [expr $x2 + $w3]
    set y1 $h1
    set y2 [expr $y1 + $h2]
    set y3 [expr $y2 + $h3]

    # paint regions
    if { [lsearch -exact $view_terrains $chip_terrain($i,1) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,1) 0 0 $x1 $y1
    }
    if { [lsearch -exact $view_terrains $chip_terrain($i,2) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,2) $x1 0 $x2 $y1
    }
    if { [lsearch -exact $view_terrains $chip_terrain($i,3) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,3) $x2 0 $x3 $y1
    }
    if { [lsearch -exact $view_terrains $chip_terrain($i,4) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,4) 0 $y1 $x1 $y2
    }
    if { [lsearch -exact $view_terrains $chip_terrain($i,5) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,5) $x1 $y1 $x2 $y2
    }
    if { [lsearch -exact $view_terrains $chip_terrain($i,6) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,6) $x2 $y1 $x3 $y2
    }
    if { [lsearch -exact $view_terrains $chip_terrain($i,7) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,7) 0 $y2 $x1 $y3
    }
    if { [lsearch -exact $view_terrains $chip_terrain($i,8) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,8) $x1 $y2 $x2 $y3
    }
    if { [lsearch -exact $view_terrains $chip_terrain($i,9) ] >= 0 } {
      highlight_terrain $win $chip_terrain($i,9) $x2 $y2 $x3 $y3
    }
  }
  
  # remember image handle
  set chip_image($i,1) $win
}


#
# Paint highlight over a chip image.
#
proc highlight_terrain { img letter x1 y1 x2 y2 } {
  # align coordinates
  set x1 [expr $x1 - ($x1 % 4)]
  set y1 [expr $y1 - ($y1 % 4)]
  
  # paint checkerboard pattern over the image
  for {set y $y1} {$y < $y2} {incr y 4} {
    for {set x $x1} {$x < $x2} {incr x 4} {
      $img put "#444444" -to $x $y [expr $x+2] [expr $y+2]
      $img put "#444444" -to [expr $x+2] [expr $y+2] [expr $x+4] [expr $y+4]
    }
  }
}