#!/bin/sh
# image.tcl - read pcx image files
# 
# Copyright (C) 2005  Toni Ronkko
# 
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of 
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Jan 16, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: image.tcl,v 1.3 2005/04/03 16:38:47 tronkko Exp $
#
# $Log: image.tcl,v $
# Revision 1.3  2005/04/03 16:38:47  tronkko
# minor optimization
#
# Revision 1.2  2005/01/19 20:07:09  tronkko
# read returns zero as "" in tcl/tk 8.0.5
#
# Revision 1.1  2005/01/16 23:34:06  tronkko
# Initial revision
#
#


#
# Read an image file fn to memory and return the image as a tk image.
# The function supports pcx file format along with file formats
# supported by Tk.
#
proc read_image { fn } {
  # read image file to memory
  set img ""
  if { [regexp -nocase {[.]pcx$} $fn] } {
    # file has .pcx extension, treat as pcx file
    set img [read_pcx $fn]
  } else {
    # read any other file format with standard tk image command
    set img [image create photo -file $fn]
  }
  return $img
}


#
# Read pcx image file fn.  Return Tk image that can be used like any
# other image created with [image create photo] command.
#
# The function is very slow, as pcx image files are read 
# completely in tcl code.
#
proc read_pcx { fn } {
  # open pcx file in binary mode
  set fd [open $fn "r"]
  fconfigure $fd -translation binary

  # read file and create image
  set img [internal_pcx $fd]
  if { $img != "" } {
    # file read in ok
    set ok 1
  } else {
    # trouble reading file
    puts "error reading file $fn"
    set ok 0
  }

  close $fd

  # return image to caller
  if { $ok } {
    # return image id
    return $img
  } else {
    # could not read image
    return ""
  }
}


#
# Read pcx image from file descriptor.  Returns image id or an empty
# string.
#
proc internal_pcx { fd } {
  # read header
  set header [read $fd 128]
  binary scan $header ccccssssssc48ccss \
      manufacturer version encoding nbits \
      xmin ymin xmax ymax hres vres \
      palette16 filler nplanes rowlen palettetype
  
  # see if the file format is supported
  if { $manufacturer != 10 } {
    puts "not a ZSoft pcx file"
    return ""
  }
  if { $encoding != 1 } {
    puts "unsupported encoding scheme"
    return ""
  }
  if { $palettetype != 1 } {
    puts "unsupported color encoding scheme"
    return ""
  }
  if { $nplanes != 1 } {
    puts "unsupported pixel format"
    return ""
  }

 # compute image size
  set w [expr $xmax - $xmin + 1]
  set h [expr $ymax - $ymin + 1]

  # seek to beginning of palette data
  seek $fd [expr -(256*3+1)] end
  set c [read_byte $fd]
  if { $c != 12 } {
    # no palette at the end of file
    puts "missing palette"
    return ""
  }

  # Read palette.  Individual color components will be formatted with
  # two digits to make it easier to construct rgb values such as
  # "#acef03" from palette.
  for {set i 0} {$i < 256} {incr i} {
    set palette($i) [format "#%02x%02x%02x" \
        [read_byte $fd] [read_byte $fd] [read_byte $fd]]
  }
  
  # seek to beginning of image data
  seek $fd 128 start

  # create void image
  set img [image create photo -width $w -height $h]

  # Read image line by line.  Reading data first into memory and then
  # assigning pixels to image consumes huge amounts of memory, so it is
  # really necessary to read lines one by one to make map editor usable
  # on ps2.
  for {set y 0} {$y < $h} {incr y} {
    # read row of data
    set x 0
    set row [list]
    while { $x < $w } {
      # read one pixel
      set c [read_byte $fd]
      if { $c >= 192 } {
        # RLE packed byte
        set rep [expr $c & 0x3f]
        set px [read_byte $fd]
        
        # store pixel value
        set color $palette($px)
        for {set i 0} {$i < $rep} {incr i} {
          lappend row $color
        }
        incr x $rep
      } else {
        # store regular pixel value
        lappend row $palette($c)
        incr x
      }
    }

    # write row of pixels into image
    $img put [list $row] -to 0 $y

    # update screen to let the user see that program is doing something
    set i [expr ($y+1)*100 / $h]
    status "Reading pcx file: $i%"
    update idletasks
  }
  return $img
}


#
# Read unsigned 8-bit byte from file descriptor fd.
#
proc read_byte { fd } {
  binary scan [read $fd 1] c c
  return [expr ($c + 0x100) % 0x100]
}

