#
# mapedit.tcl - main window of map editor
# 
# Copyright (C) 2005  Toni Ronkko
# 
# Peach Gardens is free software; you can redistribute it and/or modify 
# it under the terms of the GNU General Public License as published by the
# Free Software Foundation; either version 2 of the License, or (at your
# option) any later version.
#
# Peach Gardens is distributed in the hope that it will be useful, but 
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
# 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#
#
# Aug 23, 2004  Toni Ronkko <toni at users.playstation2-linux.com>
# 
# $Id: mapedit.tcl,v 1.9 2005/06/22 13:21:56 tronkko Exp $
#
# $Log: mapedit.tcl,v $
# Revision 1.9  2005/06/22 13:21:56  tronkko
# bug fix: division error caught after cancelling file load
#
# Revision 1.8  2005/04/03 16:39:05  tronkko
# highlight accessibility menu
#
# Revision 1.7  2005/01/19 20:07:29  tronkko
# visit home page
#
# Revision 1.6  2005/01/16 23:35:02  tronkko
# zoom_levels is a list, do not load image twice when imporing new
# chipset
#
# Revision 1.5  2005/01/09 20:54:17  tronkko
# about dialog, edit mode
#
# Revision 1.4  2004/12/09 22:16:15  tronkko
# added help menu and file handling functions
#
# Revision 1.3  2004/12/01 20:11:50  tronkko
# clarified the difference between palette and workspace
#
# Revision 1.2  2004/11/30 20:30:08  tronkko
# added edit, view and workspace menus, lock workspaces feature
#
# Revision 1.1  2004/11/28 11:08:41  tronkko
# Initial revision
#
#
#

# current protection status (for highlighting menu entries)
set cur_protect 0

# current zoom level (for highlighting menu entries)
set cur_zoom $zoom

# Menu hint texts.  Whenever the user moves the mouse over a menu,
# the system automatically tries to fetch a hint text from the array
# below, and show the text at status bar.
set menuhelp {
  # file
  "Load..."             "Load project from file"
  "Save"                "Save project to a file"
  "Save As..."          "Save project to a named file"
  "Reread"              "Re-read current chipset to update graphics"
  "Import Chipset..."   "Import chipset from image"
  "Exit"                "Exit map editor"

  # edit
  "Erase Chips"         "Erase selected chips"
  "Undo"                "Undo small changes"
  "Select"              "Select chips using left mouse button"
  "Draw"                "Point and click to paste selection"
  "Move"                "Drag selection to relocate chips"
  "Edit Chip"           "Select chip to edit its properties"
  "Pan View"            "Use left mouse button to pan view"
  "Lock Palettes"       "Protect palettes from accidental editing"
  "Accessability..."    "Define areas that can be accessed by the player"

  # view
  "Grid"                "Show/hide grid"
  "Unreachable"         "Highlight unreachable regions"
  "Forest"              "Highlight forest areas"
  "Road"                "Highlight roads"
  "Water"               "Highlight waterways"
  "Mountain"            "Highlight mountaneous areas"

  # palette
  "Reset..."            "Insert whole chipset to a palette"

  # map

  # help
  "Help Browser"        "Launch embedded help browser"
  "Keyboard Shortcuts"  "List keyboard shortcuts"
}

# file name patterns for file save and load dialogs
set file_types {
  {{Map Files}  {.map}}
  {{All Files}  *}
}


#
# Create a toplevel window for the map editor.
#
proc create_mapedit {} {
  global env
  global statusbar_widget
  global sourcedir
  global zoom_levels
  global palette_menu
  global map_menu
  global selector_widget
  global editor_widget
  global window_title
  global editor_grid
  global menubar
  global geometry

  # read resource database
  set fn [file join $env(HOME) .Xdefaults]
  if {[file isfile $fn]} {
    option readfile $fn
  }

  # create name for the main window
  set top .mapedit
  toplevel $top -class PGMapEdit
  
  # set title of the main window (which has already been opened by tk)
  wm title $top $window_title
  wm iconname $top "PGMapEdit"

  # Names of window elements.  Help menu will be right-justfied if it is
  # named "menu".  Here, the item has been name xhelp to prevent the
  # justification.
  set menubar           $top.menubar
  set file_menu         $menubar.file
  set edit_menu         $menubar.edit
  set view_menu         $menubar.palette
  set map_menu          $menubar.map
  set menu_zoom         $menubar.zoom
  set palette_menu      $menubar.ws
  set help_menu         $menubar.xhelp
  set pane              $top.pane
  set statusbar         $top.statusbar
  set statusbar_widget  $statusbar.frame

  # divide main window into left and right portions
  create_pane $pane 0 0
  divide_pane $pane 0.25
  pack $pane -expand yes -fill both

  # create selector
  set selector_widget [create_view $pane.left.selector]

  # create editor
  set editor_widget [create_view $pane.right.editor]

  # status bar
  frame $statusbar -borderwidth 1 -relief raised 
  frame $statusbar.frame -borderwidth 1 -relief sunken
  label $statusbar_widget.text -padx 4 -pady 2 -justify left
  button $statusbar_widget.lock -image actunlock16 \
      -relief flat  -width 16 -height 16 \
      -command "swap_protection" -takefocus 0
  
  # create menu bar
  menu $menubar -tearoff no
  
  # file menu
  menu $file_menu -tearoff no
  $menubar add cascade -label "File" -menu $file_menu -underline 0
  $file_menu add command -label "New..." -underline 0 \
      -command new_project
  $file_menu add separator
  $file_menu add command -label "Load..." -underline 0 \
      -command load_project
  $file_menu add command -label "Save" -underline 0 \
      -command save_project
  $file_menu add command -label "Save As..." \
      -command save_as
  $file_menu add separator
  $file_menu add command -label "Reread" -underline 0 \
      -command reread_image
  $file_menu add command -label "Import Chipset..." \
      -command import_image
  $file_menu add separator
  $file_menu add command -label "Quit" -underline 0 \
      -command confirm_exit

  # edit menu
  menu $edit_menu -tearoff no
  $menubar add cascade -label "Edit" -menu $edit_menu -underline 0
  $edit_menu add command -label "Undo" -command undo
  $edit_menu add separator
  $edit_menu add radiobutton -label "Select" -underline 0 \
      -variable mouse_mode -value "select" \
      -command "set_mouse_mode select"
  $edit_menu add radiobutton -label "Draw" -underline 0 \
      -variable mouse_mode -value "draw" \
      -command "set_mouse_mode draw"
  $edit_menu add radiobutton -label "Move" -underline 0 \
      -variable mouse_mode -value "move" \
      -command "set_mouse_mode move"
  $edit_menu add radiobutton -label "Pan View" -underline 0 \
      -variable mouse_mode -value "pan" \
      -command "set_mouse_mode pan"
  $edit_menu add radiobutton -label "Edit Chip" -underline 0 \
      -variable mouse_mode -value "edit" \
      -command "set_mouse_mode edit"
  $edit_menu add separator  
  $edit_menu add checkbutton -label "Lock Palettes" -underline 0 \
      -variable cur_protect -offvalue 0 -onvalue 1 \
      -command swap_protection
  $edit_menu add separator
  $edit_menu add command -label "Erase Chips" -command "erase_chips"

  # view menu
  menu $view_menu -tearoff no
  $menubar add cascade -label "View" -menu $view_menu -underline 0
  menu $menu_zoom -tearoff no  
  $view_menu add cascade -label "Zoom" -menu "$menu_zoom" -underline 0  
  for {set i [expr [llength $zoom_levels]-1]} {$i >= 0 } {incr i -1} {
    # get zoom level
    set z [lindex $zoom_levels $i]
    
    # format zoom factor as text
    if { $z >= 1 } {
      set text "$z:1"
    } else {
      set text "1:[expr int(1/$z)]"
    }

    # insert menu option
    $menu_zoom add radiobutton -label "$text" -variable cur_zoom \
        -value $i -command "set_zoom $i"
  }
  
  $view_menu add checkbutton -label "Grid" -underline 0 \
      -variable editor_grid -offvalue 0 -onvalue 1 \
      -command {set_border $editor_widget $editor_grid}
  
  $view_menu add checkbutton -label "Unreachable" -underline 0 \
      -variable editor_unreachable -offvalue 0 -onvalue 1 \
      -command {highlight_access}
  $view_menu add checkbutton -label "Forest" -underline 0 \
      -variable editor_forest -offvalue 0 -onvalue 1 \
      -command {highlight_access}
  $view_menu add checkbutton -label "Road" -underline 0 \
      -variable editor_road -offvalue 0 -onvalue 1 \
      -command {highlight_access}
  $view_menu add checkbutton -label "Water" -underline 0 \
      -variable editor_water -offvalue 0 -onvalue 1 \
      -command {highlight_access}  
  $view_menu add checkbutton -label "Mountain" -underline 0 \
      -variable editor_mountain -offvalue 0 -onvalue 1 \
      -command {highlight_access}

  set_border $editor_widget $editor_grid

  # palette menu
  menu $palette_menu -tearoff no  
  $menubar add cascade \
      -label "Palette" -menu "$palette_menu" -underline 0
  
  # map menu
  menu $map_menu -tearoff no
  $menubar add cascade -label "Map" -menu $map_menu -underline 0

  # help menu  
  menu $help_menu -tearoff no
  $menubar add cascade -label "Help" -menu $help_menu -underline 0
  $help_menu add command -label "Help Browser" \
      -command "help_window" -underline 0
  $help_menu add command -label "Keyboard Shortcuts" \
      -command "help_window 1" -underline 0
  $help_menu add separator
  $help_menu add command -label "Copying" \
      -command "help_window 3" -underline 0
  $help_menu add command -label "About..." -underline 0 \
      -command about_dialog
  
  # attach menu bar to main window
  $top configure -menu $menubar

  # show panes
  pack $pane -expand yes -fill both

  # show status bar
  pack $statusbar -fill x -side bottom -padx 1 -pady 1
  pack $statusbar.frame -side left -fill x -expand yes
  pack $statusbar_widget.lock -side left
  pack $statusbar_widget.text -side left -expand yes -anchor w
 
  # handle exit from the window manager
  wm protocol $top WM_DELETE_WINDOW confirm_exit

  # default key bindings
  bind $top <KeyPress-s> "set_mouse_mode select"
  bind $top <KeyPress-d> "set_mouse_mode draw"
  bind $top <KeyPress-p> "set_mouse_mode pan"
  bind $top <KeyPress-m> "set_mouse_mode move"
  bind $top <KeyPress-e> "set_mouse_mode edit"
  bind $top <KeyPress-l> "swap_protection"
  bind $top <KeyPress-Delete> "erase_chips"
  bind $top <KeyPress-F5> "reread_image"
  bind $top <KeyPress-F1> "help_window"
  bind $top <Control-s>  "save_project"
  bind $top <Control-q>  "confirm_exit"
  bind $top <Control-p>  "create_palette_dialog"
  bind $top <Control-m>  "create_map_dialog"
  bind $top <Control-l> "load_project"
  bind $top <Control-z> "undo"

  # alt-p opens view menu instead of invoking pan view
  bind $top <Alt-KeyPress-p> "$view_menu invoke 0"

  # hint text of the lock image
  bind $statusbar_widget.lock <Enter> \
    "status {Lock/release palette from accidental modifications}"
  bind $statusbar_widget.lock <Leave> "status {}"

  # move keyboard focus to selector
  focus $selector_widget

  # reset views
  refresh_view $editor_widget
  refresh_view $selector_widget
  
  # insert dynamic entries
  refresh_menu

  # Define minimum and requested size of the main window.  In tk 8.4 this
  # seems to be essential or otherwise the window will be very small, and
  # the size tends to reset when touching menu entries.
  wm minsize $top 320 256
  if { $geometry == "" } {
    # try to get geometry from resource database
    set width  [option get . width ""]
    set height [option get . height ""]
    
    if { $width != ""  &&  $height != "" } {
      # size specified in resource database
      wm geometry $top "${width}x${height}"
    } else {
      # default size
      wm geometry $top "640x450"
    }
  } else {
    # geometry from command line
    wm geometry $top "$geometry"
  }
}


#
# Create new map project.
#
proc new_project {} {
  global selector_widget
  global editor_widget
  global cur_file
  global edit_status

  # verify destruction of existing chipset
  if { $edit_status } {
    set ok [modal_dialog \
        "Create New Project" \
        "Unsaved data exists. Lose changes?" question {Yes No} 0 ]
    if { $ok > 0 } {
      return
    }
  }

  # ask for file name and size
  set result [import_dialog]
  if { $result != "" } {
    # destroy existing chipset
    reset_undo
    destroy_chipset
    destroy_workspaces

    # import chipset
    replace_image $result

    # create initial map
    status "Create map"
    update idletasks
    create_map_dialog "world"
    
    # there is no file loaded now
    set cur_file ""
  } else {
    status "Cancel"
  }
}


#
# Ask a file name, and load a file to memory.
#
proc load_project { {fn ""} } {
  global default_dir
  global file_types
  global nworkspaces
  global workspace_used
  global workspace_type
  global selector_widget
  global editor_widget
  global cur_file
  global edit_status

  # verify destruction of existing chipset
  if { $edit_status } {
    set ok [modal_dialog \
        "Load Project" \
        "Unsaved data exists. Lose changes?" question {Yes No} 0 ]
    if { $ok > 0 } {
      return
    }
  }

  # lose undo buffers
  reset_undo

  # ask for file name unless supplied
  if { $fn == "" } {
    set fn [tk_getOpenFile -parent . -filetypes $file_types \
        -title "Load..." -defaultextension .map \
        -initialdir $default_dir]
  }  
  if { $fn != "" } {

    # read file into memory
    set ok [load_data $fn]
    if { $ok } {
      # remember default directory and file for loading next map
      set default_dir [file dirname $fn]
      set cur_file $fn

      # find first palette & map
      set pal -1
      set map -1
      for {set i 0} {$i < $nworkspaces} {incr i} {
        if { $workspace_used($i) } {
          if { $workspace_type($i) == "map"  &&  $map < 0 } {
            set map $i
          }
          if { $workspace_type($i) == "palette"  &&  $pal < 0 } {
            set pal $i
          }
        }        
      }
      
      status "Preparing map for display, please wait..."
      update idletasks
      
      # refresh views
      refresh_menu
      select_palette $pal
      select_map $map
      reset_selection $selector_widget
      
      status "Ready"
    } else {
      # load ended in error => no data loaded
      refresh_view $selector_widget
      refresh_view $editor_widget
      refresh_menu
    }
    
  } else {
    status "Cancel"
  }
}


#
# Save file with current name.
#
proc save_project {} {
  global chipset_present
  global cur_file
  if { $chipset_present } {
    if { $cur_file != "" } {

      status "Saving file $cur_file, please wait..."
      update idletasks

      # overwrite existing file
      set ok [save_data $cur_file]
      if { $ok } {
        # no unsaved data in memory
        set_edited 0
        status "Saved $cur_file"

        # do not allow undo beyound save file
        reset_undo
      }

    } else {
      # no file name
      save_as
    }
  } else {
    modal_dialog "Error" "No chipset in memory" error
  }
}


#
# Ask a file name, and save the current chipset and workspaces to file.
#
proc save_as {} {
  global chipset_present
  global file_types
  global cur_file
  global default_dir
  
  if { $chipset_present } {
    # default file name
    if { $cur_file != "" } {
      set file [file tail $cur_file]
    } else {
      set file ""
    }
    
    # ask for file name
    set fn [tk_getSaveFile -parent . -filetypes $file_types \
        -title "Save As..." -defaultextension .map \
        -initialdir $default_dir -initialfile $file]
    if { $fn != "" } {

      set ok [save_data $fn]
      if { $ok } {
        # remember file name for save
        set cur_file $fn

        # no unsaved data in memory
        set_edited 0
        status "Saved $cur_file"
        
        # do not allow undo beyound save file
        reset_undo
      }
      
    } else {
      status "Cancel"
    }
    
  } else {
    modal_dialog "Error" "No chipset in memory" error
  }
}


#
# Ask file name, and import image from file.  The command is used for
# replacing current chipset with another one.
#
proc import_image {} {
  global selector_widget
  global editor_widget
  global chipset_present
  global chipset_fn

  # confirm overwrite of existing chipset
  if { $chipset_present } {
    set ok [modal_dialog \
        "Overwrite Chipset?" \
        "Importing chipset will replace the old one possibly making\
the map unusable.  Do you still wish to overwrite the existing\
chipset?" question {Yes No} 0]
    if { $ok > 0 } {
      # do not replace
      return 0
    }
  }

  # ask for file name and size
  set result [import_dialog]
  if { $result != "" } {
    replace_image $result
    return 1
  } else {
    status "Cancel"
    return 0
  }
}


#
# Import from file replacing current chipset.  The sole argument result
# is a return value of import_dialog proc.
#
proc replace_image { result } {
  global selector_widget
  global editor_widget
  global chipset_present
  global chipset_fn

  # create initial palette & map?
  if { $chipset_present } {
    set init 0
  } else {
    set init 1
  }

  # split result into components
  set fn  [lindex $result 0]
  set w   [lindex $result 1]
  set h   [lindex $result 2]
  set img [lindex $result 3]
  status ""
  update idletasks

  # remember image size
  set imgw [image width $img]
  set imgh [image height $img]

  # initialize chipset
  read_chipset_img $img $w $h
  image delete $img

  if { $init } {
    # create initial palette
    set bw [expr int(($imgw+$w-1)/$w)]
    set bh [expr int(($imgh+$h-1)/$h)]
    set pal [init_palette $bw $bh]

    # switch views
    select_palette $pal
  }

  status "Refreshing view, please wait..."
  update idletasks
    
  # update screen
  refresh_chipset
  refresh_view $selector_widget
  refresh_view $editor_widget
  refresh_menu

  # there are unsaved changes now
  set_edited 1

  # remember image file
  set chipset_fn $fn

  # select any chip so that user can start editing with keyboard
  reset_selection $selector_widget
}


#
# Show or hide certain terrain types.  This very function is called
# as a response to user clicking on a menu.  Hence, global variables
# such as editor_unreachable and editor_forest are set by Tk to
# indicate change of terrains.
#
proc highlight_access {} {
  global editor_unreachable
  global editor_forest
  global editor_road
  global editor_water
  global editor_mountain
  global view_terrains

  status "Refreshing view, please wait..."
  update idletasks  

  # update list of viewed terrains
  set view_terrains ""
  if { $editor_unreachable } {
    lappend view_terrains "U"
  }
  if { $editor_forest } {
    lappend view_terrains "F"
  }
  if { $editor_road } {
    lappend view_terrains "R"
  }
  if { $editor_water } {
    lappend view_terrains "W"
  }
  if { $editor_mountain } {
    lappend view_terrains "M"
  }
  
  # refresh display without re-creating images
  status "Done"
  refresh_chipset 0
}


#
# Re-read image file to update graphics.  The function is needed when
# the user edits the chipset image in an image editor while the map
# editor is open.  By re-reading the image file the looks are updated
# on the map editor side.
#
proc reread_image {} {
  global chipset_fn
  global chipw
  global chiph
  global selector_widget
  global editor_widget

  if { $chipset_fn != "" } {
    if { [file exists $chipset_fn] } {
      status "Re-loading chipset, please wait..."
      update idletasks
      
      set img [read_image $chipset_fn]
      if { $img != "" } {
        # re-import image without asking anything
        read_chipset_img $img $chipw $chiph
        image delete $img

        status "Refreshing view, please wait..."
        update idletasks

        # update screen
        refresh_chipset
        refresh_view $selector_widget
        refresh_view $editor_widget
        status "Ready"
      } else {
        status "Image not loaded."
      }
    } else {
      # image file has been deleted/moved
      modal_dialog "Error" "Cannot find file $chipset_fn" error
    }
  } else {
    # trying to re-read without an image being loaded first
    modal_dialog "Error" "No image loaded" error
  }
}


#
# Update menu bar.  Note that the dynamic palette and map menus must be
# updated only after disabling other menu entries.  The two menu have
# special state changing patterns, and the functions refresh_palette_menu
# and refresh_map_menu take care of disabling necessary menu entries
# by themself.
# 
proc refresh_menu {} {
  global chipset_present
  global menubar
  
  # figure out the global state
  if { $chipset_present } {
    set state normal
  } else {
    set state disabled
  }
  
  # enable/disable all menu entries except help
  set n1 [$menubar index last]
  for {set i 0} {$i < $n1} {incr i} {
    if { [$menubar type $i] == "cascade" } {
      # sub-menu
      set sub [$menubar entrycget $i -menu]
      set n2 [$sub index last]
      if { $n2 != "none" } {
        for {set j 0} {$j <= $n2} {incr j} {
          if { [$sub type $j] != "separator" } {
            $sub entryconfigure $j -state $state
          }
        }
      }
    }
  }

  # insert dynamic menu entries
  refresh_palette_menu
  refresh_map_menu

  # always enable new, load and exit
  set sub [$menubar entrycget 0 -menu]
  $sub entryconfigure [$sub index "New*"]  -state normal
  $sub entryconfigure [$sub index "Load*"] -state normal
  $sub entryconfigure [$sub index "Quit*"] -state normal
}


#
# Confirm exit from the map editor, and close the application.
#
proc confirm_exit {} {
  global edit_status
  
  # see if there is anything unsaved
  if { $edit_status != 0 } {
    # unsaved changes 
    set index [modal_dialog \
        "Exit Map Editor?" \
        "There are unsaved changes.  Really exit map editor?" \
        question {Yes No} 0 ]
    if {$index == 0} {
      # user said yes
      exit 0
    }
  } else {
    # no unsaved changes
    exit 0
  }
}


#
# Update mouse action mode
#
proc set_mouse_mode { mode } {
  global mouse_mode
  global selector_widget
  global editor_widget
  global write_protect
  global chipset_present

  # refuse to change modes unless there is a chipset
  if { ! $chipset_present } {
    return
  }
  
  # remember mode
  set mouse_mode $mode

  # update hint text
  if { $mode == "select" } {
    status "Press and hold left mouse button to select chips"
    
  } elseif { $mode == "draw" } {
    if { ! $write_protect($selector_widget) } {
      # writing allowed in workspace
      status "Press left mouse button to duplicate selection"
    } else {
      status "Press left mouse button to paste selection over map"
    }
    
  } elseif { $mode == "move" } {
    status "Drag selection to relocate chips"
    
  } elseif { $mode == "edit" } {
    status "Select chip to edit properties"
    
  } elseif { $mode == "pan" } {
    status "Press and hold left mouse button to pan view"
    
  } else {
    modal_dialog "Error" "Invalid mode: $mode" error
  }

  # change cursor to match the mode
  restore_cursor $selector_widget
  restore_cursor $editor_widget
}


#
# Enable/disable palette protection and update the image of the lock on
# status bar depending on the mode.
#
proc swap_protection {} {
  global statusbar_widget
  global write_protect
  global selector_widget
  global cur_protect
  global chipset_present

  # refuse to change modes unless there is a chipset
  if { ! $chipset_present } {
    return
  }

  # swap mode
  set old_protect $write_protect($selector_widget)
  set new_protect [expr 1-$old_protect]
  set write_protect($selector_widget) $new_protect
  set cur_protect $new_protect

  # select icon image based on lock mode
  if { $cur_protect } {
    set img actlock16
  } else {
    set img actunlock16
  }
  $statusbar_widget.lock configure -image $img

  # update cursor to match the mouse mode
  restore_cursor $selector_widget

  # hint text
  if { $cur_protect } {
    status "Work space is protected from accidental changes"
  } else {
    status "Work space is editable"
  }
}


#
# Set file status to edited.  The status is used when closing the program
# to see if the user has any unsaved changes still in the editor
#
proc set_edited { f } {
  global edit_status
  global window_status

  # remember status
  if { $f } {
    incr edit_status
  } else {
    set edit_status 0
  }

  # update window title
  if { $window_status == 0  &&  $edit_status != 0 } {
    # first edit
    refresh_title
  } elseif { $window_status != 0  &&  $edit_status == 0 } {
    # just saved
    refresh_title
  }
}


#
# Update window title.
# 
proc refresh_title {} {
  global edit_status
  global window_title
  global edit_status
  global window_status
  global editor_widget
  global workspace_name
  global cur_workspace

  # mark edited file with asterisk
  if { $edit_status } {
    set edited "* "
  } else {
    set edited ""
  }

  # get map name
  set cur $cur_workspace($editor_widget)
  if { $cur >= 0 } {
    set name "$workspace_name($cur) - "
    set icon $workspace_name($cur)
  } else {
    set name ""
    set icon "PGMapEdit"
  }

  # set title and icon name
  wm title .mapedit "$edited$name$window_title"
  wm iconname .mapedit $icon

  # do not update window title at next edit unless saved between
  set window_status $edit_status
}


#
# Set zoom level of all view widgets.
#
proc set_zoom { i } {
  global zoom
  global cur_zoom
  global selector_widget
  global editor_widget

  if { $zoom != $i } {
    # changing zoom level takes time, prepare user
    status {Changing zoom level, please wait...}
    update idletasks
  
    # remember zoom value
    set zoom $i
    set cur_zoom $i

    # re-create chipset images for the new zoom level
    refresh_chipset

    # repaint views
    status {Repainting map, please wait...}
    update idletasks
    refresh_view $selector_widget
    refresh_view $editor_widget
  }
}


#
# Show hint text in the status bar widget when the user moves the mouse
# cursor over a menu entry.
#
bind Menu <<MenuSelect>> {
  global menuhelp
  
  # find out name of the menu entry
  if {[catch {%W entrycget active -label} label]} {
    set label ""
  }

  # see if there is a matching hint text
  if {$label != ""} {
    set i [lsearch $menuhelp $label]
    if {$i >= 0} {
      # found a hint text
      set label [lindex $menuhelp [expr $i + 1]]
    } else {
      # no text
      set label ""
    }
  }

  # show hint in status bar
  status $label
  update idletasks
}


#
# Show about dialog
#
proc about_dialog {} {
  modal_dialog \
      "About..." \
"Peach Gardens Map Editor version 1.1
Copyright (C) 2005 Toni Ronkko 

This very program is crafted especially for a game called Peach Gardens.\
Please visit Peach Gardens home page at\
http://playstation2-linux.com/projects/pgardens/ for more info." \
    info {OK} 0]
}