/*
 * depth.c - depth sorting
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Jan 21 2005, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: depth.c,v 1.1 2005/04/03 16:29:46 tronkko Exp $
 *
 * $Log: depth.c,v $
 * Revision 1.1  2005/04/03 16:29:46  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "depth.h"
#include "graphics.h"
#include "image.h"
#include "debug.h"


/* divide screen estate into n vertical bands to speed up sorting */
#define BANDS 20

/* maximum number of queued operations */
#define QUEUE_LEN 2000



/*
 * A queued drawing operation.
 */
struct operation
{
  struct operation *next;
  int z;
  struct image *imp;
  int sx1, sy1, sx2, sy2;
  int dx1, dy1, dx2, dy2;
};


/* buffer for queued operations */
static struct operation ops[QUEUE_LEN];
static int nops = 0;

/* linked lists of operations per band */
typedef struct operation *band_ptr;
static band_ptr bands[BANDS];


/*
 * Remove queued objects from depth buffers.  Use this function at the
 * beginning of a display cycle to initialize depth sorting.
 */
void
zreset (void)
{
  int i;
  
  /* erase all stored operations */
  nops = 0;

  /* empty linked lists */
  for (i = 0; i < BANDS; i++) {
    bands[i] = NULL;
  }
}


/*
 * Paint depth enabled objects to screen.  Note that objects are sorted when
 * inserting them, so we do not need to sort objects here.
 */
void
zdraw (void)
{
  int band;
#ifndef NDEBUG
  int z = -999999;
#endif

  /* paint all bands */
  for (band = 0; band < BANDS; band++) {

    /* paint all queued operations of a band */
    struct operation *op = bands[band];
    while (op != NULL) {
      /* make sure that operations are sorted correctly */
      assert (op->z >= z);

      /* paint queued operation */
      bitblt(
          op->imp, 
          op->sx1, op->sy1, op->sx2, op->sy2,
          op->dx1, op->dy1, op->dx2, op->dy2);
      
#ifndef NDEBUG
      z = op->z;
#endif
      op = op->next;
    }

    /* erase band */
    bands[band] = NULL;
  }

  /* erase all stored operations */
  nops = 0;
}


/*
 * Draw region of source image to screen with depth coordinate.  Currently,
 * the operation is implemented with a depth sorting algorithm, and the
 * operation will be queued until zdraw() or zreset() function is called.
 * A pointer to image imp is stored by the function, and the image must not
 * be destroyed before zdraw() or zreset().
 *
 * Depth coordinates are specified from back to front such that an object
 * with greater depth coordinate is painted over objects having smaller
 * depth coordinates.  For example, an object with depth coordinate of 0 is
 * drawn after an object with depth coordinate of -1 but before an object
 * with depth coordinate of 100.
 */
void zbitblt(
    struct image *imp, int z, 
    int sx1, int sy1, int sx2, int sy2,
    int dx1, int dy1, int dx2, int dy2)
{
  struct operation *op;
  struct operation *p;
  struct operation *prev;
  int band;

  /* allocate space for operation */
  if (nops < QUEUE_LEN) {
    op = &ops[nops++];
  } else {
    /* no more space in queue */
    abort_msg ("too many depth enabled objects");
  }

  /* initialize operation */
  op->imp = imp;
  op->z   = z;
  op->sx1 = sx1;
  op->sy1 = sy1;
  op->sx2 = sx2;
  op->sy2 = sy2;
  op->dx1 = dx1;
  op->dy1 = dy1;
  op->dx2 = dx2;
  op->dy2 = dy2;

  /* compute suitable band */
  band = z / BANDS;
  if (band < 0) {
    band = 0;
  } else if (band >= BANDS) {
    band = BANDS - 1;
  }

  /* find object with greater depth coordinate */
  prev = NULL;
  p = bands[band];
  while (p != NULL  &&  p->z < z) {
    prev = p;
    p = p->next;
  }

  if (prev != NULL) {
    /* insert into middle of a list */
    assert (prev->z < z);
    assert (p == NULL  ||  z <= p->z);
    op->next = p;
    prev->next = op;
  } else {
    /* first operation of a band */
    op->next = p;
    bands[band] = op;
  }

  /* make sure that band is correctly sorted */
#ifndef NDEBUG
  op = bands[band];
  while (op != NULL) {
    assert (op->next == NULL  ||  op->z <= op->next->z);
    op = op->next;
  }
#endif
}
