/*
 * maemo.c - graphics interface to Nokia 770 Internet Tablet
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Jun 11 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: maemo.c,v 1.1 2005/06/22 13:10:27 tronkko Exp $
 *
 * $Log: maemo.c,v $
 * Revision 1.1  2005/06/22 13:10:27  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#ifdef MAEMO
#include <hildon-widgets/hildon-app.h>
#include <hildon-widgets/hildon-appview.h>
#include <hildon-widgets/hildon-note.h>
#include <hildon-widgets/gtk-infoprint.h>
#include <gtk/gtkmain.h>
#include <gtk/gtk.h>
#include <gdk/gdkkeysyms.h>
#include <libosso.h>

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#include "graphics.h"
#include "options.h"
#include "image.h"
#include "input.h"
#include "ticks.h"
#include "font.h"
#include "maemo.h"
#include "debug.h"


/* program title */
#define APP_TITLE "Peach Gardens"

/* Program key for storing parameters and communicating with other apps.
 * Must not contain a slash or a period character. */
#define OSSO_APP "pgardens"

/* Duration of display cycle in millisecond.  Value of 20 is equal to 50
 * frames per sec.  Undefine for dynamic frame rate. */  
#define TICKS_PER_FRAME 20

/* define to use double buffering (a must) */
#define DOUBLE_BUFFER 


/* properties of registered image */
struct reg_image {
  int used;                                    /* for allocated structs */
  GdkPixbuf *pixbuf;                           /* graphics object */
  guchar *pixels;                              /* image data */
};

/* store pixel in a format that is suitable for gtk */
struct pixel {
  unsigned char r;
  unsigned char g;
  unsigned char b;
  unsigned char a;
};


/* application data */
static HildonApp *app;
static HildonAppView *appview;

/* has graphics subsystem been initialized? */
static int gfx_initialized = 0;

/* set to true if gtk main loop has been exited for good */
static int exiting;

/* is graphics window currently open? */
static int window_open;

/* for communicating with system */
static osso_context_t *osso_context = NULL;

/* menu entries (for graying menu items) */
static GtkMenu *main_menu;
static GtkWidget *file_menu;
static GtkWidget *item_file;
static GtkWidget *item_new;
static GtkWidget *item_open;
static GtkWidget *item_close;
static GtkWidget *item_save;
static GtkWidget *item_saveas;
static GtkWidget *item_full;

/* painting canvas */
static GtkWidget *drawing_area;

/* drawing context */
static GdkGC *gc;

/* are we on fullscreen mode? */
static int fullscreen;

/* trigger mode switch at next buffer swap */
static int fullscreen_switch;

/* visible region */
static GdkRegion *visible;

/* whether application is currently in view */
static int active;

/* background and foreground color */
static GdkColor bg = { 0, 0x0000, 0x0000, 0x0000 };
static GdkColor fg = { 0, 0xff00, 0xff00, 0xff00 };

/* registered images */
static struct reg_image images[MAX_IMAGES];

/* event buffer */
#define EV_BUFFER_LEN 8
static int eventbuf[EV_BUFFER_LEN];
static int nevents;


/* invoked from menu close item */
static void quit (void);

/* invoked on key press/release */
static gboolean key_press(
    GtkWidget *widget, GdkEventKey *event, HildonApp *app);
static gboolean key_release(
    GtkWidget *widget, GdkEventKey *event, HildonApp *app);

/* listen hardware events */
static void hw_event_handler (osso_hw_state_t *state, gpointer data);

/* close request from system */
static void exit_event_handler (gboolean die_now, gpointer data);
static void im_close (HildonAppView *hildonappview, gpointer user_data);

/* deal with user switching applications */
static void deactivate (HildonAppView *hildonappview, gpointer user_data);
static void activate (HildonAppView *hildonappview, gpointer user_data);
  
/* switch between fullscreen and windowed mode */
static void toggle_fullscreen (void);

/* start off-screen paint */
static void begin_paint (void);

/* copy off-screen drawings to screen */
static void end_paint (void);

/* load image to graphics device */
static void load_image (struct image *imp);
  
/* release image memory */
static void purge_images (void);

/* release raw image data */
static void image_destroy (guchar *pixels, gpointer data);

/* store event for later retrieval */
static void push_event (int ev);


/*
 * For bitmapped fonts: blend rectangular sub-image (sx,sy)-(sx+w,sy+h) with
 * foreground color to frame buffer at (dx,dy).  This is typically used for
 * painting fonts.
 *
 * There is no equivalent function in gdk, so the function cannot be
 * implemented for Nokia 770 Internet Tablet.  This prototype will be
 * removed after bitmap fonts have been converted to native fonts.
 */
void
textblt(
    struct image *imp,
    int sx, int sy, int w, int h,
    int dx, int dy)
{
  (void) imp; (void) sx; (void) sy; (void) w; (void) h;
  (void) dx; (void) dy;
}


/*
 * Paint letter c to screen coordinate (x,y) using given font.
 */
void
maemo_draw_glyph(
    GdkFont *font,
    int x, int y,
    int c)
{
  guchar chr;

  assert (window_open);
  assert (font != NULL);

  /* paint letter */
  chr = c;
  gdk_draw_text(
      drawing_area->window,                    /* drawable */
      font,                                    /* font */
      gc,                                      /* graphics context */
      x, y,                                    /* position */
      &chr,                                    /* string */
      1);                                      /* string length */
}


/*
 * Retrieve information on common ASCII characters, and store that
 * information to an indexed glyph table.
 */
void
maemo_init_glyphs(
    GdkFont *font, struct glyph *glyphs, int nglyphs)
{
  int i;

  assert (font != NULL);
  assert (glyphs != NULL);
  assert (nglyphs > 0);

  /* zero glyph table */
  memset (glyphs, 0, sizeof (glyphs[0]) * nglyphs);

  /* initialize glyphs */
  for (i = 32; i < nglyphs; i++) {
    struct glyph *gp = &glyphs[i];
    char c;
    int w;
      
    /* invalidate fields used by bitmap fonts */
    gp->x = 0;
    gp->y = 0;
    gp->h = -1;

    /* retrieve extents for glyph i */
    c = (char) i;
    gdk_text_extents(
        font,
        &c,                                    /* string */
        1,                                     /* length */
        &gp->lead,                             /* lbearing */
        &gp->follow,                           /* rbearing */
        &w,                                    /* width */
        &gp->ascent,                           /* ascent */
        &gp->descent);                         /* descent */
    
    /* compute and store width */
    gp->w = w - gp->lead - gp->follow;

    /* add a few pixels for spacing lines */
    gp->descent += 4;
  }
}


/*
 * Draw area (sx1,sy1)-(sx2,sy2) of a multi-color image imp to screen at
 * position (dx1,dy1)-(dx2,dy2).
 *
 * For more information on image blitting, please see:
 * http://developer.gnome.org/doc/books/WGA/graphics-gdk-pixbuf.html
 */
void
bitblt(
    struct image *imp,
    int sx1, int sy1, int sx2, int sy2,
    int dx1, int dy1, int dx2, int dy2)
{
  struct reg_image *rip;

  assert (window_open);
  assert (imp != NULL);
  assert (imp->data != NULL);
  assert (sx1 < sx2);
  assert (sy1 < sy2);

  /* ensure that image pixbuf exists */
  rip = (struct reg_image*) imp->data;
  if (rip->pixbuf == NULL) {
    load_image(imp);
  }
  assert (rip->pixbuf != NULL);

  /* paint image to screen*/
  gdk_draw_pixbuf(
      drawing_area->window,                    /* drawable */
      gc,                                      /* GdkGC */
      rip->pixbuf,                             /* source pixbuf */
      sx1, sy1,                                /* src x, y */
      dx1, dy1,                                /* dest x, y */
      sx2 - sx1,                               /* size */
      sy2 - sy1,
      GDK_RGB_DITHER_NONE,                     /* dither */
      0,
      0);
}


/*
 * Clear graphics window while resetting any drawing modes.  As window is
 * only cleared once per frame, this is a good place to set up default
 * drawing modes.
 */
void
clear (void)
{
  assert (window_open);

  /* select background color */
  gdk_gc_set_rgb_fg_color (gc, &bg);

  /* paint whole screen with background color */
  gdk_draw_rectangle(
      drawing_area->window,                    /* widget */
      gc,                                      /* graphics context */
      1,                                       /* filled? */
      0, 0,                                    /* upper left coordinate */
      screenwidth(), screenheight());

  /* revert to foreground color */
  gdk_gc_set_rgb_fg_color (gc, &fg);
}


/*
 * Set background color used by clear().
 */
void
setbg(
    int r, int g, int b)
{ 
  assert (window_open);

  /* define color value, but do not make it active yet */
  bg.pixel = 0;
  bg.red   = r << 8;
  bg.green = g << 8;
  bg.blue  = b << 8;
}


/*
 * Set foregroung color.
 */
void
setfg(
    int r, int g, int b)
{
  assert (window_open);
  
  /* define color value */
  fg.pixel = 0;
  fg.red   = r << 8;
  fg.green = g << 8;
  fg.blue  = b << 8;
  
  /* set foreground color using unallocated rgb */
  gdk_gc_set_rgb_fg_color (gc, &fg);
}


/*
 * Draw line (x1,y1)-(x2,y2) using current pen.
 */
void
line(
    int x1, int y1, int x2, int y2)
{
  assert (window_open);

  /* paint line with current color */
  gdk_draw_line (drawing_area->window, gc, x1, y1, x2, y2);
}


/*
 * Draw filled triangle (x1,y1)-(x2,y2),-(x3,y3) using current pen.
 */
void
triangle(
    int x1, int y1, int x2, int y2, int x3, int y3)
{
  GdkPoint points[3];
  assert (window_open);

  /* set up coordinates */
  points[0].x = x1;
  points[0].y = y1;
  points[1].x = x2;
  points[1].y = y2;
  points[2].x = x3;
  points[2].y = y3;
  
  /* paint polygon */
  gdk_draw_polygon(
      drawing_area->window,
      gc,                                      /* graphics context */
      TRUE,                                    /* filled? */
      points,                                  /* polygon data */
      sizeof (points) / sizeof (points[0]));   /* number of points */
}


/*
 * Draw box (x1,y1) - (x2,y2) with a semi-transparent color.  Argument
 * alpha ranges from 0 to 255, and it defines how much the original color
 * shows through.  If alpha is 200, intensity of foreground color is 80%
 * while 20% of background color shows through.
 */
void
fade(
    int x1, int y1, int x2, int y2, int alpha)
{
  assert (window_open);

  /* FIXME: do something */
}


/*
 * Swap double buffers to show hidden buffer to user.
 */
void
swap_buffers (void)
{
#ifdef TICKS_PER_FRAME
  static ticks_t last = 0;
  ticks_t now;
#endif

  assert (window_open);

  /* copy off-screen drawing to on-screen */
  end_paint();
  
  /*
   * Peach Gardens is currently hidden by some other applications.  Release
   * non-essential graphics resources to allow this other applications to
   * run.
   */
  if (!active) {
    /* FIXME: do something */
  }

  /*
   * Read events from Maemo queue to internal queue.  If the applications is
   * deactivated, wait until more events are available.  Otherwise, return
   * as soon as there are no more events.
   */
  if (!exiting) {
    if (active) {
      /* process events once and exit */
      while (gtk_events_pending()  &&  active) {
        gtk_main_iteration_do (0);
      }
    } else {
      /* stay in event loop until activated */
      purge_images ();
      while (!active) {
        gtk_main_iteration_do (1);
      }
    }
  }

  /*
   * Switch between full-screen and windowed mode.  Mode is changed only
   * during buffer switches so that change does not interfere with drawing.
   */
  if (fullscreen_switch) {
    fullscreen_switch = 0;
    
    /* change window size */
    fullscreen = !fullscreen;
    hildon_appview_set_fullscreen (appview, fullscreen);
  }

  /*
   * Make display cycle take about the same time no matter how much actual
   * drawing takes.  This is mostly to make animation smooth, but this can
   * also save battery when there is very little drawing to do.  At least
   * the screen is not updated 1000 times per second.
   *
   * If defined, TICKS_PER_FRAME sets the minimum duration of a display
   * cycle in milliseconds.  This can be used for both slowing down frame
   * rate and making frame rate more reliable on varying CPU load.  The
   * feature can also be used for decreasing CPU load caused by the
   * application on a multi-user environments.
   */
#ifdef TICKS_PER_FRAME
  do{      
    now = ticks ();
    if (now - last < TICKS_PER_FRAME) {
      /* wait some micro seconds to lenghten the display cycle */
      usleep ((TICKS_PER_FRAME - (now - last)) * 1000);
      now = ticks ();
    }
  } while (now - last < TICKS_PER_FRAME);
  last = now;
#endif

  /* start a new drawing cycle */
  begin_paint ();
}


/*
 * In graphics interface of Peach Gardes, this function is supposed to flush
 * drawing queues and return immediately.  However, drawing queues in GTK
 * are managed automatically, and we have no control over the queues, so
 * nothing needs to be done at this time.
 */
void
flush_gfx (void)
{
  /*NOP*/;
}

  
/*
 * Start drawing to off-screen buffer.
 */
static void
begin_paint (void)
{
  assert (window_open);

#ifdef DOUBLE_BUFFER

  /* allocate off-screen buffer */
  if (visible == NULL) {
    visible = gdk_drawable_get_visible_region (drawing_area->window);
  }
  
  /* start drawing to off-screen buffer */
  gdk_window_begin_paint_region (drawing_area->window, visible);

#endif /*DOUBLE_BUFFER*/
}


/*
 * Finish drawing to an off-screen buffer, and copy contents of that buffer
 * to on-screen.
 */
static void
end_paint (void)
{
  assert (window_open);

#ifdef DOUBLE_BUFFER

  if (visible != NULL) {
    /* copy off-screen drawing area to on-screen */
    gdk_window_end_paint (drawing_area->window);
    
    /* destroy drawing region */
    gdk_region_destroy (visible);
    visible = NULL;
  }

#endif /*DOUBLE_BUFFER*/
}


/*
 * Get width of graphics screen.
 */
int
screenwidth (void)
{
  gint w;
  assert (window_open);
  
  gdk_drawable_get_size (drawing_area->window, &w, NULL);
  return w;
}


/*
 * Get height of graphics window.
 */
int
screenheight (void)
{
  gint h;
  assert (window_open);
  
  gdk_drawable_get_size (drawing_area->window, NULL, &h);
  return h;
}


/*
 * Open graphics window.
 */
int
open_screen (void)
{
  /*
   * Initialize graphics libraries.  This cannot be done in init_graphics,
   * as test programs will not start otherwise.
   */
  if (app == NULL) {
    osso_hw_state_t state;
    
    /*
     * Initialize gtk with application name.  This is important, as without
     * this Peach Gardens will not show up in application switcher.
     */
    {
      int argc = 1;                            /* number of args in argv */
      char appname[] = "pgardens";             /* argv[0] */
      char *args[] = { appname };              /* argv table */
      char **argv = args;                      /* pointer to pointer table */
      gtk_init (&argc, &argv);
    }
    
    /* create hildon application */
    app = HILDON_APP (hildon_app_new());
    assert (app != NULL);
    
    /* initialize system interface */
    osso_context = osso_initialize (OSSO_APP, VERSION, TRUE, NULL);
    if (osso_context == NULL) {
      /* failed to initialize osso */
      abort_msg("cannot initialize osso");
    }
    
    /* define notifications of interest */
    memset (&state, 0, sizeof (state));
    state.shutdown_ind = 1;
    state.memory_low_ind = 1;
    state.system_inactivity_ind = 1;
    state.save_unsaved_data_ind = 1;
    
    /* add handler for hardware D-BUS messages */
    if (osso_hw_set_event_cb(
        osso_context, &state, hw_event_handler, NULL) != OSSO_OK) {
      abort_msg("cannot add event callback");
    }
    
    /* add handler for Exit D-BUS messages */
    if (osso_application_set_exit_cb(
        osso_context, exit_event_handler, NULL) != OSSO_OK) {
      abort_msg("cannot add exit callback");
    }
  }
  
  /* begin the main application */
  if (app != NULL  &&  !window_open) {
    /* state-saving not supported: must not kill application */
    hildon_app_set_killable (app, 0);

    /* set application title */
    hildon_app_set_title (app, APP_TITLE);
    hildon_app_set_two_part_title (app, 0);

    /* create view */
    appview = HILDON_APPVIEW (hildon_appview_new(""));
    hildon_app_set_appview (app, appview);

    /* deny fullscreen key (it is handled manually) */
    hildon_appview_set_fullscreen_key_allowed (appview, 0);

    /* create drawing area with minimum size of 100x100 pixels */
    drawing_area = gtk_drawing_area_new();
    gtk_widget_set_size_request (drawing_area, 100, 100);
    gtk_container_add(GTK_CONTAINER(appview), drawing_area);

    /*
     * Turn off automatic double buffering.  We do double buffering on our
     * own, and system's buffering causes flicker when moving in menus.
     */
    gtk_widget_set_double_buffered (drawing_area, 0);

    /* create new graphics context for painting to drawing area */
    gc = gdk_gc_new (drawing_area->window);
    assert (gc != NULL);
    
    /* create menus */
    main_menu = hildon_appview_get_menu (appview);
    file_menu = gtk_menu_new();
    item_file = gtk_menu_item_new_with_label ("Game");
    gtk_menu_item_set_submenu(
        GTK_MENU_ITEM(item_file), file_menu);
    
    /* create menu items */
    item_new    = gtk_menu_item_new_with_label("New...");
    item_open   = gtk_menu_item_new_with_label("Load...");
    item_save   = gtk_menu_item_new_with_label("Save");
    item_saveas = gtk_menu_item_new_with_label("Save As...");
    item_full   = gtk_menu_item_new_with_label("Full Screen");
    item_close  = gtk_menu_item_new_with_label("Close");
    
    /* bind menu items to menu */
    gtk_menu_append (main_menu, item_file);
    gtk_menu_append (file_menu, item_new);
    gtk_menu_append (file_menu, item_open);
    gtk_menu_append (file_menu, item_save);
    gtk_menu_append (file_menu, item_saveas);
    gtk_menu_append (main_menu, item_full);
    gtk_menu_append (main_menu, gtk_separator_menu_item_new());
    gtk_menu_append (main_menu, item_close);

    /* gray out invalid menu items for now */
    gtk_widget_set_sensitive (item_save, 0);
    gtk_widget_set_sensitive (item_saveas, 0);
    
    /* register callbacks */
    g_signal_connect (G_OBJECT (item_close), "activate",
        GTK_SIGNAL_FUNC(quit), NULL);
    g_signal_connect (G_OBJECT (item_full), "activate",
        GTK_SIGNAL_FUNC(toggle_fullscreen), NULL);
    g_signal_connect (G_OBJECT (app), "im-close",
        GTK_SIGNAL_FUNC(im_close), NULL);
    g_signal_connect (G_OBJECT (app), "topmost-status-lose",
        GTK_SIGNAL_FUNC(deactivate), NULL);
    g_signal_connect (G_OBJECT (app), "topmost-status-acquire",
        GTK_SIGNAL_FUNC(activate), NULL);
    
    /* show menus */
    gtk_widget_show_all (GTK_WIDGET(main_menu));
    
    /* add hardware button listener */
    g_signal_connect (G_OBJECT(app),
        "key_press_event", G_CALLBACK(key_press), app);
    g_signal_connect (G_OBJECT(app),
        "key_release_event", G_CALLBACK(key_release), app);
    
    /* reset local variables */
    fullscreen_switch = 0;
    fullscreen = 0;
    visible = NULL;
    
    /* mark window initialized */
    window_open = 1;

    /* start drawing to an off-screen buffer */
    setbg (0, 0, 0);
    begin_paint ();
  }

  /* show main window */
  if (app != NULL  &&  window_open) {
    gtk_widget_show_all (GTK_WIDGET(appview));
    gtk_widget_show_all (GTK_WIDGET(app));
  }
  return window_open;
}


/*
 * In multi-device graphics interface of Peach Gardens, this function is
 * supposed to close graphics window temporarily.  However, as N770 tablet
 * has no console screen to return to, there is no much point closing the
 * graphics window.
 */
void
close_screen (void)
{
  if (window_open) {
    clear();
    purge_images();
    end_paint();
  }
}


/*
 * Switch from full-screen mode to windowed or vice versa.
 */
static void
toggle_fullscreen (void)
{
  /* trigger switch */
  fullscreen_switch = 1;
}


/*
 * Application has been de-activated.
 */
static void
deactivate(
    HildonAppView *hildonappview, gpointer user_data)
{
  (void) hildonappview; (void) user_data;
  active = 0;
}


/*
 * Application has been activated.
 */
static void
activate(
    HildonAppView *hildonappview, gpointer user_data)
{
  (void) hildonappview; (void) user_data;
  active = 1;
}


/*
 * Application closing.
 */
static void
im_close(
    HildonAppView *hildonappview, gpointer user_data)
{
  (void) hildonappview; (void) user_data;
  /* FIXME: do something */
}



/*
 * Callback for exit D-BUS event.
 */
static void
exit_event_handler(
    gboolean die_now, gpointer data)
{
  /* FIXME: do something */
  quit ();
}


/*
 * Callback for hardware D-BUS events.
 */
static void
hw_event_handler(
    osso_hw_state_t *state, gpointer data)
{
  (void) data;

  if (state->shutdown_ind) {
    gtk_infoprint (GTK_WINDOW(app), "Shutdown event!");
  }
  if (state->memory_low_ind) {
    /* memory low: purge pixbufs from memory */
    purge_images ();
    gtk_infoprint (GTK_WINDOW(app), "Running low on memory!");
  }
  if (state->save_unsaved_data_ind) {
    gtk_infoprint (GTK_WINDOW(app), "Must save unsaved data event!");
  }
  if (state->system_inactivity_ind) {
    gtk_infoprint(GTK_WINDOW(app), "Minimize application inactivity event!");
  }
}


/*
 * Callback for hardware keys.  Function key F1 acts as an alias for
 * full-screen key F6, as my personal keyboard has key F6 configured for
 * other purposes.
 */
static gboolean
key_press(
    GtkWidget *widget, GdkEventKey *event, HildonApp *app)
{
  (void) widget;

  switch (event->keyval) {
  case GDK_Up:
    push_event (UP | PRESS);
    return TRUE;
    
  case GDK_Down:
    push_event (DOWN | PRESS);
    return TRUE;
    
  case GDK_Left:
    push_event (LEFT | PRESS);
    return TRUE;
    
  case GDK_Right:
    push_event (RIGHT | PRESS);
    return TRUE;
    
  case GDK_Return:
    push_event (ACTION | PRESS);
    return TRUE;

  /* toggle between full-screen and windowed modes */
  case GDK_F1:
  case GDK_F6:
    toggle_fullscreen ();
    return TRUE;
    
  case GDK_F7:
    gtk_infoprint(GTK_WINDOW(app), "Increase (zoom in)");
    return TRUE;
    
  case GDK_F8:
    gtk_infoprint(GTK_WINDOW(app), "Decrease (zoom out)");
    return TRUE;
    
  case GDK_Escape:
    push_event (EXIT | PRESS);
    return TRUE;        
  }
  return FALSE;
}


/*
 * Key released.
 */
static gboolean
key_release(
    GtkWidget *widget, GdkEventKey *event, HildonApp *app)
{
  (void) widget; (void) app;

  switch (event->keyval) {
  case GDK_Up:
    push_event (UP | RELEASE);
    return TRUE;
    
  case GDK_Down:
    push_event (DOWN | RELEASE);
    return TRUE;
    
  case GDK_Left:
    push_event (LEFT | RELEASE);
    return TRUE;
    
  case GDK_Right:
    push_event (RIGHT | RELEASE);
    return TRUE;
    
  case GDK_Return:
    push_event (ACTION | RELEASE);
    return TRUE;
    
  case GDK_Escape:
    push_event (EXIT | RELEASE);
    return TRUE;        
  }
  return FALSE;
}


/*
 * Callback for "Close" menu entry.
 */
static void
quit (void)
{
  assert (window_open);
  gtk_infoprint(GTK_WINDOW(app), "quit");
  
  /* FIXME: ask if it is OK to close application */
  exiting = 1;
  
  /* FIXME: push event telling to quit no */
}


/*
 * Initialize graphics libraries with default settings.
 */
void
init_graphics (void)
{
  assert (!gfx_initialized);

  /* mark all images free */
  memset (&images[0], 0, sizeof (images));

  /* initialize other module local variables */
  app = NULL;
  exiting = 0;
  window_open = 0;
  fullscreen_switch = 0;
  gc = NULL;
  active = 1;
  nevents = 0;
  
  gfx_initialized = 1;
}


/*
 * De-initialize graphics libraries.  Here, almost everything is left
 * for the GTK.  In fact, I could not even find information on how to
 * destroy all the allocated data structures explicitly.  GTK does implement
 * reference counting, and I must rely on it here.
 */
void
done_graphics (void)
{
  if (gfx_initialized) {
    /* close graphics screen */
    if (window_open) {
      close_screen();
    }
    
    /* release graphics context */
    if (gc != NULL) {
      g_object_unref (gc);
      gc = NULL;
    }
    
    /* de-initialize graphics libraries */
    if (osso_context != NULL) {
      /* de-initialize osso */
      osso_deinitialize (osso_context);
      osso_context = NULL;
    }
    
    window_open = 0;
    gfx_initialized = 0;
  }
}


/*
 * Print error message to graphics window.
 */
void
error(
    const char *msg)
{
  if (app != NULL) {
    GtkWidget *dialog = hildon_note_new_information_with_icon_stock(
        GTK_WINDOW(app), msg, "qgn_note_gene_syserror");
    hildon_note_set_button_text (HILDON_NOTE(dialog), "OK");
    gtk_dialog_run (GTK_DIALOG(dialog));
    gtk_widget_destroy (dialog);
  }
}


/*
 * Register image for drawing.  Color map of an image should not be modified
 * after registration: possible changes to color map only take effect after
 * image is purged and reloaded to graphics memory.
 *
 * Also note that no graphics resources should be used at this time.
 * Graphics screen may not be open at this time.
 */
int
register_image(
    struct image *imp, const char *buffer, int len)
{
  int ok = 0;
  int i;
  struct reg_image *rip = NULL;
  
  assert (gfx_initialized);
  assert (imp != NULL);
  assert (buffer != NULL);
  assert (imp->data == NULL);
  assert (imp->width * imp->height <= len);
  
  /* find a free reg_image structure */
  for (i = 0; i < MAX_IMAGES; i++) {
    if (images[i].used == 0) {
      /* found an unused image */
      rip = &images[i];
      rip->used = 1;
      break;
    }
  }
  DEBUG (("%p: alloc reg_image", rip));
  if (rip != NULL) {
    /* initialize structure */
    rip->pixbuf = NULL;
    imp->data = rip;

    /* duplicate image data */
    rip->pixels = malloc (len);
    DEBUG (("%p: alloc pixels", rip->pixels));
    if (rip->pixels != NULL) {
      memcpy (rip->pixels, buffer, len);
      ok = 1;
    } else {
      out_of_memory();
    }
    
  } else {
    /* no slots left image array */
    abort_msg ("too many images");
  }
  
  if (ok == 0) {
    unregister_image (imp);
  }
  return ok;
}


/*
 * Unregister image to release all hardware specific resources allocated for
 * it.
 * 
 * Note that this very function can be called without init_graphics()
 * function being invoked first.  This may happen in batch programs that
 * only modify images, without any intent on showing them.
 */
void
unregister_image(
    struct image *imp)
{
  struct reg_image *rip;
  assert (imp != NULL);

  /* get pointer to registered image */
  rip = (struct reg_image*) imp->data;  
  if (rip != NULL) {
    /* has hardware dependent data */
    assert (&images[0] <= rip  &&  rip < &images[MAX_IMAGES]);
    assert (rip->used);

    /* release pixbuf */
    if (rip->pixbuf != NULL) {
      DEBUG (("%p: free pixbuf", rip->pixbuf));
      gdk_pixbuf_unref (rip->pixbuf);
    }

    /* release pixel data */
    if (rip->pixels != NULL) {
      DEBUG (("%p: free pixels", rip->pixels));
      free (rip->pixels);
    }

    /* destroy structure to detect dangling pointers */
#ifndef NDEBUG
    memset (rip, 0xcc, sizeof (*rip));
#endif
    
    /* free image structure */
    DEBUG (("%p: free reg_image", rip));
    rip->used = 0;
  }
}


/*
 * Ensure that image data is loaded onto gdk and create pixbuf object if
 * needed.  This allows drawing image to a display surface.
 */
static void
load_image(
    struct image *imp)
{
  struct reg_image *rip;

  assert (window_open);
  assert (imp != NULL);
  assert (imp->data != NULL);

  /* get pointer to registered image */
  rip = (struct reg_image*) imp->data;
  assert (rip->used);
  assert (rip->pixels != NULL);

  /* create pixbuf if needed */
  if (rip->pixbuf == NULL) {
    /* allocate rgb buffer for image */
    struct pixel *buf = (struct pixel*)
        malloc (imp->width * imp->height * sizeof(struct pixel));
    DEBUG (("%p: alloc rgb data", buf));
    if (buf != NULL) {
      /*
       * Copy image data from pixels array, where pixel is an 8-bit index to
       * color table, to pixel buffer that contains distinct red, green,
       * blue and alpha components.
       */
      int x, y;
      unsigned char *src = rip->pixels;
      struct pixel *pix = buf;
      for (y = 0; y < imp->height; y++) {
        for (x = 0; x < imp->width; x++, src++, pix++) {
          int idx;
          
          /* get pixel from buffer */
          idx = *src;
          
          /* fetch corresponding color from palette and store pixel */
          assert (0 <= idx  &&  idx < imp->ncolors);
          pix->r = imp->palette[idx].r;
          pix->g = imp->palette[idx].g;
          pix->b = imp->palette[idx].b;
          pix->a = imp->palette[idx].a;
        }
      }

      /* create pixbuf */
      rip->pixbuf = gdk_pixbuf_new_from_data(
          (guchar*) buf,
          GDK_COLORSPACE_RGB, 
          TRUE,                                /* alpha */
          8,                                   /* bits per sample */
          imp->width,                          /* width */
          imp->height,                         /* height */
          imp->width*sizeof(struct pixel),     /* rowstride */
          image_destroy,                       /* function to destroy buf */
          rip);                                /* argument to image_destroy */
      DEBUG (("%p: alloc pixbuf", rip->pixbuf));
      assert (rip->pixbuf != NULL);
      
    } else {
      out_of_memory();
    }
  }
}


/*
 * Release image data.  This will be called by gdk when a pixbuf object is
 * deleted.
 */
static void
image_destroy(
    guchar *pixels, gpointer data)
{
  (void) data;
  DEBUG (("%p: free rgb data", pixels));
  free (pixels);
}


/*
 * Purge gdk pixbufs from memory.  This is mostly for responding to low
 * memory events.  Purging images causes registered but currently unused
 * images to be removed from memory.  However, raw image data is still kept
 * in memory, and pixbuf objects will be re-created if needed.
 */
static void
purge_images (void)
{
  int i;
  struct reg_image *rip;
  
  for (i = 0; i < MAX_IMAGES; i++) {
    rip = &images[i];
    if (rip->used  &&  rip->pixbuf != NULL) {
      /* pixbuf allocated */
      DEBUG (("%p: free pixbuf", rip->pixbuf));
      gdk_pixbuf_unref (rip->pixbuf);
      rip->pixbuf = NULL;
    }
  }
}



/*
 * Initialize keyboard interface.  Here, nothing needs to be done, as
 * console mode programs are not supported (by this interface).
 */
void
init_key (void)
{
  /*NOP*/;
}


/*
 * De-initialize keyboard interface.  Keyboard handling is tightly
 * associated to window management, and interface will be de-initialized
 * implicitly when an application is closed.
 */
void
done_key (void)
{
  /*NOP*/;
}


/*
 * Get raw hardware specific keyboard event.
 */
int
key_event (void)
{
  /* get latest event */
  if (nevents > 0) {
    assert (nevents <= EV_BUFFER_LEN);
    return eventbuf[--nevents];
  } else {
    /* no events in queue */
    return NOEVENT;
  }
}


/*
 * Push event to event buffer.
 */
static void
push_event(
    int ev)
{
  /* insert event to buffer */
  if (nevents == 0) {
    /* empty */
    eventbuf[0] = ev;
    nevents = 1;
  } else if (nevents < EV_BUFFER_LEN) {
    /* room for one more event */
    memmove (&eventbuf[1], &eventbuf[0], nevents*sizeof (eventbuf[0]));
    eventbuf[0] = ev;
    nevents++;
  } else {
    /* buffer full */
    memmove(&eventbuf[1], &eventbuf[0], (EV_BUFFER_LEN-1)*sizeof(eventbuf[1]));
    eventbuf[0] = ev;
    DEBUG (("event buffer full, ignoring event %d",eventbuf[EV_BUFFER_LEN-1]));
  }
}


/*
 * Print help on X11 specific options.  As tablet device is not usually
 * connected to a terminal, nothing needs to be done here.
 */
void
print_graphics_options (void)
{
  /*NOP*/;
}


/*
 * Parse and read away argv[i] if the switch is a graphics related option.
 * The function should return the number of options it reads away.
 *
 * The function is un-implemented in Maemo interface, as regular users cannot
 * specify a command lines anyway.  There are currently no options for
 * developers either.
 */
int
read_graphics_options(
    int *argc, char **argv, int i)
{
  (void) argc; (void) argv; (void) i;
  /*NOP*/;
  return 0;
}

#endif /*MAEMO*/
