/*
 * map.c - map creation and drawing
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Dec 23 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: map.c,v 1.2 2005/04/03 16:31:54 tronkko Exp $
 *
 * $Log: map.c,v $
 * Revision 1.2  2005/04/03 16:31:54  tronkko
 * move with view, queue depth enabled chips for depth sorting
 *
 * Revision 1.1  2005/01/09 21:12:24  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "graphics.h"
#include "depth.h"
#include "chip.h"
#include "chipset.h"
#include "map.h"
#include "view.h"
#include "image.h"
#include "debug.h"


/*
 * Draw visible portion of a map to screen.  Objects with depth coordinates
 * will be queued for drawing.  Be sure to call zdraw() to draw these depth
 * enabled parts of the map.
 */
void
draw_map(
    struct view *vp, struct map *mp)
{
  int x1, y1, x2, y2;
  int x, y;
  int chipw, chiph;
  int sw, sh;
  int offx, offy;
  float s, s1;
  struct chipset *chp;

  /* check args */
  assert (mp != NULL);
  assert (mp->w > 0);
  assert (mp->h > 0);
  
  /* get chipset of the map */
  chp = mp->chp;
  assert (chp != NULL);
  
  /* chip width & height in chipset image */
  chipw = chp->chipw;
  chiph = chp->chiph;
  
  /* chip width & height in screen in pixels */
  s = scale(vp);
  s1 = 1.0 / s;
  sw = chipw * s;
  sh = chiph * s;
  
  /* origin of view */
  offx = viewx(vp);
  offy = viewy(vp);
  
  /* find out visible part of map */
  x1 = offx / chipw;
  if (x1 < 0) x1 = 0;
  
  y1 = offy / chiph;
  if (y1 < 0) y1 = 0;
  
  x2 = (offx + screenwidth()*s1) / chipw + 1;
  if (x2 > mp->w) x2 = mp->w;
  
  y2 = (offy + screenheight()*s1) / chiph + 1;
  if (y2 > mp->h) y2 = mp->h;
  
  /* draw visible part of map */
  for (y = y1; y < y2; y++) {
    for (x = x1; x < x2; x++) {
      int sx1, sy1, sx2, sy2;
      int dx1, dy1, dx2, dy2;
      struct map_block *mbp;
      struct chip *cp;
      
      /* get chip */
      mbp = get_block (mp, x, y);
      if (mbp->id > 0) {
        /* valid chip */
        cp = get_chip (chp, mbp->id);
        
        /* compute screen coordinates for the chip */
        dx1 = (x * chipw - offx) * s;
        dy1 = (y * chiph - offy) * s;
        dx2 = dx1 + chipw * s;
        dy2 = dy1 + chiph * s;
        
        /* draw background chip */
        if (cp->bg > 0) {
          /* get pointer to chip at background */
          struct chip *bg = get_chip (chp, cp->bg);
          assert (bg != NULL);
          
          /* compute source coordinate within chipset image */
          sx1 = bg->x * chipw;
          sy1 = bg->y * chiph;
          sx2 = sx1 + chipw;
          sy2 = sy1 + chiph;
          
          /* paint background chip */
          bitblt (&chp->img,
              sx1, sy1, sx2, sy2,
              dx1, dy1, dx2, dy2);
        }
        
        /* compute source coordinate within chipset image */
        sx1 = cp->x * chipw;
        sy1 = cp->y * chiph;
        sx2 = sx1 + chipw;
        sy2 = sy1 + chiph;
        
        /* paint chip  to screen */
        if (cp->z <= 0) {
          /* paint immediately */
          bitblt (&chp->img,
              sx1, sy1, sx2, sy2,
              dx1, dy1, dx2, dy2);
        } else {
          /* queue for drawing */
          zbitblt (&chp->img,
              dy1 + cp->z * s, 
              sx1, sy1, sx2, sy2,
              dx1, dy1, dx2, dy2);
        }
      }
      
    } /* for x */
  }
}


/*
 * Get pointer to map block at (x,y).  The function is only used in debug
 * mode.  Otherwise, get_block_macro() is used directly.
 */
struct map_block *
get_block_function(
    struct map *mp, int x, int y)
{
  /* check input args */
  assert (mp != NULL);
  assert (mp->data != NULL);
  assert (0 <= x  &&  x < mp->w);
  assert (0 <= y  &&  y < mp->h);

  /* return block */
  return get_block_macro (mp, x, y);
}


/*
 * Make map visible or hidden.
 */
void
set_map_visibility(
    struct map *mp, int visible)
{
  int x, y;

  for (y = 0; y < mp->h; y++) {
    for (x = 0; x < mp->w; x++) {
      struct map_block *mbp = get_block (mp, x, y);
      mbp->visible = visible;
    }
  }
}


/*
 * Initialize map by allocating memory for w x h map blocks.
 */
void
init_map(
    struct map *mp, int w, int h, struct chipset *chp)
{
  struct map_block *p;
  int size;

  /* check args */
  assert (mp != NULL);
  assert (3 < w  &&  w < 10000);
  assert (3 < h  &&  h < 10000);
  
  /* allocate memory for the map data */
  size = w * h * sizeof (struct map_block);
  p = malloc (size);
  DEBUG (("%p: alloc map", p));
  if (p) {
    /* duplicate map properties */
    mp->w = w;
    mp->h = h;
    
    /* bind chipset to the map */
    mp->chp = chp;

    /* reset map */
    mp->data = p;
    memset (p, 0, size);
  } else {
    out_of_memory ();
  }
}


/*
 * Release map from memory.
 */
void
done_map(
    struct map *mp)
{
  assert (mp != NULL);

  /* forget chipset */
  if (mp->chp != NULL) {
    mp->chp = NULL;
  }

  /* destroy data */
  if (mp->data != NULL) {
    DEBUG (("%p: free map", mp->data));
    free (mp->data);
    mp->data = NULL;
  }
}
