/*
 * options.c - read command line options
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Oct 31 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: options.c,v 1.3 2005/04/03 16:32:09 tronkko Exp $
 *
 * $Log: options.c,v $
 * Revision 1.3  2005/04/03 16:32:09  tronkko
 * not needing feat.h anymore
 *
 * Revision 1.2  2005/01/09 21:12:32  tronkko
 * updated copyright
 *
 * Revision 1.1  2004/11/20 10:59:24  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "options.h"
#include "split.h"
#include "graphics.h"


/*
 * Read command line arguments.  Arguments having some meaning to us are
 * read and removed from the arguments array.  Command line switches
 * accepted by the function are removed from the commands array.
 *
 * The function does not return on failure.
 */
void
read_options (int *argc, char **argv)
{
  int cur = 1;
  while (cur < *argc) {
    if (strcmp (argv[cur], "--") == 0) {
      /*
       * Signal the end of arguments.  Note that the flag is left in
       * the arguments array so that the caller can also note it.
       */
      break;
      
    } else if (argv[cur][0] != '-') {
      /* the very first file name signals the end of options */
      break;

    } else {
      int valid = 0;
      
      /* see if the argument belongs to the graphics sub-system */
      int opts = read_graphics_options (argc, argv, cur);
      if (opts > 0) {
        /* the option was read away and argc modified (do not touch cur) */
        valid = 1;
      }
      
      /*
       * The very first un-known option stops the processing of the command
       * line arguments.  This allows the caller to have its very own
       * options listed right after the general options.
       */
      if (!valid) {
        break;
      }
    }
  }
}


/*
 * Read and parse default parameters from an environmental variable.  The
 * variable can be used to set the default graphics mode, for example.
 * The default values set in the environment can be overridden in the
 * command line, though.
 *
 * The function refuses to accept input files or application specific
 * options from the environment.  Option -- is also un-supported.  The
 * function does not return in the case of error.
 */
void
get_defaults (void)
{ 
  /* get the variable */
  const char *options=getenv(DEFAULTS_VAR);
  if (options != NULL) {
    /* the variable is defined */
    char **ps2argv;
    int ps2argc;

    /* split the contents of the variable to the ps2argv array */
    if (split (&ps2argc, &ps2argv, options, " \t", 1)) {

      int i = 0;
      while (i < ps2argc) {
        int valid = 0;

        /* see if the option has something to do with graphics */
        int opts = read_graphics_options (&ps2argc, ps2argv, i);
        if (opts > 0) {
          /* the option was read away and ps2argc modified */
          valid = 1;
        }

        /* complain about invalid option */
        if (!valid)  {
          fprintf (stderr,
              "invalid option in %s: %s\n", DEFAULTS_VAR, ps2argv[i]);
          free (ps2argv);
          exit (EXIT_FAILURE);
        }
      }

      /* release the arguments array */
      free (ps2argv);
    } else {
      fprintf (stderr,
          "unterminated string in %s: %s\n", DEFAULTS_VAR, options);
      exit (EXIT_FAILURE);
    }
  }
}
