/*
 * ps2dev.cpp - interface to PS2 graphics library (libps2dev)
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Jan  3 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: ps2dev.c,v 1.6 2005/06/22 13:07:03 tronkko Exp $
 *
 * $Log: ps2dev.c,v $
 * Revision 1.6  2005/06/22 13:07:03  tronkko
 * added error function
 *
 * Revision 1.5  2005/04/03 16:32:23  tronkko
 * fade()
 *
 * Revision 1.4  2005/01/09 22:11:22  tronkko
 * added a texture slot for 1024x1024 texture, dropped all slots for
 * 512x512 textures
 *
 * Revision 1.3  2005/01/09 21:13:54  tronkko
 * bug fix: initialize texture wrap modes correctly
 *
 * Revision 1.2  2004/11/20 10:58:42  tronkko
 * major changes in texture handling and video mode initialization, added
 * signal handlers
 *
 * Revision 1.1  2004/11/02 18:48:07  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#ifdef PS2DEV

#if HAVE_UNISTD_H
# include <unistd.h>
#endif

#if HAVE_FCNTL_H
# include <fcntl.h>
#endif

#if HAVE_SYS_MMAN_H
# include <sys/mman.h>
#endif

#ifdef HAVE_SIGNAL_H
# include <signal.h>
#endif

#include <ps2gs.h>
#include <ps2dma.h>

#include "graphics.h"
#include "options.h"
#include "image.h"
#include "input.h"
#include "tty.h"
#include "ticks.h"
#include "debug.h"


/* set to 1 to optimize drawing performance with queues */
#define ASYNCHRONOUS 1

/* number of 4K pages to allocate for dma buffers */
#define MAX_PAGES 1000 /*200=820K*/

/* size of cpu page  */
#define PAGE_SIZE 4096

/* size of buffer in scratchpad RAM */
#define SPR_BUFFER_SIZE 0x2000

/* starting offsets of scratchpad buffers (in spr ram) */
#define SPR_BUFFER_1 0x0000
#define SPR_BUFFER_2 (SPR_BUFFER_1 + SPR_BUFFER_SIZE)

/* pixel format of frame buffer */
#define PIXEL_FORMAT PS2_GS_PSMCT32

/* number of bytes per pixel in frame buffer */
#define BYTES_PER_PIXEL 4

/* number of bytes in gs memory */
#define GS_MEMORY_SIZE 0x400000

/* min & max number of textures in gs memory */
#define MIN_RESIDENT_TEXS 4
#define MAX_RESIDENT_TEXS 20

/* maximum width of texture */
#define MAX_TEXTURE_WIDTH 1024

/* byte size of a palette in gs memory */
#define PALETTE_SIZE (256*4)

/* transform clut index according to storage mode 1 */
#define CLUTSWIZZLE(i) (((i)&~0x18) | (((i)>>1) & 0x8) | (((i)<<1) & 0x10))

/* set priviledged gs register r to value v */
#define PRIVGS(r,v) \
  do{ \
    int error; \
    struct ps2_gssreg _gsreg; \
    _gsreg.reg = (r); \
    _gsreg.val = (__u64) (v); \
    error = ioctl (ps2gs, PS2IOC_SGSSREG, &_gsreg); \
    assert (error == 0); \
  }while(0)


/* type of previous giftag (for giftags that are modified) */
enum gif_type {
  None, Line, Triangle, BitBlt, Blend, Fade
};


/*
 * Image registered for drawing.  The structure contains the gs specific
 * variables that are needed to maintain the image in dma buffer and in gs
 * memory (where it is called texture).  The general parameters are kept
 * with the generic image structure.
 */
struct reg_image
{
  void *data;                                  /* image data in giftags */
  int data_len;                                /* length of image data */
  int data_size;                               /* max size of data */
  int resident;                                /* is resident? */
  int buffer;                                  /* texture index */
  struct transferenv *image;                   /* for fixing dest address */
  struct transferenv *palette;
};


/*
 * Structure containing giftags that clear the screen and reset any drawing
 * modes.  As a screen is cleared only once per frame, this is a good place
 * to change various gs registers to their common value.  This way the
 * registers do not need to be modified during drawing.
 * 
 * Note that ps2_giftag structure must be at offset 0 in any structure that
 * is queued to gs.  Never insert fields at the beginning of the structure.
 * Also remember that a certain memory layout is specified in giftag.  Never
 * re-order fields without considering the addressing mode.
 */
struct resetenv {
  ps2_giftag giftag;
  __u64 xyoffs;                                /* window offset */
  __u64 xyoffs_adr;
  __u64 scissor;                               /* clipping region */
  __u64 scissor_adr;
  __u64 prim;                                  /* sprite */
  __u64 prim_adr;
  __u64 rgb0;                                  /* bg color */
  __u64 rgb0_adr;
  __u64 v0;                                    /* cleared rectangle */
  __u64 v0_adr;
  __u64 v1;
  __u64 v1_adr;
  __u64 tex1;                                  /* texture lod and mipmap */
  __u64 tex1_adr;
  __u64 texa;                                  /* texture alpha */
  __u64 texa_adr;
  __u64 fba;                                   /* alpha correction */
  __u64 fba_adr;
  __u64 alpha;                                 /* alpha blending */
  __u64 alpha_adr;
  __u64 rgb1;                                  /* fg color */
  __u64 rgb1_adr;
  __u64 clamp;                                 /* texture clamp mode */
  __u64 clamp_adr;
  __u64 prmodecont;
  __u64 prmodecont_adr;
};


/*
 * Structure for drawing single-colored lines.  If multiple lines are drawn
 * in sequence, the structure is changed by adding more vertices in which
 * case the size of the structure larger that what would appear at first
 * sight.
 */
struct lineenv {
  ps2_giftag giftag;
  struct {
    __u64 v0;
    __u64 v1;
  } coords[1];
};


/*
 * Structure for drawing single-colored triangles.  If multiple triangles
 * are drawn in sequence, the structure is changed by adding more vertices
 * in which case the size of the structure larger that what would appear at
 * first sight.
 *
 * BEWARE: all dma transfers must be divisable by 16 bytes.  As the number
 * of vertices is odd with triangle, the prerequisite may not hold with the
 * trienv structure.  To fix that, we need to insert a dummy field just to
 * align the structure properly.
 */
struct trienv {
  ps2_giftag giftag;
  struct {
    __u64 rgb;
    __u64 v0;
    __u64 v1;
    __u64 v2;
  } coords[1];
};


/*
 * Structure for changing drawing mode and pen color.  If you insert fields
 * here, be sure to aligned the structure to 16 bytes.  In fact, that is why
 * the rgb was added to the structure at the first place.
 *
 * This very structure is often used as prim register cannot be specified in
 * giftag when using reglist mode.  Moreover, gif packets must be aligned to
 * 128 bytes, so a single field complicates construction of gif packets too
 * much.  It is easier to send prim register as a separate packet when
 * drawing primitives can be easily aligned.
 */
struct primenv {
  ps2_giftag giftag;
  __u64 prim;
  __u64 rgb;
};


/*
 * Structure for switching drawing buffers.  Also disables z-buffer.
 */
struct drawbufenv
{
  ps2_giftag giftag;
  __u64 test;
  __u64 test_adr;
  __u64 frame;
  __u64 frame_adr;
  __u64 zbuf;
  __u64 zbuf_adr;
  __u64 scanmsk;
  __u64 scanmsk_adr;
};


/*
 * Structure containing parameters for image/palette transfer.  The
 * structure is used when transferring data from main memory to gs.
 */
struct transferenv {
  ps2_giftag giftag;
  __u64 bitbltbuf;
  __u64 bitbltbuf_adr;
  __u64 trxpos;
  __u64 trxpos_adr;
  __u64 trxreg;
  __u64 trxreg_adr;
  __u64 trxdir;
  __u64 trxdir_adr;
};


/*
 * Structure containing pixels.  The structure is used for transferring
 * textures to gs.
 */
struct dataenv {
  ps2_giftag giftag;
  struct {
    __u64 lo;
    __u64 hi;
  } data[1];
};


/*
 * Structure containing palette colors.  The structure is used for
 * transferring color lookup tables to gs.
 */
struct paletteenv {
  ps2_giftag giftag;
  struct {
    unsigned char r;
    unsigned char g;
    unsigned char b;
    unsigned char a;
  } color[1];
};


/*
 * Flush image cache and stop dma packet.
 */
struct imageenv {
  ps2_giftag giftag;
  __u64 texflush;
  __u64 texflush_adr;
};


/*
 * Structure for specifying texture.  This is used when drawing images and
 * fonts.  The fragment color is only needed when drawing fonts, though.
 */
struct textureenv {
  ps2_giftag giftag;
  __u64 tex0;
  __u64 tex0_adr;
  __u64 prim;
  __u64 prim_adr;
  __u64 rgb;                                   /* fragment color */
  __u64 rgb_adr;
};


/*
 * Structure for drawing with bitblt().
 */
struct bitbltenv {
  ps2_giftag giftag;
  struct {
    __u64 t0;
    __u64 v0;
    __u64 t1;
    __u64 v1;
  } rect[1];
};


/*
 * Structure for fading.  Color of the fade will be set using primenv
 * structure.
 */
struct fadeenv {
  ps2_giftag giftag;
  struct {
    __u64 v0;
    __u64 v1;
  } rect[1];
};


/*
 * Screen settings.  The following definitions have been taken from file
 * ps2gs.c found at directory drivers/video under Linux kernel source tree.
 * The code is copyrighted by Sony Computer Entertainment Inc.  Please see
 * the kernel source files for more info.  Fields involved with
 * synchronization has been removed.
 */
#define vSMODE1(VHP,VCKSEL,SLCK2,NVCK,CLKSEL,PEVS,PEHS,PVS,PHS,GCONT,SPML,PCK2,XPCK,SINT,PRST,EX,CMOD,SLCK,T1248,LC,RC)	\
  (((__u64)(VHP)<<36)   | ((__u64)(VCKSEL)<<34) | ((__u64)(SLCK2)<<33) | \
   ((__u64)(NVCK)<<32)  | ((__u64)(CLKSEL)<<30) | ((__u64)(PEVS)<<29)  | \
   ((__u64)(PEHS)<<28)  | ((__u64)(PVS)<<27)    | ((__u64)(PHS)<<26)   | \
   ((__u64)(GCONT)<<25) | ((__u64)(SPML)<<21)   | ((__u64)(PCK2)<<19)  | \
   ((__u64)(XPCK)<<18)  | ((__u64)(SINT)<<17)   | ((__u64)(PRST)<<16)  | \
   ((__u64)(EX)<<15)    | ((__u64)(CMOD)<<13)   | ((__u64)(SLCK)<<12)  | \
   ((__u64)(T1248)<<10) | ((__u64)(LC)<<3)      | ((__u64)(RC)<<0))
#define vSYNCH1(HS,HSVS,HSEQ,HBP,HFP)	\
  (((__u64)(HS)<<43) | ((__u64)(HSVS)<<32) | ((__u64)(HSEQ)<<22) | \
   ((__u64)(HBP)<<11) | ((__u64)(HFP)<<0))
#define vSYNCH2(HB,HF) \
  (((__u64)(HB)<<11) | ((__u64)(HF)<<0))
#define vSYNCV(VS,VDP,VBPE,VBP,VFPE,VFP) \
  (((__u64)(VS)<<53) | ((__u64)(VDP)<<42) | ((__u64)(VBPE)<<32) | \
   ((__u64)(VBP)<<20) | ((__u64)(VFPE)<<10) | ((__u64)(VFP)<<0))

struct rdisplay {
  int magv, magh, dy, dx;
};
#define vDISPLAY(DH,DW,MAGV,MAGH,DY,DX) \
  { (MAGV), (MAGH), (DY), (DX) }
#define wDISPLAY(DH,DW,MAGV,MAGH,DY,DX) \
  (((__u64)(DH)<<44) | ((__u64)(DW)<<32) | ((__u64)(MAGV)<<27) | \
   ((__u64)(MAGH)<<23) | ((__u64)(DY)<<12) | ((__u64)(DX)<<0))

struct syncparam {
  int modename;
  int width, height, rheight, dvemode;
  __u64 smode1, smode2, srfsh, synch1, synch2, syncv;
  struct rdisplay display;
};

/* type of signal handler function */
typedef void sigaction_handler (int, siginfo_t*, void *);


/* acquire virtual console */
static int acquire_vc (void);

/* release virtual console */
static int release_vc (void);

/* get a piece of dma buffer for sending giftag to gs */
static void *get_giftag (size_t size, enum gif_type type);

/* post GIFtag to gs queue */
static void queue_giftag (void *p);

/* flush spr buffers */
static void flush_spr (void);

/* wait until DMA transfers have been completed */
static void wait_gs (void);

/* allocate dma buffers */
static void alloc_spr_buffers (void);

/* set display buffer */
static void set_dispbuf (int i);

/* set draw buffer */
static void set_drawbuf (int i);

/* release one texture from gs buffer */
static void purge_texture (int i);

/* remove all textures from gs memory */
static void purge_textures (void);

/* load texture to gs */
static void load_texture (struct image *imp);

/* set up gs mode to use a texture */
static void use_texture (int i, int mode);

/* allocate locked memory for dma transfers */
static void *alloc_locked (int bytes);

/* release locked memory */
static void free_locked (void *p, int bytes);

/* handle sudden termination */
static void sudden_death (int sig, int code, struct sigcontext *ctx);

/* handle ctrl+z */
static void suspend (int sig, int code, struct sigcontext *ctx);
static void resume (int sig, int code, struct sigcontext *ctx);
  
/* enter hibernation mode after ctrl+z */
static void hibernate (void);

/* initialize video mode */
static void init_video (void);
  

/* have the necessary structures been initialized? */
static int initialized = 0;

/* screen resolution */
static int resolution = PS2_GS_640x480;

/* graphics mode */
static int graphics_mode = PS2_GS_PAL;

/* width and height of graphics screen (in pixels) */
static int scrwidth = 640;
static int scrheight = 480;

/* are we in graphics screen? */
static int in_graphics_mode = 0;

/* user customizable colors */
static __u64 bgcolor = PS2_GS_SETREG_RGBAQ(0,0,0,0,0);
static __u64 fgcolor = PS2_GS_SETREG_RGBAQ(255,255,255,0,0);

/* x and y offset of pixel at upper left corner of screen */
static int xoffs = 0;
static int yoffs = 0;

/* address and size of scratchpad RAM */
static char *spr_adr = (char*) -1;
static size_t spr_size = 0;

/* start of current gs buffer (in scratchpad ram) */
static char *spr_start = NULL;

/* pointer to end of current spr buffer */
static char *spr_end = NULL;

/* spr buffers allocated for communicating with gs */
static char *spr_buffers[2] = { 0 };

/* index of current scratchpad buffer */
static int cur_spr_buffer = -1;

/* pointer to last giftag in spr buffer */
static ps2_giftag *cur_giftag = NULL;

/* drawing primitive pointed by cur_giftag */
static enum gif_type cur_primitive = None;

/* buffer used for drawing/display */
static int display_buffer = 0;
static int draw_buffer = 1;

/* addresses of frame buffers */
static unsigned int frame_adr[2] = { 0 };

/* registered images */
static struct reg_image images[MAX_IMAGES];

/* addresses of textures/palettes in gs memory */
static unsigned int tex_adr[MAX_RESIDENT_TEXS] = { 0 };
static unsigned int palette_adr[MAX_RESIDENT_TEXS] = { 0 };

/* free/allocated status of textures and palettes */
static struct reg_image *tex_owner[MAX_RESIDENT_TEXS] = { 0 };

/* number of usable textures/palettes */
static int nbuffers = 0;

/* when was the last time the resident texture was used */
static unsigned int tex_time[MAX_RESIDENT_TEXS] = { 0 };

/* counter for aging resident textures */
static unsigned int tex_counter = 0;

/* current texture loaded into tex0 register */
static int cur_texture = -1;

/*
 * Pixel sizes of gs memory regions that are allocated for textures.  At
 * least MIN_RESIDENT_TEXS slots are allocated at all times.  If memory
 * allows, all MAX_RESIDENT_TEXS buffers are allocated.  Note that GS has
 * only 4 megabytes of memory, and two display buffers take 2621440 bytes of
 * memory (640*512*2*4), and only the remaining memory is available for 
 * allocation.
 *
 * Also note that the selection of texture slots is highly dependent on the
 * game.  In Peach Gardens, 1024x1024 slot is taken by current chipset
 * image, one of 128x128 slots is used for 10x16 font while the remaining
 * 128x128 slots are for objects.  Furthermore, 256x256 slot is used by
 * 16x24 font.  Also note that there is no slot for 512x512 images as Peach
 * Gardens does not need those.
 */
static const int tex_size[MAX_RESIDENT_TEXS] = {
  1024, 256, 128, 64, 
  128, 128, 128, 128, 
  128, 128, 128, 128, 
  128, 128, 128, 128, 
  128, 128, 128, 128
};

/* handle to /dev/ps2gs devce */
static int ps2gs = -1;

/* handle to scratchpad RAM device */
static int ps2spr = -1;

/* handle to ps2mem device */
static int ps2mem = -1;

/* start of memory allocated for dma transfers between core and gs */
static void *page_adr = (void*) -1;

/* page allocation status */
static int page_used[MAX_PAGES] = { 0 };

/* ctrl+z pressed to suspend execution */
static int suspend_request = 0;

/* return to graphics mode at resume? */
static int resume_graphics = 0;

/* measure length of display cycle? */
static int measure = 0;

/*
 * The table below is also taken from Linux kernel source, and the
 * modes not used by Peach Gardens have been removed.  Moreover, the video
 * output mode has been added.
 */
static const struct syncparam syncdata[] = {
  /* ntsc interlace 640x480 */
  { PS2_GS_NTSC, 
    640, 480, 448, 0,
    vSMODE1(0, 1,1,1,1, 0,0, 0,0,0,4,0,0,1,1,0,2,0, 1,32,4), 1, 8,
    vSYNCH1(254,1462,124,222,64), vSYNCH2(1652,1240),
    vSYNCV(6,480,6,26,6,1), vDISPLAY(479,2559,0,3,50,632) },

  /* pal interlace 640x512 */
  { PS2_GS_PAL, 
    640, 576, 512, 1,
    vSMODE1(0, 1,1,1,1, 0,0, 0,0,0,4,0,0,1,1,0,3,0, 1,32,4), 1, 8,
    vSYNCH1(254,1474,127,262,48), vSYNCH2(1680,1212),
    vSYNCV(5,576,5,33,5,1), vDISPLAY(575,2559,0,3,72,652) },
  
  /* vesa 640x480 59.940Hz */
  { PS2_GS_VESA, 
    640, 480, -1, 2,
    vSMODE1(1, 0,1,1,1, 0,0, 0,0,0,2,0,0,1,0,0,0,0, 1,15,2), 0, 4,
    vSYNCH1(192,608,192,84,32), vSYNCH2(768,524),
    vSYNCV(2,480,0,33,0,10), vDISPLAY(479,1279,0,1,34,276) }
};


/*
 * Allocate size bytes from scratchpad RAM for sending a GIFtag to gs.  If
 * there is no room in scratchpad, flush DMA queues to release some space.
 *
 * The function is placed here at the top of this very source file so that
 * gcc could better inline the function.
 */
static void *
get_giftag(
    size_t size, enum gif_type type)
{
  void *p;

  /* flush buffers to make more room available */
  if (spr_end + size > spr_start + SPR_BUFFER_SIZE) {
    flush_spr ();
  }
  assert (spr_end + size <= spr_start + SPR_BUFFER_SIZE);

  /* allocate region from scratchpad RAM */
  p = spr_end;
  spr_end += size;

  /* remember the most recent giftag */
  cur_primitive = type;  
  return p;
}


/*
 * Send buffer p in scratchpad RAM to DMA queue for sending a GIFtag to 
 * gs.  The buffer must have been allocated by get_giftag().
 */
static void
queue_giftag(
    void *p)
{
  assert (spr_start <= (char*)p  &&  (char*)p < spr_end);
  
  /* make giftag p the last giftag of the packet */
  cur_giftag = (ps2_giftag*) p;  
  assert (cur_giftag->EOP == 0);

  /* send giftag immediately to gs */
#if !ASYNCHRONOUS
  flush_spr ();
#endif 
}


/*
 * Load texture into gs memory.  If necessary, the least frequently used
 * texture of suitable size is purged from gs memory to make room for the
 * new texture.  (However, note that a texture may be purged even if there
 * is plenty of gs memory left as we only use memory especially allocated
 * for the textures.  Furthermore, only textures of similar size are
 * considered for purging.)
 *
 * Note that the function is not ideal when working with limited memory.
 * The problem with the routine is that it cannot rescale the memory blocks
 * set at the initialization time.  For example, with 800x600 resolution
 * there is only space for one 512x512 texture, four 256x256 or sixteen
 * 128x128 textures.  If the space is allocated dynamically, textures of all
 * sizes can be used.  This very function, however, does not know how to
 * change the memory alignments such that the space allocated for a 512x512
 * texture could be later divided among four textures of 256x256 each.  The
 * buffers are set in advance.
 */
static void
load_texture(
    struct image *imp)
{
  int i;
  struct reg_image *it;
  int sz;
  struct ps2_packet dma;
  int oldest = -1;                             /* index of oldest texture */
  unsigned int maxcount = tex_counter;

  /*
   * Flush scratchpad to gs so that all current primitives are drawn
   * before transferring the texture to gs.  This may be important if we
   * replace an existing texture.
   */
  flush_spr ();
  
  /* find out image tag */
  it = (struct reg_image*) imp->data;
  assert (it != NULL);
  assert (&images[0] <= it  &&  it < &images[MAX_IMAGES]);
  assert (it->data != NULL);
    
  /*
   * Select texture slot for the image.  The logic below should be
   * considered together with tex_size table.  In all cases there must be a
   * slot for any given texture.  For example, Peach Gardens does not have a
   * slot for 512x512 textures, so the next suitable block must be used.
   * That is why one if is currently commented out.
   */
  sz = imp->width;
  if (imp->height> sz) sz = imp->height;
  if (sz < 0x40)       sz = 0x40;
  else if (sz < 0x80)  sz = 0x80;
  else if (sz < 0x100) sz = 0x100;
/*else if (sz < 0x200) sz = 0x200;*/
  else if (sz < 0x400) sz = 0x400;
  assert (sz <= MAX_TEXTURE_WIDTH);
  
  /* find space from gs memory */
  for (i = 0; i < nbuffers; i++) {    
    if (tex_size[i] == sz) {
      /* slot is suitable for the texture */
      if (tex_owner[i] == NULL) {
        /* found a free slot */
        break;
      } else {
        /* remember oldest allocated slot of suitable size */
        if (oldest < 0  ||  tex_time[i] < maxcount) {
          /* texture in this slot has not been used for a while */
          oldest = i;
          maxcount = tex_time[i];
        }
      }      
    }
  }
  
  /* replace the oldest texture */
  if (i >= nbuffers) {
    /*
     * The assertion below will fail if there was no suitable slot in GS
     * memory.  This might be because of image size.  That is, there are no
     * slots for 512x512 images.
     */
    assert (oldest >= 0);
    
    i = oldest;
    purge_texture (i);
  }
  
  /* set destination address of texture data */
  assert (it->image != NULL);
  assert ((tex_size[i] % 64) == 0);
  it->image->bitbltbuf =
      PS2_GS_SETREG_BITBLTBUF(
          0, 0, 0,                             /* source (in gs) */
          tex_adr[i],                          /* dest buffer ptr */
          tex_size[i]/64,                      /* width/64 */
          PS2_GS_PSMT8);                       /* pixel format */
  
  /* set destination address of palette data */
  assert (it->palette != NULL);
  assert ((tex_size[i] % 64) == 0);
  it->palette->bitbltbuf =
      PS2_GS_SETREG_BITBLTBUF(
          0, 0, 0,                             /* source addr (in gs) */
          palette_adr[i],                      /* dest buffer ptr */
          16/16,                               /* width/64 */
          PS2_GS_PSMCT32);                     /* pixel format */
  
  /* post the texture with direct non-blocking dma to gs */
  dma.ptr = it->data;
  dma.len = it->data_len;
  if (ioctl (ps2gs, PS2IOC_SENDA, &dma) != /*OK*/0) {
    DEBUG(("cannot send DMA packet: addr %p, size %d", dma.ptr, dma.len));
    abort_msg ("DMA error");
  }
    
  /* mark the texture resident */
  tex_owner[i] = it;
  it->buffer   = i;
  it->resident = 1;
  
  /* force update of texture environment */
  cur_texture = -1;
}


/*
 * Set up texture environment to use ith texture with mode.  Mode is
 * typically PS2_GS_TEX_TFX_MODULATE or PS2_GS_TEX_TFX_DECAL.  The first one
 * is used with fonts whereas the latter is used with regular images.
 */
static void 
use_texture(
    int i, int mode)
{
  int tw;
  struct textureenv *pr;

  /* create new giftag */
  pr = get_giftag (sizeof (*pr), None);
  PS2_GIFTAG_CLEAR_TAG(&pr->giftag);   
  pr->giftag.NLOOP = sizeof(*pr)/16 - 1;
  pr->giftag.NREG  = 1;
  pr->giftag.REGS0 = PS2_GIFTAG_REGS_AD;       /* address mode */
  
  /* compute texture width (2^tw == width) */
  if (tex_size[i] <= 64)       tw = 6;
  else if (tex_size[i] <= 128) tw = 7;
  else if (tex_size[i] <= 256) tw = 8;
  else if (tex_size[i] <= 512) tw = 9;
  else tw = 10;
  
  /* texture parameters */
  pr->tex0 = PS2_GS_SETREG_TEX0(
      tex_adr[i],                              /* texture base ptr */
      tex_size[i] / 64,                        /* buffer width */
      PS2_GS_PSMT8,                            /* pixel storage format */
      tw,                                      /* texture width */
      tw,                                      /* height */
      PS2_GS_TEX_TCC_RGBA,                     /* color component */
      mode,                                    /* texture function */
      palette_adr[i],                          /* clut base pointer */
      PS2_GS_PSMCT32,                          /* clut pixel storage fmt */
      PS2_GS_TEX_CSM_CSM1,                     /* clut storage mode */
      0,                                       /* clut offset */
      PS2_GS_TEX_CLD_LOAD);                    /* clut buffer load ctrl */
  pr->tex0_adr = PS2_GS_TEX0_1;
  
  /* enable alpha blending */
  pr->prim = PS2_GS_SETREG_PRIM(
      PS2_GS_PRIM_PRIM_SPRITE,                 /* draw primative */
      PS2_GS_PRIM_IIP_FLAT,                    /* colour type */
      PS2_GS_PRIM_TME_ON,                      /* enable texture */
      PS2_GS_PRIM_FGE_OFF,                     /* fog off */
      PS2_GS_PRIM_ABE_ON,                      /* Alpha BlEnd */ 
      PS2_GS_PRIM_AA1_OFF,                     /* AA off */
      PS2_GS_PRIM_FST_UV,                      /* texture coord */
      PS2_GS_PRIM_CTXT_CONTEXT1,               /* context set to 1 */
      PS2_GS_PRIM_FIX_NOFIXDDA);               /* dont fix interpolation */
  pr->prim_adr = PS2_GS_PRIM;

  /* fragment color (for fonts) */
  pr->rgb = fgcolor;
  pr->rgb_adr = PS2_GS_RGBAQ;
  
  /* send giftag to gs */
  queue_giftag (pr);
  
  /* no need to set drawing env if using the very same texture again */
  cur_texture = i;
}


/*
 * Blend rectangular sub-image (sx,sy)-(sx+w,sy+h) with foreground color to
 * frame buffer at (dx,dy).  This is typically used for printing fonts.
 */
void
textblt(
    struct image *imp,
    int sx, int sy, int w, int h,
    int dx, int dy)
{
  int i;                                       /* current texture */
  struct reg_image *it;
  
  assert (w > 0  &&  h > 0);
  assert (in_graphics_mode);
  assert (imp != NULL);
  assert (imp->width > 0);
  assert (0 <= sx  &&  sx+w <= imp->width);
  assert (0 <= sy  &&  sy+h <= imp->height);
  
  /* find out image tag */
  it = (struct reg_image*) imp->data;
  assert (it != NULL);
  assert (&images[0] <= it  &&  it < &images[MAX_IMAGES]);
  assert (it->data != NULL);
  
  /* load texture to gs unless it is already resident */
  if (!it->resident) {
    load_texture (imp);    
  }
  i = it->buffer;
  assert (0 <= i  &&  i < nbuffers);
  
  /*
   * Set up drawing environment for using the texture.
   */
  if (cur_texture != i  ||  cur_primitive != Blend) {
    use_texture (i, PS2_GS_TEX_TFX_MODULATE);
  }

  /* wrap drawing primitive to a gif packet */
  {
    int j;                                     /* index to rect */
    int update;                                /* 1=fix existing packet */
    struct bitbltenv *pr;
    int wx1, wy1, wx2, wy2;                    /* dest in window coords */
    
    if (cur_primitive != Blend
        ||  spr_end + sizeof(struct bitbltenv) > spr_start+SPR_BUFFER_SIZE)
    {
      /* create new giftag */
      pr = get_giftag (sizeof (*pr), Blend);    
      PS2_GIFTAG_CLEAR_TAG(&pr->giftag);   
      pr->giftag.NREG  = 4;
      pr->giftag.REGS0 = PS2_GIFTAG_REGS_UV;
      pr->giftag.REGS1 = PS2_GIFTAG_REGS_XYZF2;
      pr->giftag.REGS2 = PS2_GIFTAG_REGS_UV;
      pr->giftag.REGS3 = PS2_GIFTAG_REGS_XYZF2;
      pr->giftag.FLG   = PS2_GIFTAG_FLG_REGLIST;
      
      /* do post giftag */
      update = 0;
      j = 0;        
    } else {    
      /* update existing giftag */
      assert (cur_giftag != NULL);
      pr = (struct bitbltenv*) cur_giftag;
      j = pr->giftag.NLOOP;
      update = 1;
    }
    
    /*
     * Convert destination address to window coordinates.  Here, we must
     * substract 8 (0.5) from the output address in order to align textures
     * with other graphic.
     */    
    wx1 = xoffs + dx*16 - 8;
    wy1 = yoffs + dy*16 - 8;
    wx2 = wx1 + w*16;
    wy2 = wy1 + h*16;
    
    /* texture must fall within screen */
    assert (0 <= wx1  &&  wx1 <= 0xffff);
    assert (0 <= wy1  &&  wy1 <= 0xffff);
    assert (0 <= wx2  &&  wx2 <= 0xffff);
    assert (0 <= wy2  &&  wy2 <= 0xffff);
    
    /* insert rectangle to giftag */
    pr->rect[j].t0 = PS2_GS_SETREG_UV ((sx<<4), (sy<<4));
    pr->rect[j].v0 = PS2_GS_SETREG_XYZF2(wx1, wy1, 0xffffff, 0xff);
    pr->rect[j].t1 = PS2_GS_SETREG_UV ((sx<<4)+w*16, (sy<<4)+h*16);
    pr->rect[j].v1 = PS2_GS_SETREG_XYZF2(wx2, wy2, 0xffffff, 0xff);
    assert (j < 0x7fff);
    pr->giftag.NLOOP = j + 1;
    
    /* update scratchpad buffer */
    if (update) {
      /* fix starting address of next giftag (in spr memory) */
      spr_end += sizeof (pr->rect[0]);
      assert (spr_end <= spr_start + SPR_BUFFER_SIZE);
    } else {
      /* send gif packet to gs queue */
      queue_giftag (pr);
    }
  }
  
  /* update counter so that the very buffer used is not purged soon */
  tex_time[i] = tex_counter++;
}


/*
 * Draw sub-image (sx1,sy1)-(sx2,sy2) of image imp to screen at
 * (dx1,dy1)-(dx2,dy2).  The image will be re-sized automatically to fit the
 * destination area.
 *
 * The function first allocates memory from gs buffer for the image, loads
 * the image to gs (unless it is not already resident) and finally draws
 * portion of the image to frame buffer.  The function tries to maintain a
 * resident set of textures in gs memory to enhance performance.  Linux kit
 * is known for its lousy dma performance compared to the native kit, and we
 * try to overcome this by keeping a large set of resident textures.  If a
 * texture is already in gs memory, we do not need to load it, and in
 * many cases we can thus regain performance lost because of slow dma
 * transfers.  In fact, I have seen pixel writing rates of about 800
 * Mpixels/s which is quite close to maximum performance of 1.2 Gpixels/s
 * that could be achieved in the native developer kit.
 */
void 
bitblt(
    struct image *imp,
    int sx1, int sy1, int sx2, int sy2,
    int dx1, int dy1, int dx2, int dy2)
{
  int i;                                       /* current texture */
  struct reg_image *it;
  
  assert (in_graphics_mode);
  assert (imp != NULL);
  assert (imp->width > 0);
  assert (0 <= sx1  &&  sx1 <= imp->width);
  assert (0 <= sy1  &&  sy1 <= imp->height);
  assert (0 <= sx2  &&  sx2 <= imp->width);
  assert (0 <= sy2  &&  sy2 <= imp->height);
  
  /* find out image tag */
  it = (struct reg_image*) imp->data;
  assert (it != NULL);
  assert (&images[0] <= it  &&  it < &images[MAX_IMAGES]);
  assert (it->data != NULL);

  /* load texture to gs unless it is already resident */
  if (!it->resident) {
    load_texture (imp);    
  }
  i = it->buffer;
  assert (0 <= i  &&  i < nbuffers);

  /* set up drawing environment for using the texture */
  if (cur_texture != i  ||  cur_primitive != BitBlt) {
    use_texture (i, PS2_GS_TEX_TFX_DECAL);
  }

  /* wrap drawing primitive to a gif packet */
  {
    int j;                                     /* index to rect */
    int update;                                /* 1=fix existing packet */
    struct bitbltenv *pr;
    int wx1, wy1, wx2, wy2;                    /* dest in window coords */
    
    if (cur_primitive != BitBlt
        ||  spr_end + sizeof(struct bitbltenv) > spr_start+SPR_BUFFER_SIZE)
    {
      /* create new giftag */
      pr = get_giftag (sizeof (*pr), BitBlt);    
      PS2_GIFTAG_CLEAR_TAG(&pr->giftag);   
      pr->giftag.NREG  = 4;
      pr->giftag.REGS0 = PS2_GIFTAG_REGS_UV;
      pr->giftag.REGS1 = PS2_GIFTAG_REGS_XYZF2;
      pr->giftag.REGS2 = PS2_GIFTAG_REGS_UV;
      pr->giftag.REGS3 = PS2_GIFTAG_REGS_XYZF2;
      pr->giftag.FLG   = PS2_GIFTAG_FLG_REGLIST;

      /* do post giftag */
      update = 0;
      j = 0;        
    } else {    
      /* update existing giftag */
      assert (cur_giftag != NULL);
      pr = (struct bitbltenv*) cur_giftag;
      j = pr->giftag.NLOOP;
      update = 1;
    }

    /*
     * Convert destination address to window coordinates.  Value of 8 (0.5)
     * is substracted from the output coordinate to align the texture in
     * respect to other graphic.
     */
    wx1 = xoffs + dx1*16 - 8;
    wy1 = yoffs + dy1*16 - 8;
    wx2 = xoffs + dx2*16 - 8;
    wy2 = yoffs + dy2*16 - 8;
    
    /* texture must fall within screen estate */
    assert (0 <= wx1  &&  wx1 <= 0xffff);
    assert (0 <= wy1  &&  wy1 <= 0xffff);
    assert (0 <= wx2  &&  wx2 <= 0xffff);
    assert (0 <= wy2  &&  wy2 <= 0xffff);
    
    /*
     * Insert rectangle to giftag.  Notice how we substract 1 from the lower
     * right coordinate.  This is to make 1024x1024 textures paint propely.
     * Due to PS2 hardware, the counter wraps around at 1024.
     */    
    pr->rect[j].t0 = PS2_GS_SETREG_UV (sx1<<4, sy1<<4);
    pr->rect[j].v0 = PS2_GS_SETREG_XYZF2(wx1, wy1, 0xffffff, 0xff);
    pr->rect[j].t1 = PS2_GS_SETREG_UV ((sx2<<4) - 1, (sy2<<4) - 1);
    pr->rect[j].v1 = PS2_GS_SETREG_XYZF2(wx2, wy2, 0xffffff, 0xff);
    assert (j < 0x7fff);
    pr->giftag.NLOOP = j + 1;

    /* update scratchpad buffer */
    if (update) {
      /* fix starting address of next giftag (in spr memory) */
      spr_end += sizeof (pr->rect[0]);
      assert (spr_end <= spr_start + SPR_BUFFER_SIZE);
    } else {
      /* send gif packet to gs queue */
      queue_giftag (pr);
    }
  }
  
  /* update counter so that the very buffer used is not purged soon */
  tex_time[i] = tex_counter++;
}


/*
 * Draw triangle (x1,y1)-(x2,y2),-(x3,y3) using current pen.  The function
 * can draw about 25000 uni-colored triangles per frame (1.5M triangles/s)
 * assuming that triangles are grouped together.  If other drawing commands
 * are intermixed with the drawing of triangles, then the performance will
 * suffer.
 */
void
triangle(
    int x1, int y1, int x2, int y2, int x3, int y3)
{
  /* set gs to accept triangles */
  if (cur_primitive != Triangle) {
    struct primenv *pr = get_giftag (sizeof (*pr), None);
    
    PS2_GIFTAG_CLEAR_TAG(&pr->giftag);
    pr->giftag.NLOOP = 1;
    pr->giftag.NREG  = 2;
    pr->giftag.REGS0 = PS2_GIFTAG_REGS_PRIM;
    pr->giftag.REGS1 = PS2_GIFTAG_REGS_RGBAQ;
    pr->giftag.FLG   = PS2_GIFTAG_FLG_REGLIST;

    /* type of drawing primitive */
    pr->prim = 
        PS2_GS_SETREG_PRIM(
            PS2_GS_PRIM_PRIM_TRIANGLE,         /* draw primative */
            PS2_GS_PRIM_IIP_FLAT,              /* colour type */
            PS2_GS_PRIM_TME_OFF,               /* texture */
            PS2_GS_PRIM_FGE_OFF,               /* fog off */
            PS2_GS_PRIM_ABE_OFF,               /* Alpha BlEnd */ 
            PS2_GS_PRIM_AA1_OFF,               /* AA */
            PS2_GS_PRIM_FST_STQ,               /* texture coord */
            PS2_GS_PRIM_CTXT_CONTEXT1,         /* context set to 1 */
            PS2_GS_PRIM_FIX_NOFIXDDA);         /* dont fix interpolation */
    pr->rgb = fgcolor;
    
    /* send prim register to GS */
    queue_giftag (pr);
  }

  /* draw triangle */
  {
    struct trienv *pr;
    int update;
    int j;
    int wx1, wy1, wx2, wy2, wx3, wy3;
        
    if (cur_primitive != Triangle
      ||  spr_end + sizeof(struct trienv) > spr_start + SPR_BUFFER_SIZE)
    {
      /* create new giftag */
      pr = get_giftag (sizeof (*pr), Triangle);    
      PS2_GIFTAG_CLEAR_TAG(&pr->giftag);
      pr->giftag.NREG  = 4;
      pr->giftag.REGS0 = PS2_GIFTAG_REGS_RGBAQ;
      pr->giftag.REGS1 = PS2_GIFTAG_REGS_XYZ2;
      pr->giftag.REGS2 = PS2_GIFTAG_REGS_XYZ2;
      pr->giftag.REGS3 = PS2_GIFTAG_REGS_XYZ2;
      pr->giftag.FLG   = PS2_GIFTAG_FLG_REGLIST;

      /* do post giftag */
      update = 0;
      j = 0;    
    } else {    
      /* update existing giftag */
      assert (cur_giftag != NULL);
      pr = (struct trienv*) cur_giftag;
      j = pr->giftag.NLOOP;
      update = 1;
    }

    /* convert triangle to window coordinates */
    wx1 = xoffs + x1*16;
    wy1 = yoffs + y1*16;
    wx2 = xoffs + x2*16;
    wy2 = yoffs + y2*16;
    wx3 = xoffs + x3*16;
    wy3 = yoffs + y3*16;
    
    /* triangle must fall into screen */
    assert (0 <= wx1  &&  wx1 <= 0xffff);
    assert (0 <= wy1  &&  wy1 <= 0xffff);
    assert (0 <= wx2  &&  wx2 <= 0xffff);
    assert (0 <= wy2  &&  wy2 <= 0xffff);
    assert (0 <= wx3  &&  wx3 <= 0xffff);
    assert (0 <= wy3  &&  wy3 <= 0xffff);
    
    /* insert triangle to gif packet */
    pr->coords[j].rgb = fgcolor;
    pr->coords[j].v0  = PS2_GS_SETREG_XYZ2(wx1, wy1, 0xffffff);
    pr->coords[j].v1  = PS2_GS_SETREG_XYZ2(wx2, wy2, 0xffffff);
    pr->coords[j].v2  = PS2_GS_SETREG_XYZ2(wx3, wy3, 0xffffff);
    assert (j < 0x7fff);
    pr->giftag.NLOOP = j + 1;
    
    /* fix spr memory */
    if (update) {
      spr_end += sizeof (pr->coords[0]);
      assert (spr_end <= spr_start + SPR_BUFFER_SIZE);
    } else {
      /* send triangle to GS queue */
      queue_giftag (pr);
    }
  }
}


/*
 * Draw single colored line (x1,y1)-(x2,y2) using current pen. The function
 * is quite slow drawing at most 25000 lines per frame (about 1.5M lines/s).
 * However, we will not be drawing many lines so the performance does not
 * really matter.
 */
void
line(
    int x1, int y1, int x2, int y2)
{
  assert (in_graphics_mode);
  
  /* set gs to accept lines */
  if (cur_primitive != Line) {
    struct primenv *pr = get_giftag (sizeof (*pr), None);
    
    PS2_GIFTAG_CLEAR_TAG(&pr->giftag);
    pr->giftag.NLOOP = 1;
    pr->giftag.NREG  = 2;
    pr->giftag.REGS0 = PS2_GIFTAG_REGS_PRIM;
    pr->giftag.REGS1 = PS2_GIFTAG_REGS_RGBAQ;
    pr->giftag.FLG   = PS2_GIFTAG_FLG_REGLIST;

    /* define current drawing primitive */
    pr->prim = 
        PS2_GS_SETREG_PRIM(
            PS2_GS_PRIM_PRIM_LINE,             /* draw primative */
            PS2_GS_PRIM_IIP_FLAT,              /* colour type */
            PS2_GS_PRIM_TME_OFF,               /* texture */
            PS2_GS_PRIM_FGE_OFF,               /* fog off */
            PS2_GS_PRIM_ABE_OFF,               /* Alpha BlEnd */ 
            PS2_GS_PRIM_AA1_OFF,               /* AA */
            PS2_GS_PRIM_FST_STQ,               /* texture coord */
            PS2_GS_PRIM_CTXT_CONTEXT1,         /* context set to 1 */
            PS2_GS_PRIM_FIX_NOFIXDDA);         /* dont fix interpolation */
    pr->rgb = fgcolor;
    
    /* send prim register to GS */
    queue_giftag (pr);
  }
  
  /* draw line */
  {
    int j;
    int update;
    struct lineenv *pr;
    int wx1, wy1, wx2, wy2;
    
    if (cur_primitive != Line
        ||  spr_end + sizeof(struct lineenv) > spr_start + SPR_BUFFER_SIZE)
    {
      /* create new giftag */
      pr = get_giftag (sizeof (*pr), Line);    
      PS2_GIFTAG_CLEAR_TAG(&pr->giftag);   
      pr->giftag.NREG  = 2;
      pr->giftag.REGS0 = PS2_GIFTAG_REGS_XYZ2;
      pr->giftag.REGS1 = PS2_GIFTAG_REGS_XYZ2;
      pr->giftag.FLG   = PS2_GIFTAG_FLG_REGLIST;

      /* do post buffer */
      update = 0;
      j = 0;
    } else {
      /* update existing giftag */
      assert (cur_giftag != NULL);
      pr = (struct lineenv*) cur_giftag;
      j = pr->giftag.NLOOP;
      update = 1;
    }

    /* convert to window coordinates */
    wx1 = xoffs + x1*16;
    wy1 = yoffs + y1*16;
    wx2 = xoffs + x2*16;
    wy2 = yoffs + y2*16;
    
    /* line must fall into screen */
    assert (0 <= wx1  &&  wx1 <= 0xffff);
    assert (0 <= wy1  &&  wy1 <= 0xffff);
    assert (0 <= wx2  &&  wx2 <= 0xffff);
    assert (0 <= wy2  &&  wy2 <= 0xffff);
    
    /* insert coordinate pair to gif packet */
    pr->coords[j].v0 = PS2_GS_SETREG_XYZ2(wx1, wy1, 0xffffff);
    pr->coords[j].v1 = PS2_GS_SETREG_XYZ2(wx2, wy2, 0xffffff);
    assert (j < 0x7fff);
    pr->giftag.NLOOP = j + 1;
    
    /* fix gs memory */
    if (update) {
      /* update starting addr of very next packet in spr mem */
      spr_end += sizeof (pr->coords[0]);
      assert (spr_end <= spr_start + SPR_BUFFER_SIZE);
    } else {
      /* send drawing environment to GS */
      queue_giftag (pr);
    }
  }
}


/*
 * Draw box (x1,y1) - (x2,y2) with a semi-transparent color.  Argument
 * alpha ranges from 0 to 255, and it defines how much the original color
 * shows through.  If alpha is 200, intensity of foreground color is 80%
 * while 20% of background color shows through.
 *
 * BEWARE: the function modifies alpha component of current color.  This
 * should not affect other primitives unless alpha blending is turned on
 * without re-setting foreground color.
 */
void
fade(
    int x1, int y1, int x2, int y2, int alpha)
{
  assert (in_graphics_mode);
  assert (0 <= alpha  &&  alpha < 256);
  
  /* set up drawing mode */
  if (cur_primitive != Line) {
    struct primenv *pr = get_giftag (sizeof (*pr), None);
    ps2_gsreg_rgbaq *cp = (ps2_gsreg_rgbaq*) &fgcolor;
    
    PS2_GIFTAG_CLEAR_TAG(&pr->giftag);
    pr->giftag.NLOOP = 1;
    pr->giftag.NREG  = 2;
    pr->giftag.REGS0 = PS2_GIFTAG_REGS_PRIM;
    pr->giftag.REGS1 = PS2_GIFTAG_REGS_RGBAQ;
    pr->giftag.FLG   = PS2_GIFTAG_FLG_REGLIST;
    
    /* define current drawing primitive */
    pr->prim = 
        PS2_GS_SETREG_PRIM(
            PS2_GS_PRIM_PRIM_SPRITE,           /* draw primative */
            PS2_GS_PRIM_IIP_FLAT,              /* colour type */
            PS2_GS_PRIM_TME_OFF,               /* texture */
            PS2_GS_PRIM_FGE_OFF,               /* fog off */
            PS2_GS_PRIM_ABE_ON,                /* Alpha BlEnd */ 
            PS2_GS_PRIM_AA1_OFF,               /* AA */
            PS2_GS_PRIM_FST_STQ,               /* texture coord */
            PS2_GS_PRIM_CTXT_CONTEXT1,         /* context set to 1 */
            PS2_GS_PRIM_FIX_NOFIXDDA);         /* dont fix interpolation */

    /* specify alpha color */
    pr->rgb = PS2_GS_SETREG_RGBAQ(cp->R,cp->G,cp->B, (alpha >> 1), cp->Q);
    
    /* send prim register to GS */
    queue_giftag (pr);
  }

  /* paint a sprint */
  {
    int j;                                     /* index to rect */
    int update;                                /* 1=fix existing packet */
    struct fadeenv *pr;
    int wx1, wy1, wx2, wy2;                    /* dest in window coords */
    
    /* set up gif packet */  
    if (cur_primitive != Fade
        ||  spr_end + sizeof(struct fadeenv) > spr_start+SPR_BUFFER_SIZE)
    {
      /* create new giftag */
      pr = get_giftag (sizeof (*pr), Fade);
      PS2_GIFTAG_CLEAR_TAG(&pr->giftag);   
      pr->giftag.NREG  = 2;
      pr->giftag.REGS0 = PS2_GIFTAG_REGS_XYZF2;
      pr->giftag.REGS1 = PS2_GIFTAG_REGS_XYZF2;
      pr->giftag.FLG   = PS2_GIFTAG_FLG_REGLIST;
    
      /* post giftag to queue */
      update = 0;
      j = 0;
    } else {    
      /* update existing giftag */
      assert (cur_giftag != NULL);
      pr = (struct fadeenv*) cur_giftag;
      j = pr->giftag.NLOOP;
      update = 1;
    }
    
    /* convert destination address to window coordinates */
    wx1 = xoffs + x1*16;
    wy1 = yoffs + y1*16;
    wx2 = xoffs + x2*16;
    wy2 = yoffs + y2*16;
  
    /* must fall inside screen space */
    assert (0 <= wx1  &&  wx1 <= 0xffff);
    assert (0 <= wy1  &&  wy1 <= 0xffff);
    assert (0 <= wx2  &&  wx2 <= 0xffff);
    assert (0 <= wy2  &&  wy2 <= 0xffff);
    
    /* Insert rectangle to giftag */    
    pr->rect[j].v0 = PS2_GS_SETREG_XYZF2(wx1, wy1, 0xffffff, 0xff);
    pr->rect[j].v1 = PS2_GS_SETREG_XYZF2(wx2, wy2, 0xffffff, 0xff);
    assert (j < 0x7fff);
    pr->giftag.NLOOP = j + 1;
    
    /* update scratchpad buffer */
    if (update) {
      /* fix starting address of next giftag (in spr memory) */
      spr_end += sizeof (pr->rect[0]);
      assert (spr_end <= spr_start + SPR_BUFFER_SIZE);
    } else {
      /* send gif packet to gs queue */
      queue_giftag (pr);
    }
  }
}


/*
 * Set background color.  Note that we should not modify reset structure
 * here as switching to text mode and back would otherwise destroy the
 * background color.
 */
void 
setbg(
    int r, int g, int b)
{
  bgcolor = PS2_GS_SETREG_RGBAQ(r,g,b,0x80,0);
}


/*
 * Set foregroung color used by many drawing functions.
 */
void
setfg(
    int r, int g, int b)
{
  /* save color for future */
  fgcolor = PS2_GS_SETREG_RGBAQ(r,g,b,0x80,0);

  /* invalidate previous giftag to force sending prim tag with new color */
  cur_primitive = None;
}


/*
 * Flush dma buffers in scratchpad memory to gs.  Does not return until at
 * least one buffer is available for use.
 */
void
flush_gfx (void)
{
  assert (in_graphics_mode);
  flush_spr ();
}


/*
 * Switch display and draw buffers, and process postponed suspend and
 * release requests.  The function is called at the end of each display
 * cycle to mark that the draw buffer is ready for display.
 *
 * The two requests are also processes here, as they cannot be handled in
 * the middle of the drawing cycle or otherwise we would lose all contents
 * of the gs memory without the drawing logic knowing about it.  However, 
 * calling this very function means that the program has done all the
 * drawing for the current display cycle, and wants to wait for the vertical
 * retrace signal.  Hence, it is possible to destroy gs memory here, exit to
 * text mode, and then later re-allocate memory for the very next display
 * cycle.
 */
void
swap_buffers (void)
{
  assert (in_graphics_mode);

  /* send graphics primitives still in buffer to gs */
  flush_spr ();
  
  /* switch front and back buffers */  
  display_buffer = 1 - display_buffer;
  draw_buffer    = 1 - draw_buffer;
  
  /* paint performance meter on right edge of screen */
  if (measure) {
    __u64 c = fgcolor;
    wait_gs ();
    set_drawbuf (draw_buffer);
    setfg (255, 0, 0);
    triangle (scrwidth-4,0, scrwidth,0, scrwidth,scrheight);
    triangle (scrwidth-4,0, scrwidth-4,scrheight, scrwidth,scrheight);
    fgcolor = c;
    flush_spr ();
    wait_gs ();
  }
  
  /* reset texture environment */
  cur_texture = -1;
    
  /* wait for vertical retrace */
  wait_gs ();
  ps2_gs_sync_v(0);
 
  /* allow user to switch consoles */
  if (release_request) {
    int status;

    /* exit from graphics mode */
    status = release_vc ();
    assert (status == PS2_GS_VC_REL_SUCCESS);

    /* wait until user switches consoles back */
    switch_tty ();

    /* return to graphics mode */
    status = acquire_vc ();
    assert (status == PS2_GS_VC_REL_SUCCESS);
    
    /* start a fresh display cycle */
    ps2_gs_sync_v(0);
  }

  /* go to sleep as a result of ctrl+z */
  if (suspend_request) {
    hibernate ();
    
    /* start a fresh display cycle */
    ps2_gs_sync_v(0);
  }

  /* update buffer addresses */
  set_dispbuf (display_buffer);
  set_drawbuf (draw_buffer);
}


/*
 * Get width of graphics screen.  For Peach Gardens, the width is usually
 * 640, but you should not depend on this as OpenGL port may have different
 * limits.  (The amount of memory on-board GS prevents Peach Gardens from
 * working as is if the screen resolution is other than 640x480 or 640x512.)
 */
int
screenwidth (void)
{
  assert (in_graphics_mode);
  return scrwidth;
}


/*
 * Get height of graphics screen.  Note that the height varies depending on
 * the monitor or the tv used by the player.  In NTSC and VESA modes the
 * screen is typically 480 pixels tall whereas in PAL mode the screen is 512
 * pixels tall.  Never draw anything beyound 480 pixels without first
 * checking whether or not the user can see it.  Better yet, make all
 * drawing at the bottom of the screen depend on the screen height.
 */
int
screenheight (void)
{
  assert (in_graphics_mode);
  return scrheight;
}


/*
 * Clear screen using background color.  The function also resets
 * viewport and all drawing states.
 */
void 
clear (void)
{
  int x0, y0, x1, y1;
  struct resetenv *pr;

  assert (in_graphics_mode);
  
  /* define clipping region */
  x0 = 2048 - scrwidth/2;
  y0 = 2048 - scrheight/2;
  x1 = x0 + scrwidth;
  y1 = y0 + scrheight;
  
  /* offset to top left corner */
  xoffs = x0 * 16;
  yoffs = y0 * 16;
  
  /* allocate room from scratchpad */
  pr = get_giftag (sizeof (*pr), None);
  
  /* initialize drawing environment */
  PS2_GIFTAG_CLEAR_TAG(&pr->giftag);
  pr->giftag.NLOOP = sizeof(*pr)/16-1;         /* number of tags */
  pr->giftag.NREG = 1;                         /* number of REGS used */
  pr->giftag.REGS0= PS2_GIFTAG_REGS_AD;        /* address mode */

  /* offset to upper left corner of window */
  pr->xyoffs = PS2_GS_SETREG_XYOFFSET (xoffs, yoffs);
  pr->xyoffs_adr = PS2_GS_XYOFFSET_1;
  
  /* clipping region */
  pr->scissor = PS2_GS_SETREG_SCISSOR (0, scrwidth - 1, 0, scrheight - 1);
  pr->scissor_adr = PS2_GS_SCISSOR_1;
  
  /* set graphics primitive */
  pr->prim =
      PS2_GS_SETREG_PRIM(
          PS2_GS_PRIM_PRIM_SPRITE,             /* draw primative */
          PS2_GS_PRIM_IIP_FLAT,                /* colour type */
          PS2_GS_PRIM_TME_OFF,                 /* texture */
          PS2_GS_PRIM_FGE_OFF,                 /* fog off */
          PS2_GS_PRIM_ABE_OFF,                 /* Alpha BlEnd */ 
          PS2_GS_PRIM_AA1_OFF,                 /* AA */
          PS2_GS_PRIM_FST_STQ,                 /* texture coord */
          PS2_GS_PRIM_CTXT_CONTEXT1,           /* context set to 1 */
          PS2_GS_PRIM_FIX_NOFIXDDA);           /* dont fix interpolation */
  pr->prim_adr = PS2_GS_PRIM;
    
  /* set background color */
  pr->rgb0 = bgcolor;
  pr->rgb0_adr = PS2_GS_RGBAQ;
  
  /* coordinates of the area to clear */
  pr->v0 = PS2_GS_SETREG_XYZ(xoffs, yoffs, 0xffffff);
  pr->v0_adr = PS2_GS_XYZF2;
  pr->v1 = PS2_GS_SETREG_XYZ(x1*16, y1*16, 0xffffff);
  pr->v1_adr = PS2_GS_XYZF2;

  /* disable lod and mipmap (for bitblt) */
  pr->tex1 = PS2_GS_SETREG_TEX1(
      PS2_GS_TEX1_LCM_K,                       /* lod calculation method */
      0,                                       /* max mip level */
      PS2_GS_TEX1_MMAG_NEAREST,                /* expansion filter */
      PS2_GS_TEX1_MMAG_NEAREST,                /* reduction filter */
      PS2_GS_TEX1_MTBA_NOAUTO,                 /* base addr specification */
      0,                                       /* lod parameter l */
      0);                                      /* lod parameter k */
  pr->tex1_adr = PS2_GS_TEX1_1;

  /* texture alpha (for bitblt) */
  pr->texa = PS2_GS_SETREG_TEXA(
      0,                                       /* for RGB16 and RGB24 */
      PS2_GS_TEXA_AEM_NORMAL,                  /* black isn't transparent */
      0);                                      /* for RGB16 */
  pr->texa_adr = PS2_GS_TEXA;

  /* alpha correction */
  pr->fba = PS2_GS_SETREG_FBA (1);
  pr->fba_adr = PS2_GS_FBA_1;

  /*
   * Set alpha blending function to match OpenGL's GL_SRC_ONE_MINUS_ALPHA.
   * The function set by this very register is of the form (A-B)*C+D while
   * we want the function to be Cs*As+Cd*(1-As) where Cs is texture color,
   * As texture alpha and Cd frame buffer color.  Luckily, our formula can
   * be expressed as Cs*As+Cd-As*Cd which can be further modified to 
   * (Cs-Cd)*aS+Cd.  Therefore, A is Cs, B is Cd, C is As and D is Cd.
   * Simple, eh?
   */  
  pr->alpha = PS2_GS_SETREG_ALPHA(
      PS2_GS_ALPHA_A_CS,                       /* input color A */
      PS2_GS_ALPHA_B_CD,                       /* input color B */
      PS2_GS_ALPHA_C_AS,                       /* input alpha C */
      PS2_GS_ALPHA_D_CD,                       /* input color D */
      0);                                      /* fixed alpha value */
  pr->alpha_adr = PS2_GS_ALPHA_1;

  /* set foreground color */
  pr->rgb1 = fgcolor;
  pr->rgb1_adr = PS2_GS_RGBAQ;

  /* reset clamping mode */
  pr->clamp = PS2_GS_SETREG_CLAMP(
      PS2_GS_CLAMP_REPEAT,                     /* horiz wrap */
      PS2_GS_CLAMP_REPEAT,                     /* vert wrap */
      0, 0, 0, 0);
  pr->clamp_adr = PS2_GS_CLAMP_1;
    
  /* attributes live in prim register */
  pr->prmodecont = PS2_GS_SETREG_PRMODECONT (PS2_GS_PRMODECONT_REFPRIM);
  pr->prmodecont_adr = PS2_GS_PRMODECONT;
  
  /* send giftag to gs */
  queue_giftag (pr);
}


/*
 * Make frame buffer visible to the user.
 *
 * To enhance the quality of TV output the function sets up the merge
 * circuit to smooth the image.  That is, the function makes display circuit
 * 1 process even scan lines while display circuit 2 reads odd lines.  The
 * merge circuit combines the two scan lines to form single line.  This
 * makes the resulting image smooth, and interlace does not need to be used
 * even with TV resolution.
 */
static void
set_dispbuf(
    int i)
{
  int w;
  ps2_gssreg_dispfb dispfb1;                   /* frame buffer addr */
  ps2_gssreg_dispfb dispfb2;
  
  assert (in_graphics_mode);
  assert (ps2gs >= 0);
  assert (i == 0  ||  i == 1);

  /* compute length of row in frame buffer */
  w = (scrwidth + 63) / 64;
  
  /* clear structures */
  memset (&dispfb1, 0, sizeof (dispfb1));
  memset (&dispfb2, 0, sizeof (dispfb2));  
    
  /* read addr of display circuit 1 */
  dispfb1.FBP = frame_adr[i];
  dispfb1.FBW = w;
  dispfb1.PSM = PIXEL_FORMAT;
  dispfb1.DBX = 0;
  dispfb1.DBY = 0;
  PRIVGS (PS2_GSSREG_DISPFB1, *(__u64*) (void*) &dispfb1);
  
  /* read addr of display circuit 2 */
  dispfb2.FBP = frame_adr[i];
  dispfb2.FBW = w;
  dispfb2.PSM = PIXEL_FORMAT;
  dispfb2.DBX = 0;
  dispfb2.DBY = 1;
  PRIVGS (PS2_GSSREG_DISPFB2, *(__u64*) (void*) &dispfb2);
}


/*
 * Select buffer for drawing.  Note that the function should only ge called
 * when gs is idle.  Changing drawing buffer at the middle may have some
 * unwanted side effects.  Also remember that wait_gs() function only waits
 * until all the graphics primitives are sent to gs -- it does not wait for
 * gs to draw them!
 */
static void
set_drawbuf(
    int i)
{
  int w;
  struct drawbufenv *pr;
  
  /* get room from spr */
  pr = get_giftag (sizeof (*pr), None);

  /* compute length of row in frame buffer */
  w = (scrwidth + 63) / 64;
  
  /* initialize giftag */
  PS2_GIFTAG_CLEAR_TAG(&pr->giftag);
  pr->giftag.NLOOP = (sizeof(*pr) - sizeof(ps2_giftag)) / 16;
  pr->giftag.NREG  = 1;
  pr->giftag.REGS0 = PS2_GIFTAG_REGS_AD;

  /* disable depth test */
  pr->test = PS2_GS_SETREG_TEST(
      PS2_GS_TEST_ATE_OFF,                     /* alpha test off */
      PS2_GS_TEST_ATST_ALWAYS,                 /* all pixels pass alpha */
      0,                                       /* compare to alpha value */
      PS2_GS_TEST_AFAIL_KEEP,                  /* method for failed */
      PS2_GS_TEST_DATE_OFF,                    /* dest alpha test */
      PS2_GS_TEST_DATM_PASS0,                  /* dest alpha test mode */
      PS2_GS_TEST_ZTE_OFF,                     /* depth test */
      PS2_GS_TEST_ZTST_ALWAYS);                /* depth test method */
  pr->test_adr = PS2_GS_TEST_1;

  /* set frame buffer address */
  pr->frame = PS2_GS_SETREG_FRAME(
      frame_adr[i],                            /* frame buffer base */
      w,                                       /* buffer width */
      PIXEL_FORMAT,                            /* pixel storage */
      0);                                      /* draw masking */
  pr->frame_adr = PS2_GS_FRAME_1;

  /* do not update z buffer */
  pr->zbuf = PS2_GS_SETREG_ZBUF(
      0,                                       /* z buffer base pointer */
      PS2_GS_PSMCT16,                          /* pixel storage format */
      PS2_GS_ZBUF_ZMSK_NOMASK);                /* do not write to z buffer*/
  pr->zbuf_adr = PS2_GS_ZBUF_1;

  /* draw every scan line */
  pr->scanmsk = PS2_GS_SETREG_SCANMSK(0);
  pr->scanmsk_adr = PS2_GS_SCANMSK;
  
  /* send giftag to gs */
  queue_giftag (pr);
}


/*
 * Initialize video mode by poking into GS registers directly.  I did not
 * want this, but I was not able to take control of the GS memory *and*
 * manage video modes otherwise.
 *
 * The code that computes display location and magnification is taken from
 * file drivers/video/ps2gs.c in xRhino kernel.  The file is under GPL and
 * copyrighted by Sony Computer Entertainment Inc.
 */
static void
init_video (void)
{
  unsigned int mode;                           /* index of gfx parameters */
  int found;
  ps2_gssreg_display display1;                 /* pos of circuit 1 */
  ps2_gssreg_display display2;                 /* pos of circuit 2 */
  ps2_gsreg_pmode pmode;                       /* circuit 1 & 2 enable */
  __u64 smode1;
  int magh, magv;
  int pdh, pdw, pmagv, pmagh, pdy, pdx;
  
  /* reset structures */
  memset (&display1, 0, sizeof (display1));
  memset (&display2, 0, sizeof (display2));
  memset (&pmode, 0, sizeof (pmode));
  
  /* find parameters of the graphics mode */
  found = 0;
  for (mode = 0; mode < sizeof(syncdata) / sizeof (syncdata[0]); mode++) {
    if (syncdata[mode].modename == graphics_mode) {
      found = 1;
      break;
    }
  }
  if (!found) {
    abort_msg("no suitable video mode");
  }

  assert (0 < scrwidth  &&  scrwidth < 2048);
  assert (0 < scrheight  &&  scrheight < 2048);

  /* wait for vertical retrace before changing video parameters */
  ps2_gs_sync_v(0);    
  
  /* compute magnification for display1/2 register */
  /* Copyright (C) 2000, 2001  Sony Computer Entertainment Inc. */
  magh = syncdata[mode].display.magh + 1;
  magv = syncdata[mode].display.magv + 1;
  pmagh = syncdata[mode].width * magh / scrwidth;
  pmagv = syncdata[mode].height * magv / scrheight;
  assert (0 < pmagh  &&  pmagh <= 16);
  assert (0 < pmagv  &&  pmagv <= 4);
  
  /* compute location and dimensions of display area */
  pdh = scrheight * pmagv;
  if (syncdata[mode].rheight > 0) {
    if (syncdata[mode].rheight * magv > pdh) {
      pdy = syncdata[mode].display.dy
          + (syncdata[mode].rheight*magv-pdh) / 2;
    } else {
      pdy = syncdata[mode].display.dy;
    }
  } else {
    pdy = syncdata[mode].display.dy
        + (syncdata[mode].height * magv - pdh) / 2;
  }  
  pdw = scrwidth * pmagh;
  pdx = syncdata[mode].display.dx
      + (syncdata[mode].width * magh - pdw) / 2;
  
  assert (pdx > 0  &&  pdy > 0);
  assert (0 < pdh  &&  pdh <= 0x7ff);
  assert (0 < pdw  &&  pdw <= 0xfff);
    
  /* fetch smode1 for modification */
  smode1 = syncdata[mode].smode1;

  /* set up graphics mode */
  PRIVGS (PS2_GSSREG_SMODE1, smode1 | ((__u64)1 << 16));
  PRIVGS (PS2_GSSREG_SYNCH1, syncdata[mode].synch1);
  PRIVGS (PS2_GSSREG_SYNCH2, syncdata[mode].synch2);
  PRIVGS (PS2_GSSREG_SYNCV,  syncdata[mode].syncv);
  PRIVGS (PS2_GSSREG_SMODE2, syncdata[mode].smode2);
  PRIVGS (PS2_GSSREG_SRFSH,  syncdata[mode].srfsh);
  PRIVGS (PS2_GSSREG_SMODE1, smode1 & ~((__u64)1<<16) & ~((__u64)1<<17));

  /* position of display circuit 1 */
  display1.DX = pdx;
  display1.DY = pdy;
  display1.MAGV = pmagv - 1;
  display1.MAGH = pmagh - 1;
  display1.DW = pdw - 1;  
  display1.DH = pdh - 1;
  PRIVGS (PS2_GSSREG_DISPLAY1, *(__u64*) (void*) &display1);
  
  /* position of display circuit 2 */
  display2.DX = pdx;
  display2.DY = pdy;
  display2.MAGV = pmagv - 1;
  display2.MAGH = pmagh - 1;
  display2.DW = pdw - 1;
  display2.DH = pdh - 2;
  PRIVGS (PS2_GSSREG_DISPLAY2, *(__u64*) (void*) &display2);
  
  /* read circuit enable */
  if (graphics_mode == PS2_GS_NTSC  ||  graphics_mode == PS2_GS_PAL) {
    /* enable both circuits and blend results */
    pmode.EN1 = 1;
    pmode.EN2 = 1;
    pmode.MMOD = 1;
    pmode.ALP = 0x80;
  } else {
    /* enable just the circuit 1 */
    pmode.EN1 = 1;
    pmode.MMOD = 1;
    pmode.ALP = 0xff;
  }
  pmode.CRTMD = 1;
  PRIVGS (PS2_GSSREG_PMODE, *(__u64*) (void*) &pmode);
}


/*
 * Wait until all DMA transfers have been completed.  Please see
 * ps2dev_ioctls_en.txt for more info on waiting.
 */
static void
wait_gs (void)
{
  assert (ps2gs >= 0);
  if (ioctl (ps2gs, PS2IOC_SENDQCT, 1) != /*OK*/0) {
    DEBUG (("cannot wait for DMA completion"));
  }
}


/*
 * Flush dma buffers in scratchpad memory to gs.
 */
static void
flush_spr (void)
{
  struct ps2_packet dma;

  assert (spr_start != NULL);
  assert (spr_start <= spr_end);
  assert (ps2gs != -1);  
  
  /* is there anything to be sent to gs? */
  if (spr_start < spr_end) {   
    /* specify region to send */
    dma.ptr = spr_start;
    dma.len = spr_end - spr_start;
    assert ((dma.len % 16) == 0);

    /* end DMA packet */
    assert(spr_start <= (char*)cur_giftag && (char*)cur_giftag < spr_end);
    cur_giftag->EOP = 1;
    
    /* swap spr buffers */
    cur_spr_buffer = 1 - cur_spr_buffer;
    spr_start = spr_buffers[cur_spr_buffer];
    spr_end = spr_start;
    cur_giftag = NULL;
    cur_primitive = None;

    /* new buffer must completely fit into the scratchpad */
    assert (spr_start + SPR_BUFFER_SIZE <= spr_adr + spr_size);
    
    /* wait until current dma transfers are completed */
    wait_gs ();

    /*
     * Post buffer with direct non-blocking DMA.  Compared to
     * ps2_dma_start_n(), the ioctl() below is about 50% faster.
     */
    if (ioctl (ps2gs, PS2IOC_SENDA, &dma) != /*OK*/0) {
      DEBUG(("cannot send DMA packet: addr %p, size %d", dma.ptr, dma.len));
      abort_msg ("DMA error");
    }
    
    /* fill current spr buffer with garbage to spot uninitialized vars */
#ifndef NDEBUG
    memset (spr_start, 0xcc, SPR_BUFFER_SIZE);
#endif
  }
}


/*
 * Open graphics screen.  Opening the screen preserves some graphics related
 * services for this process.
 */
int 
open_screen (void)
{
  int adr = 0;                                 /* byte address to gs mem */
  int sz, i, w;
  
  assert (initialized);
    
  /* open ps2gs device at the very first time */
  if (acquire_vc() != PS2_GS_VC_ACQ_FAILURE) {

    /* switch to graphics mode */
    ps2_gs_vc_graphicsmode();
    in_graphics_mode = 1;
    
    /* compute width of row in frame buffer */
    w = (scrwidth + 63) & ~63;
    
    /* compute byte size of a screen */
    sz = (w*scrheight*BYTES_PER_PIXEL + 0x1fff) & ~0x1fff;
    assert ((sz & 0x1fff) == 0);

    /* divide gs memory among various buffers */
    adr = 0;
    
    /* allocate frame buffers from gs memory */
    frame_adr[0] = adr/8192;
    adr += sz;
    frame_adr[1] = adr/8192;
    adr += sz;
    display_buffer = 0;
    draw_buffer = 1;
   
    /* allocate texture buffers and palettes from gs memory */
    tex_counter = 0;
    cur_texture = -1;
    for (i = 0; i < MAX_RESIDENT_TEXS; i++) {
      assert (tex_size[i] <= MAX_TEXTURE_WIDTH);
      
      /* compute texture size assuming one byte per pixel */
      sz = (tex_size[i]*tex_size[i] + 0xff) & ~0xff;
      
      /* allocate space for a texture */
      assert ((adr & 0xff) == 0);
      tex_adr[i] = adr / 256;
      tex_time[i] = 0;
      tex_owner[i] = NULL;
      adr += sz;
      
      /* set up palette for the texture */
      assert ((adr & 0xff) == 0);
      palette_adr[i] = adr / 256;
      adr += PALETTE_SIZE;
      
      /* stop allocating buffers when all the memory is used */
      if (adr > GS_MEMORY_SIZE) {
        break;
      }
    }
        
    /* remember the number of usable texture slots */
    nbuffers = i;
    assert (nbuffers >= MIN_RESIDENT_TEXS);
    if (nbuffers < MIN_RESIDENT_TEXS) {
      abort_msg("out of gs memory");
    }
    
    /* initialize drawing environments for double buffering */
    set_drawbuf (display_buffer);
    clear ();
    set_drawbuf (draw_buffer);
    clear ();
    set_dispbuf (display_buffer);    
    
    return 1;
  } else {
    /* some other process is using graphics */
    fprintf (stderr, "graphics window allocated to some other process\n");
    return 0;
  }
}


/*
 * Close graphics screen and return to text mode.  The contents of the
 * screen are lost for good.
 */
void
close_screen (void)
{ 
  if (in_graphics_mode) {
    /* return to text mode */
    ps2_gs_vc_textmode ();
    
    /* release virtual console */
    release_vc ();
    
    in_graphics_mode = 0;
  }
}


/*
 * Initialize Playstation 2 graphics libraries with allocating as little
 * resources permanently as possible.  The initialization should not open
 * graphics window or reserve any single-user resources such as gs since we
 * do not yet know if we ever enter the graphics mode.  Allocating such
 * resources here would make it harder to run any graphics capable
 * applications in background, for example.
 */
void
init_graphics (void)
{
  int fd;
  struct sigaction handler;
  assert (!initialized);

  /* initialize terminal driver */
  init_tty ();
  
  /* reset module local variables */
  memset (images, 0, sizeof (images));
  
  /*
   * Open GS for a while to read the default screen mode.  If the user is
   * using VESA display for text mode, then he is likely to want VESA
   * graphics mode too.  For more information on this code, please see 
   * the ps2gs_en.txt file at /usr/doc/PlayStation2.
   */
  fd = open (PS2_DEV_GS, O_RDWR);
  if (fd != -1) {
    /* get GS parameters */
    struct ps2_screeninfo scr;
    if (ioctl (fd, PS2IOC_GSCREENINFO, &scr) == /*OK*/0) {
      /* get graphics mode */
      graphics_mode = scr.mode;

      /* select default graphics mode */
      switch (graphics_mode) {
      case PS2_GS_VESA:
        resolution = PS2_GS_640x480;
        scrwidth  = 640;
        scrheight = 480;
        break;
        
      case PS2_GS_NTSC:
        resolution = PS2_GS_640x480;
        scrwidth  = 640;
        scrheight = 480;
        break;
        
      case PS2_GS_PAL:
        resolution = PS2_GS_640x480;
        scrwidth  = 640;
        scrheight = 512;
        break;
      }
    }
    
    /* close GS so that other programs can use it */
    close (fd);
  } else {
    /*
     * Could not open GS for reading the default.  Most likely some other
     * program is already using the console.  This is not much of a problem
     * since we do not even know if the graphics mode ever needs to be
     * used.  Therefore we just issue a warning and linger on.
     */
    DEBUG(("cannot acquire default graphics mode"));
  }

  /* open memory device */
  ps2mem = open (PS2_DEV_MEM, O_RDWR);
  if (ps2mem != -1) {
    /*
     * Get large chunk of non-swapped memory for dma buffers.  This has to
     * be done only once as otherwise the contents of the buffer could be
     * lost.  Hence, we allocate a large chunk of memory here and then
     * divide it among the images using the memory.  Use function
     * alloc_locked() to retrieve the memory allocated here in small pieces.
     */    
    page_adr = mmap (NULL, MAX_PAGES * PAGE_SIZE,
        PROT_READ|PROT_WRITE, MAP_SHARED, ps2mem, 0);
    if (page_adr == (void*)-1) {
      abort_msg("cannot allocate dma buffer");
    }
  } else {
    abort_msg("cannot open "PS2_DEV_MEM);
  }    

  /* set signal handlers to de-initialize graphics mode at sudden exit */
  memset (&handler, 0, sizeof (handler));
  handler.sa_sigaction = (sigaction_handler*) sudden_death;
  sigaction (SIGINT,  &handler, NULL);
  sigaction (SIGABRT, &handler, NULL);
  sigaction (SIGTERM, &handler, NULL);
  sigaction (SIGSEGV, &handler, NULL);

  /* trap suspend and resume */
  handler.sa_sigaction = (sigaction_handler*) suspend;
  sigaction (SIGTSTP, &handler, NULL);
  handler.sa_sigaction = (sigaction_handler*) resume;
  sigaction (SIGCONT, &handler, NULL);  
  
  /* default graphics mode is now detected */
  initialized = 1;
}


/*
 * De-initialize graphics library and return to text mode that was active
 * when the graphics screen was first opened.
 */
void
done_graphics (void)
{
  /* return to text mode if we have opened the graphics screen */
  close_screen ();

  /* release dma buffers */
  if (page_adr != (void*) -1) {
    munmap (page_adr, MAX_PAGES * PAGE_SIZE);
  }

  /* close memory device */
  if (ps2mem != -1) {
    close (ps2mem);
    ps2mem = -1;
  }

  /* de-initialize console input driver */
  done_tty();
  
  initialized = 0;
}


/*
 * This function is supposed to print an error message to graphics window.
 * However, PS2 interface module has no knowledge of fonts or other
 * graphics resources, so it is better to print the message to stderr and
 * quit the application so that the user may see the message.
 */
void
error(
    const char *msg)
{
  fprintf (stderr, "%s\n", msg);
  abort();
}


/*
 * Acquire virtual console for the ps2basic.  This will be called by the
 * ps2dev library when the user wants to switch consoles.
 */
static int
acquire_vc (void)
{
  int ok = 1;
  
  assert (ps2gs < 0);
  assert (ps2spr < 0);
  assert (spr_adr == (void*) -1);

  /* switch to raw console mode */
  raw_tty ();
  
  /* open scratchpad RAM device and map it to memory */
  ps2spr = open(PS2_DEV_SPR, O_RDWR);
  if (ps2spr >= 0) {
    /* find out size of scratchpad */
    struct ps2_sprinfo sprinfo;
    if (ioctl (ps2spr, PS2IOC_SPRINFO, &sprinfo) == /*OK*/0) {
      /* map scratchpad to memory */
      spr_size = sprinfo.size;
      spr_adr = mmap(NULL, spr_size,
          PROT_READ|PROT_WRITE, MAP_SHARED, ps2spr, 0);
      if (spr_adr != (void*)-1) {
        /* allocate DMA buffers from scratchpad RAM */
        alloc_spr_buffers ();
      } else {
        DEBUG(("cannot map scratchpad to user address area"));
        ok = 0;
      }
    } else {
      DEBUG(("cannot determinate size of scratchpad RAM"));
      ok = 0;
    }
  } else {
    /* scratchpad used by some other application */
    DEBUG(("scratchpad device reserved by some other application"));
    ok = 0;
  }
  
  /* open graphics synthetizer */
  if (ok) {
    ps2gs = ps2_gs_open (-1);
    if (ps2gs >= 0) {
      /* set up video mode */
      init_video ();
    } else {
      /* could not acquire console */
      DEBUG(("graphics synthetizer reserved by some other application"));
      ok = 0;
    }
  }

  if (ok) {
    /* virtual console succesfully acquired */
    return PS2_GS_VC_ACQ_SUCCESS;
  } else {
    /* clean up before returning with error */
    release_vc ();
    DEBUG(("cannot return to graphics mode"));
    return PS2_GS_VC_ACQ_FAILURE;
  }
}


/*
 * Release virtual console to allow user to switch virtual consoles.
 */
static int
release_vc (void)
{
  /* empty gs memory */
  purge_textures ();
  
  /* close scratchpad ram device */
  if (ps2spr >= 0) {
    /* release mapped memory area */
    if (spr_adr != (void*) -1) {
      munmap (spr_adr, spr_size);
      spr_adr = (void*) -1;
      spr_size = 0;
    }

    /* close device */
    close (ps2spr);
    ps2spr = -1;
  }
  
  /* close graphics synthetizer */
  if (ps2gs >= 0) {
    ps2_gs_close();
    ps2gs = -1;
  }

  /* restore input mode of the console */
  restore_tty ();
  
  return PS2_GS_VC_REL_SUCCESS;
}


/*
 * Allocate DMA buffers from scratchpad RAM.  The buffers are used for
 * transferring graphics primitives to gs.
 */
static void
alloc_spr_buffers (void)
{
  assert (spr_adr != (void*) -1);

  /* allocate buffers for gs */
  spr_buffers[0] = spr_adr + SPR_BUFFER_1;
  spr_buffers[1] = spr_adr + SPR_BUFFER_2;
  cur_spr_buffer = 0;

  /* set up current buffer */
  spr_start = spr_buffers[cur_spr_buffer];
  spr_end = spr_start;
  cur_giftag = NULL;

  /* buffer must completely fit into the scratchpad */
  assert (spr_start + SPR_BUFFER_SIZE <= spr_adr + spr_size);
}


/*
 * Release texture from gs buffer.  This is called both for removing
 * specific texture from memory and for replacing it with another one.
 */
static void
purge_texture(
    int i)
{
  assert (0 <= i  &&  i < nbuffers);
  assert (tex_owner[i] != NULL);

  /* free gs memory */
  if (tex_owner[i] != NULL) {
    tex_owner[i]->resident = 0;
    tex_owner[i]->buffer = -1;
    tex_owner[i] = NULL;
  }
}


/*
 * Unload all textures from gs memory.  This is often used to recover from
 * gs reset.  That is, contents of the gs memory may be lost if the user 
 * temporarily switches to a text mode.  To recover from that, we need to
 * re-load all textures to gs.  This very function marks buffers so that we
 * know to reload the textures when we use them for the next time.
 */
static void
purge_textures (void)
{
  int i;

  /* no texture is currently used */
  cur_texture = -1;
      
  /* purge all textures from gs */
  for (i = 0; i < nbuffers; i++) {
    if (tex_owner[i] != NULL) {
      purge_texture (i);
    }
  }
}


/*
 * Allocate a dma buffers for transferring textures to gs.  Note that
 * textures may be registered before opening graphics window.  Hence, we
 * must not try to upload the image to gs just yet.
 *
 * Also note that both the palette and the image pixels are wrapped to a dma
 * packet at this time.  Therefore, any changes made to the palette 
 * after returning from this very function will not change the image.  If
 * such changes are necessary, please be sure to fix the palette before
 * registering an image.  This may require you to initialize image in
 * multiple steps without the help of init_image().
 */
int
register_image(
    struct image *imp, const char *buffer, int len)
{
  struct reg_image *it = NULL;
  int i;
  int ok;
  
  assert (initialized);
  assert (imp != NULL);
  assert (buffer != NULL);
  assert (imp->data == NULL);

  /* ensure that image width valid */
  if ((imp->width % 8) != 0) {
    fprintf (stderr, "image width must be multiple of 8\n");
    return 0;
  }
  
  /* refuse to register too large images */
  if (imp->width > MAX_TEXTURE_WIDTH) {
    fprintf (stderr, "maximum texture width exceeded\n");
    return 0;
  }
  
  /* find a free reg_image structure */
  for (i = 0; i < MAX_IMAGES; i++) {
    if (images[i].data == NULL) {
      /* found a free image */
      it = &images[i];
      break;
    }
  }
  assert (it != NULL);
  if (it == NULL) {
    abort_msg ("too many images registered");
  }
  
  /* compute maximum size of dma packet that contain the image */
  it->data_size =
      imp->width * imp->height +               /* image data */
      sizeof (struct paletteenv) + 256 * 4 +   /* palette + 16 */
      sizeof (struct transferenv) * 2 +
      sizeof (struct dataenv) +                /* image header + 16 */
      sizeof (struct imageenv) +
      32;                                      /* for extra giftags */

  /* align dma buffer to page boundary */
  it->data_size = (it->data_size + 0xfff) & ~0xfff;
  
  /* allocate dma buffer for the image */
  it->data = alloc_locked (it->data_size);
  if (it->data != NULL) {
    char *p = (char*) it->data;
    assert (((long)p) % 16 == 0);

    /* initialize static transfer parameters */
    {
      struct transferenv *pr = (struct transferenv*) (void*) p;
      p += sizeof (struct transferenv);
      assert (((long)p) % 16 == 0);
      PS2_GIFTAG_CLEAR_TAG(&pr->giftag);
      pr->giftag.NLOOP = sizeof(*pr)/16-1;     /* number of tags */
      pr->giftag.NREG = 1;                     /* number of REGS used */
      pr->giftag.REGS0= PS2_GIFTAG_REGS_AD;    /* address mode */

      /* init destination address (address set at bitblt) */
      it->image = pr;
      pr->bitbltbuf_adr = PS2_GS_BITBLTBUF;

      /* pixel offset in gs buffer */
      pr->trxpos = PS2_GS_SETREG_TRXPOS (0,0, 0,0,PS2_GS_TRXPOS_DIR_LR_UD);
      pr->trxpos_adr = PS2_GS_TRXPOS;
      
      /* size of input image */
      pr->trxreg = PS2_GS_SETREG_TRXREG (imp->width, imp->height);
      pr->trxreg_adr = PS2_GS_TRXREG;
      
      /* transmission direction */
      pr->trxdir = PS2_GS_SETREG_TRXDIR (PS2_GS_TRXDIR_HOST_TO_LOCAL);
      pr->trxdir_adr = PS2_GS_TRXDIR;
    }    

    /* copy image data */
    assert ((len % 8) == 0);
    {
      struct dataenv *pr = NULL;
      __u64 *src = (__u64*) (void*) buffer;
      __u64 *e   = (__u64*) (void*) (buffer + len);
      
      i = 0;
      while (src < e) {
        /*
         * Start a new giftag if necessary.  This is needed as NLOOP is only
         * 15 bits in length and cannot thus transfer more than 32678 tags
         * (which is 522848 pixels with IDTEX8 mode).
         */
        if (i == 0) {
          pr = (struct dataenv*) (void*) p;
          p += sizeof (pr->giftag);
          PS2_GIFTAG_CLEAR_TAG (&pr->giftag);
          pr->giftag.FLG = PS2_GIFTAG_FLG_IMAGE;
        }

        /* store 128 bits to dma packet */
        assert (pr != NULL);
        pr->data[i].lo = *src++;
        pr->data[i].hi = *src++;
        p += sizeof (pr->data[i]);
        assert (((long)p) % 16 == 0);
        ++i;

        /* end giftag once NLOOP cannot contain more data */
        if (i >= 0x7fff) {
          assert (i <= 0x7fff);
          pr->giftag.NLOOP = i;
          i = 0;
        }
      }
      
      /* fix size in final giftag */
      assert (pr != NULL);
      assert (i <= 0x7fff);
      pr->giftag.NLOOP = i;
    }
    
    /*
     * Giftag for transferring palette to gs.  Note that the palette is just
     * a two dimensional array just like any other gs object.  Thus, we need
     * to reorder the data in order to use the palette.  For more info,
     * please see page 32 of gs user's manual.
     */
    {
      struct transferenv *pp;
      struct paletteenv *dp;
      
      pp = (struct transferenv*) (void*) p;
      p += sizeof (struct transferenv);
      assert (((long)p) % 16 == 0);
      
      PS2_GIFTAG_CLEAR_TAG(&pp->giftag);
      pp->giftag.NLOOP = sizeof(*pp)/16-1;     /* number of tags */
      pp->giftag.NREG = 1;                     /* number of REGS used */
      pp->giftag.REGS0= PS2_GIFTAG_REGS_AD;    /* address mode */
      
      /* init destination address (set at bitblt) */
      it->palette = pp;
      pp->bitbltbuf_adr = PS2_GS_BITBLTBUF;
      
      /* pixel offset in gs buffer */
      pp->trxpos = PS2_GS_SETREG_TRXPOS(0,0, 0,0, PS2_GS_TRXPOS_DIR_LR_UD);
      pp->trxpos_adr = PS2_GS_TRXPOS;
      
      /* size of input data */
      pp->trxreg = PS2_GS_SETREG_TRXREG (16, 16);
      pp->trxreg_adr = PS2_GS_TRXREG;
      
      /* transmission direction */
      pp->trxdir = PS2_GS_SETREG_TRXDIR (PS2_GS_TRXDIR_HOST_TO_LOCAL);
      pp->trxdir_adr = PS2_GS_TRXDIR;

      /* always transfer all 256 colors */
      dp = (struct paletteenv*) (void*) p;
      p += sizeof (pp->giftag) + 256 * sizeof (dp->color[0]);
      assert (((long)p) % 16 == 0);
      PS2_GIFTAG_CLEAR_TAG (&dp->giftag);
      dp->giftag.NLOOP = 256 * sizeof (dp->color[0]) / 16;
      dp->giftag.FLG = PS2_GIFTAG_FLG_IMAGE;

      /*
       * Copy palette.  Note that in the generic image structure the alpha
       * component of 0xff means 1.0.  However, in gs the alpha component is
       * only 7 bits wide, and we must drop the least significant bit.
       */      
      for (i = 0; i < 256; i++) {
        int j = CLUTSWIZZLE(i);
        dp->color[j].r = imp->palette[i].r;
        dp->color[j].g = imp->palette[i].g;
        dp->color[j].b = imp->palette[i].b;
        dp->color[j].a = imp->palette[i].a >> 1;
      }
    }

    /* flush textures and close dma packet */
    {
      struct imageenv *pr = (struct imageenv*) (void*) p;
      p += sizeof (struct imageenv);
      assert (((long)p) % 16 == 0);
      
      /* mark the end of data */
      PS2_GIFTAG_CLEAR_TAG(&pr->giftag);
      pr->giftag.NLOOP = sizeof(*pr)/16-1;     /* number of tags */
      pr->giftag.NREG = 1;                     /* number of REGS used */
      pr->giftag.EOP = 1;                      /* end of packet */
      pr->giftag.REGS0= PS2_GIFTAG_REGS_AD;    /* address mode */

      /* flush texture cache */
      pr->texflush = 0;
      pr->texflush_adr = PS2_GS_TEXFLUSH;
    }

    /* remember the size of packet */
    it->data_len = p - (char*)it->data;
    assert ((it->data_len % 16) == 0);

    /* must not overrun the dma buffer */
    assert (it->data_len <= it->data_size);    
    
    /* texture is not resident yet */
    it->resident = 0;
    it->buffer = -1;

    /* associate the reg_image structure with the generic image struct */
    imp->data = it;
    ok = 1;
    
  } else {
    /* cannot allocate dma buffer */
    abort_msg ("out of dma buffers");
  }
  return ok;
}


/*
 * Release dma buffers allocated for an image.
 */
void 
unregister_image(
    struct image *imp)
{
  struct reg_image *it;  
  assert (imp != NULL);

  /* get pointer to reg_image structure */
  it = (struct reg_image*) imp->data;
  if (it != NULL) {
    /* image has been registered */
    assert (&images[0] <= it  &&  it < &images[MAX_IMAGES]);
    
    /* release dma pages */
    if (it->data != NULL) {
      free_locked (it->data, it->data_size);
    }

    /* unload texture */
    if (it->resident) {
      purge_texture (it->buffer);
    }
    
    /* release image */
    it->data = NULL;
    it->resident = 0;
    it->buffer = -1;
    it->data_size = 0;
    it->data_len = 0;
    imp->data = NULL;
  }
}


/*
 * Allocate locked memory that can be used for dma transfers between main
 * memory and gs.  The function does not return if there is no memory
 * available.
 *
 * The implementation below assumes that (1) buffer allocations are rare and
 * (2) there are no more than a few thousand pages available to allocate.
 * If thede assumptions hold, then we can simplify the operation by using a
 * linear search algorithm.  However, if there were lots of pages to
 * allocate, this would be very slow.  Hence, I advise you to never use the
 * code below for any other purpose!
 */
static void *
alloc_locked(
    int bytes)
{
  int i;
  int n;

  /* compute the number of pages needed to hold the bytes */
  n = ((bytes + PAGE_SIZE - 1) & ~(PAGE_SIZE-1)) / PAGE_SIZE;

  /* find a range of free pages */
  for (i = 0; i < MAX_PAGES - n; i++) {
    
    /* is the page used? */
    if (page_used[i] == 0) {
      int avail = 1, j;
      
      /* see whether ith page starts a of consequtive range of free pages */
      for (j = 1; j < n; j++) {
        assert (i + j < MAX_PAGES);
        if (page_used[i + j] != 0) {
          /* not a consequtive range of memory */
          avail = 0;
          break;
        }
      }

      if (avail) {
        /* found a consequtive range of pages */
        int j;

        /* mark the pages used */
        for (j = 0; j < n; j++) {
          assert (i + j < MAX_PAGES);
          page_used[i + j] = 1;
        }

        /* return the address of the starting page */
        return ((char*) page_adr) + PAGE_SIZE * i;
      }
    }
  }

  abort_msg ("out of DMA buffers");
  /*NOTREACHED*/
}


/*
 * Release pages allocated for dma transfers.
 */
static void
free_locked(
    void *p, int bytes)
{
  int n, i, j;
  
  assert (bytes > 0);
  assert (page_adr <= p && (char*)p < (char*)page_adr+MAX_PAGES*PAGE_SIZE);
  
  /* compute the number of pages needed to hold the bytes */
  n = ((bytes + PAGE_SIZE - 1) & ~(PAGE_SIZE-1)) / PAGE_SIZE;
  
  /* compute index to page_used table */
  i = ((char*)p - (char*)page_adr) / PAGE_SIZE;

  /* release n pages */
  for (j = 0; j < n; j++) {
    assert (page_used[i + j]);
    assert (i + j < MAX_PAGES);
    page_used[i + j] = 0;
  }
}


/*
 * Signal handler for exiting the application in a hurry.  The handler will
 * be called if user aborts the program with ctrl+c, a segmentation fault
 * occurs or the application aborts due to assert() or abort().
 */
static void
sudden_death(
    int sig, int code, struct sigcontext *ctx)
{
  (void) sig; (void) code; (void) ctx;
  
  DEBUG (("aborting due to signal %d, code %d", sig, code));
  done_graphics ();
  _exit (3);
}


/*
 * Handle ctrl+z signal pressed from keyboard by registering the request or
 * going directly to hibernation mode.  For the suspend/resume to work
 * correctly the application must paint the frame buffer every frame.  If
 * the application does not swap buffers for a long period of time, then the
 * screen will not be updated after we return from the hibernation mode.
 */
static void
suspend(
    int sig, int code, struct sigcontext *ctx)
{
  (void) sig; (void) code; (void) ctx;
  DEBUG (("suspend request"));

  if (in_graphics_mode) {
    /*
     * We are in graphics mode, and we cannot suspend right now as switching
     * to text mode would destroy contents of the frame buffer.  Set flag to
     * go into hibernation mode after the whole frame has been drawn.
     */
    suspend_request = 1;
  } else {
    /* we are in console mode so we can go to sleep at any time */
    hibernate ();    
  }
}


/*
 * Continue processing after returning from sleep induced by ctrl+z.  The
 * return happens when the user types something like %1 to return to stopped
 * program.  However, this can also happen as a result of user typing "bg"
 * in which case the application should continue in background.  We, on the
 * other hand, cannot run in background...
 */
static void
resume(
    int sig, int code, struct sigcontext *ctx)
{
  struct sigaction handler;
  (void) sig; (void) code; (void) ctx;

  /* are we continuing in foreground */
  if (tty_owner () == getpid()) {
    
    /* continuing in foreground */
    if (resume_graphics) {
      open_screen ();
    }
    
    /* do not go to hibernation unless ctrl+z is pressed again */
    suspend_request = 0;
    resume_graphics = 0;
    
    /* re-install ctrl+z handler */
    memset (&handler, 0, sizeof (handler));  
    handler.sa_sigaction = (sigaction_handler*) suspend;
    sigaction (SIGTSTP, &handler, NULL);
    
  } else {

    /* process continued with bg */
    DEBUG (("cowardly refusing to continue in background"));
    
  }
}


/*
 * Go into hibernation mode following ctrl+z.  Note how the tick count is
 * saved and restored after returning from hibernation.  This behaviour
 * cancels the time spent in hibernation such that any delay loops that are
 * dependent of the tick count are not exhausted when we return from sleep.
 * For example, the following code works correctly even though user stops
 * the program for half an hour.  The loop is still executed for 10 seconds.
 *
 *  for (start = ticks(); ticks() < start + 10000;) {
 *    clear ();
 *    bitblt (&inside,
 *        0,0,inside.width,inside.height,
 *        0,0,screenwidth(),screenheight());
 *    swap_buffers ();
 *  }
 */
static void
hibernate (void)
{
  ticks_t n;
  struct sigaction handler;

  /* return to text mode */
  if (in_graphics_mode) {
    resume_graphics = 1;
    close_screen ();    
  } else {
    resume_graphics = 0;
  }
  
  /* reset handler for ctrl+z */
  memset (&handler, 0, sizeof (handler));  
  handler.sa_sigaction = (sigaction_handler*) SIG_DFL;
  sigaction (SIGTSTP, &handler, NULL);

  /* get current time */
  n = ticks ();
  
  /* sleep until parent wakes us up as foreground process */
  do{
    raise (SIGTSTP);
  } while (tty_owner () != getpid());
      

  /* restore tick counter to "forget" time spent in sleeping */
  set_ticks (n);
}


/*
 * Initialize keyboard interface.
 */
void
init_key (void)
{
  /*NOP*/
}


/*
 * De-initialize keyboard interface.
 */
void
done_key (void)
{
  /*NOP*/
}


/*
 * Get keyboard event from linux console.  Note that although it would be
 * technically possible, we do not want to read the console when not in
 * graphics mode.
 */
int
key_event (void)
{
  if (in_graphics_mode) {
    return tty_event ();
  } else {
    return NOEVENT;
  }
}


/*
 * Print help on graphics library dependent options.
 */
void
print_graphics_options (void)
{
  printf(
      "Graphics options:\n"
      "  --ntsc           NTSC video output (USA and Japan).\n"
      "  --pal            PAL video output (Europe).\n"
      "  --vesa           VESA video output.\n"
      "  -m               Measure length of display cycle.\n"
  );
}


/*
 * Parse and read away option argv[i] if that is specific to the Playstation
 * 2 graphics library.  The function must return the number of arguments it
 * reads away from the arguments array.  If the Playstation 2 graphics
 * library does not use an option, the function should return zero without
 * printing an error message.
 *
 * The graphics library must have been initialized prior to using this very
 * function.
 */
int
read_graphics_options(
    int *argc, char **argv, int i)
{ 
  int opts = 1;
  const char *p;

  assert (initialized);
  assert (argv != NULL);
  assert (argc != NULL);
  assert (i < *argc);

  p = argv[i];
   
  /* graphics modes */
  if (strcmp(p, "--ntsc") == 0) {
    graphics_mode = PS2_GS_NTSC;
    scrwidth = 640;
    scrheight = 480;
  } else if (strcmp (p, "--pal") == 0) {
    graphics_mode = PS2_GS_PAL;
    scrwidth = 640;
    scrheight = 512;
  } else if (strcmp (p, "--vesa") == 0) {
    graphics_mode = PS2_GS_VESA;
    scrwidth = 640;
    scrheight = 480;
  } else if (strcmp (p, "-m") == 0) {
    measure = 1;
  } else {
    opts = 0;
  }

  /* read away options that we processed */
  if (opts != 0) {
    int j;
    
    /* remove arguments from the argv array */
    for (j = i + opts; j < *argc; j++) {
      argv[j-opts] = argv[j];
    }

    /* fix argc */
    *argc -= opts;
    assert (*argc >= 0);
  }

  /* return number of options read */
  return opts;
}
#endif /*PS2DEV*/
