/*
 * menu.c - simple menu system
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * 
 * Mar 25 2005, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: say.c,v 1.2 2005/06/22 13:11:05 tronkko Exp $
 *
 * $Log: say.c,v $
 * Revision 1.2  2005/06/22 13:11:05  tronkko
 * separated person and say to two different modules
 *
 * Revision 1.1  2005/04/03 16:25:29  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "say.h"
#include "person.h"
#include "graphics.h"
#include "text.h"
#include "font10x16.h"
#include "debug.h"


/* transparency of bg in percent, 20 means that 80 of color shows through */
#define TRANSPARENCY 50

/* height of text box */
#define TEXT_LINES 4

/* amount of while space around face (not including rim) */
#define FACE_LEFT   20
#define FACE_TOP    16
#define FACE_RIGHT  20
#define FACE_BOTTOM 10

/* size of face image on screen */
#define FACE_WIDTH  64
#define FACE_HEIGHT 64

/* width of empty region around text */
#define TEXT_RIGHT  5
#define TEXT_TOP    10
#define TEXT_BOTTOM 10

/* width of rim around face */
#define FACE_RIM 6


/* has the module been initialized already? */
static int initialized = 0;



/*
 * Paint a quotation to screen along with person's name and face.
 *
 * BEWARE: changes font, wrap region and position of text cursor.  Module
 * resource must be initialized before using this very function.
 */
void
say(
    enum person_t who, const char *what)
{
  int rx1, ry1, rx2, ry2;                      /* region boundaries */
  int tx1, ty1, tx2, ty2;
  int htface, httext;                          /* height of face & text */
  int bw;                                      /* border width */
  
  assert (initialized);
  assert (what != NULL  &&  *what != '\0');
  
  /* use small font */
  assert (font10x16 != NULL);
  setfont (font10x16);
  
  /* compute height of face and text area */
  htface = FACE_HEIGHT + FACE_TOP + FACE_BOTTOM;
  httext = TEXT_LINES * textheight("M") + TEXT_TOP + TEXT_BOTTOM;
  
  /* compute boundaries of text box */
  rx1 = 0;
  if (htface > httext) {
    ry1 = screenheight() - htface;
  } else {
    ry1 = screenheight() - httext;
  }
  rx2 = screenwidth();
  ry2 = screenheight();
  
  /* darken background */
  setfg (0, 0, 0);
  fade (rx1, ry1, rx2, ry2, 255 - TRANSPARENCY * 255 / 100);
  
  /* compute location of face */
  tx1 = rx1 + FACE_LEFT;
  ty1 = ry1 + FACE_TOP;
  tx2 = tx1 + FACE_WIDTH;
  ty2 = ty1 + FACE_HEIGHT;
  
  /* paint rim around face */
  bw = FACE_RIM;
  setfg (0xcf, 0xcf, 0xcf);
  triangle (tx1-bw, ty1-bw, tx2+bw, ty1-bw, tx1-bw, ty2+bw);
  
  bw = FACE_RIM / 2;
  setfg (0x80, 0x80, 0x80);
  triangle (tx1-bw, ty1-bw, tx2+bw, ty1-bw, tx1-bw, ty2+bw);
  
  bw = FACE_RIM;
  setfg (0x80, 0x80, 0x80);
  triangle (tx2+bw, ty2+bw, tx2+bw, ty1-bw, tx1-bw, ty2+bw);
  
  bw = FACE_RIM / 2;
  setfg (0xcf, 0xcf, 0xcf);
  triangle (tx2+bw, ty2+bw, tx2+bw, ty1-bw, tx1-bw, ty2+bw);

  /* paint face */
  draw_face (who, tx1, ty1, tx2, ty2);
  
  /* define auto-fill region */
  tx1 = rx1 + FACE_LEFT + FACE_WIDTH + FACE_RIGHT;
  ty1 = ry1 + TEXT_TOP;
  tx2 = rx2 - TEXT_RIGHT;
  ty2 = ry2;
  settextpos (tx1, ty1 + textheight("M"));
  setwrap (tx1, ty1, tx2, ty2);
  
  /* paint person's name */
  setfg (255, 255, 0);
  wrap_text (person_name (who));
  wrap_text (": ");

  /* paint text */
  setfg (255, 255, 255);
  wrap_text (what);
}


/*
 * Initialize module.
 *
 * Note: modules font10x16 and person must be initialized in order to say
 * anything.
 */
void
init_say (void)
{
  assert (!initialized);
  
  initialized = 1;
}


/*
 * De-initialize module.
 */
void
done_say (void)
{
  assert (initialized);
  
  initialized = 0;
}
