/*
 * tty.c - keyboard input module for linux terminals
 * 
 * Copyright (C) 2005  Toni Ronkko
 * 
 * Peach Gardens is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by the
 * Free Software Foundation; either version 2 of the License, or (at your
 * option) any later version.
 *
 * Peach Gardens is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU 
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with Peach Gardens; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * 
 * Nov 23 2004, Toni Ronkko <toni at users.playstation2-linux.com>
 *
 * $Id: tty.c,v 1.2 2005/01/09 21:09:20 tronkko Exp $
 *
 * $Log: tty.c,v $
 * Revision 1.2  2005/01/09 21:09:20  tronkko
 * use port1 when reporting events, check if continuing as a background
 * process
 *
 * Revision 1.1  2004/11/24 20:15:06  tronkko
 * Initial revision
 *
 *
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <assert.h>

#include "config.h"
#ifdef PS2DEV

#include "input.h"
#include "tty.h"
#include "debug.h"
#include "ticks.h"

#ifdef HAVE_TERMIOS_H
# include <termios.h>
#endif

#ifdef HAVE_UNISTD_H
# include <unistd.h>
#endif

#ifdef HAVE_SYS_IOCTL_H
# include <sys/ioctl.h>
#endif

#ifdef HAVE_SYS_SELECT_H
# include <sys/select.h>
#endif

#ifdef HAVE_SIGNAL_H
# include <signal.h>
#endif

#ifdef HAVE_LINUX_KD_H
# include <linux/kd.h>
#endif

#ifdef HAVE_LINUX_VT_H
# include <linux/vt.h>
#endif


/* scan codes */
#define ESC        1
#define LSHIFT    42
#define RSHIFT    54
#define LCONTROL  29
#define LALT      56
#define F1        59
#define F2        60
#define F3        61
#define F4        62
#define F5        63
#define F6        64
#define F7        65
#define F8        66
#define F9        67
#define F10       68
#define F11       87
#define F12       88
#define C         46
#define Z         44
#define A         30
#define W         17
#define D         32
#define X         45
#define P         25
#define E         18
#define TAB       15
#define DEL      111
#define BACKSPC   14
#define CLEFT    105                           /* cursor keys */
#define CRIGHT   106
#define CUP      103
#define CDOWN    108
#define KPLEFT    75                           /* keypad */
#define KPRIGHT   77
#define KPUP      72
#define KPDOWN    80
#define KPDEL     83                           /* keypad delete */
#define KPPLUS    78                           /* keypad plus */
#define KPMINUS   74
#define SPACE     57
#define RET       28

/* maximum number of keys in a keyboard */
#define MAX_KEYS 128


/* type of signal handler function */
typedef void sigaction_handler (int, siginfo_t*, void *);


/* handle console acquire request */
static void acquire_req (int sig, int code, struct sigcontext *ctx);

/* handle console release request */
static void release_req (int sig, int code, struct sigcontext *ctx);


/* has the terminal been initialized? */
static int initialized = 0;

/* permit use of raw input mode */
static int raw_enable = 0;

/* original input mode of the console */
static long omode = 0;

/* original input flags */
static struct termios old;

/* original console switch mode */
static struct vt_mode oswitch;

/* input flags of raw mode */
static struct termios raw;

/* read keyboard in raw mode */
static int raw_mode = 0;

/* key states */
static int key[MAX_KEYS] = { 0 };

/* postponed graphics console release request */
int release_request = 0;


/*
 * Read key code from tty and translate the key code to an event code.  Note
 * that the function does not make a difference between left and right
 * shift, control or enter.
 *
 * Keyboard mappings:
 *
 * Event      Port  Key
 * LEFT       1     Numpad left or cursor left
 * RIGHT      1     Numpad right or cursor right
 * UP         1     Numpad up or cursor up
 * DOWN       1     Numpad down or cursor down
 * ACTION     1     Numpad enter, return or space
 * EXIT       1     Numpad plus or esc
 * ERASE      1     Backspace, delete or numpad delete
 * MENU       1     Numpad minus
 * PAUSE      1     P
 *
 * The function interprets alt+fn, ctrl+c and ctrl+z key from the stream,
 * and executes them as if the console were in native mode.  This makes
 * debugging easier, as you can switch to other console to see the
 * debug messages or the source code.  Hint: start your application as
 * "t-clouds x > /dev/tty4".  All output to stdout will now appear on 4th
 * console.
 *
 * The function cannot be used in normal io mode.  Be sure to use raw_tty()
 * to switch to proper mode before using the function.
 */
int
tty_event (void)
{
  int ev = NOEVENT;
  assert (initialized);
  assert (raw_mode);
  
  if (raw_mode) {
    /* read key code from linux console */
    int retval;
    fd_set rfds;
    struct timeval timeout;
      
    /* watch stdin (fd 0) to see when it has input */
    do{
      FD_ZERO (&rfds);
      FD_SET (0, &rfds);
      timeout.tv_sec = 0;
      timeout.tv_usec = 0;
      retval = select (1, &rfds,  NULL, NULL, &timeout);
      assert (retval >= 0);
      if (retval > 0) {
        unsigned char c;
        
        /* input pending */
        if (read (0, &c, 1) == 1) {
          int press, i;
          
          /* extract key code */
          i = (c & 0x7f);
          
          /* is the key pressed? */
          if ((c & 0x80) == 0) {
            /* press */
            press = PRESS;
            key[i] = 1;
          } else {
            /* release */
            press = RELEASE;
            key[i] = 0;
          }
          
          /* convert scan code to event */
          switch (i) {
          case F1: case F2: case F3: case F4: case F5: case F6:
          case F7: case F8: case F9: case F10: case F11: case F12: {
            if (key[i] && key[LALT] && !key[LSHIFT] && !key[RSHIFT]) {
              /*
               * Left alt with a function key pressed.  Normally, linux
               * console interprets this key press, and switches consoles.
               * However, as we are reading in raw mode, linux console never
               * sees the key, and, to fix the console switching, we must
               * switch the consoles manually.
               */              
              int error;
              int cons;
              int ncons;

              /* get console number */
              switch (i) {
              case F1:  cons =  1; break;
              case F2:  cons =  2; break;
              case F3:  cons =  3; break;
              case F4:  cons =  4; break;
              case F5:  cons =  5; break;
              case F6:  cons =  6; break;
              case F7:  cons =  7; break;
              case F8:  cons =  8; break;
              case F9:  cons =  9; break;
              case F10: cons = 10; break;
              case F11: cons = 11; break;
              case F12: cons = 12; break;
              default:
                abort_msg ("invalid console switch");
              }

              /* get maximum number of consoles */
              error = ioctl (0, VT_OPENQRY, &ncons);
              if (error) {
                ncons = 0;
              }
              
              /* switch to console */
              if (cons < ncons) {
                struct vt_stat cur;
                
                /* get currently active console */
                error = ioctl (0, VT_GETSTATE, &cur);
                assert (!error);

                /* switch consoles unless the console is already active */
                if (cur.v_active != cons) {
                  error = ioctl (0, VT_ACTIVATE, cons);
                  assert (!error);
                }
              }
            }
            break;}
            
          case C:
            if (key[i] && key[LCONTROL] && !key[LSHIFT] && !key[RSHIFT]) {
              /*
               * Control+c received.  Throw SIGINT signal to possible
               * handler function to make the application stop.  If the
               * signal has been catched and ignored, then the application
               * will continue normally.
               */
              DEBUG (("ctrl-c pressed"));
              restore_tty ();
              raise (SIGINT);
              /* signal ignored if raise() returns */
              raw_tty ();
            }
            break;
            
          case Z:
            if (key[i] && key[LCONTROL] && !key[LSHIFT] && !key[RSHIFT]) {
              /*
               * Control+z pressed.  Send SIGTSTP signal to possible handler
               * function to suspend execution of the application.  If this
               * succeeds, the processing resumes when the application is
               * resumed.
               */
              DEBUG (("ctrl-z pressed"));
              restore_tty ();
              raise (SIGTSTP);
              raw_tty ();
            }
            break;

          case P:       ev = PAUSE | PORT1 | press; break;

          case DEL:     ev = ERASE | PORT1 | press; break;
          case KPDEL:   ev = ERASE | PORT1 | press; break;
          case BACKSPC: ev = ERASE | PORT1 | press; break;
          case CUP:     ev = UP    | PORT1 | press; break;
          case KPUP:    ev = UP    | PORT1 | press; break;
          case CDOWN:   ev = DOWN  | PORT1 | press; break;
          case KPDOWN:  ev = DOWN  | PORT1 | press; break;
          case CLEFT:   ev = LEFT  | PORT1 | press; break;
          case KPLEFT:  ev = LEFT  | PORT1 | press; break;
          case CRIGHT:  ev = RIGHT | PORT1 | press; break;
          case KPRIGHT: ev = RIGHT | PORT1 | press; break;
          case KPPLUS:  ev = EXIT  | PORT1 | press; break;
          case KPMINUS: ev = MENU  | PORT1 | press; break;
          case RET:     ev = ACTION| PORT1 | press; break;

          case SPACE:   ev = ACTION| PORT1 | press; break;
          case ESC:     ev = EXIT  | PORT1 | press; break;
            
          default:
            ;
            /*DEBUG (("ignode key code %d", c));*/
          }
        }
      }

      /* repeat as long as there is input pending */
    } while (retval > 0  &&  ev == NOEVENT);
  }
  return ev;
}


/*
 * Switch tty to raw mode.  The function does not return until the current
 * console is owned by this very process.
 */
void
raw_tty (void)
{
  int error;
  assert (initialized);
  
  if (raw_enable) {
    
    /*
     * Set input flags.  This will fail unless this very program is the sole
     * owner of the console.  For example, the function fails if trying to
     * run this very application in background, in which case the parent
     * shell is also using the console, 
     */
    error = tcsetattr (0, TCSANOW, &raw);
    assert (!error);
    
    /* switch to raw mode */
    error = ioctl (0, KDSKBMODE, K_RAW);
    assert (!error);

    /* now in raw mode */
    raw_mode = 1;
  }
}


/*
 * Restore original input more and input flags.
 * 
 * The function is registered by atexit() so the tty mode can be restored
 * properly at program exit.  However, this also means that this very
 * function can be invoked multiple times even though the tty is no longer
 * in raw mode.  Be sure to reset the input state properly to allow multiple
 * execution.  Also note that this very module has not necessarily been
 * initialized at all.
 */
void
restore_tty (void)
{
  if (initialized) {
    if (raw_mode) {
      int error;
      
      /* return to original input mode */
      error = ioctl (0, KDSKBMODE, omode);
      assert (!error);
      
      /* restore flags */
      error = tcsetattr (0, TCSANOW, &old);
      assert (!error);

      raw_mode = 0;
    }
  }
}


/*
 * Switch to another console, and wait until the current console is
 * activated again.
 */
void
switch_tty (void)
{
  int error;

  /* restore console modes */
  restore_tty ();
  
  /* allow the kernel to switch consoles */
  error = ioctl (0, VT_RELDISP, /*OK*/1);
  if (!error) {
    ticks_t n;
    
    /* get current time */
    n = ticks ();
    
    /* sleep until SIGUSR2 signal is received */
    assert (release_request != 0);
    while (release_request != 0) {
      pause ();
    }

    /* restore tick counter to "forget" time spent in sleeping */
    set_ticks (n);
    
    /* tell kernel we are back! */
    error = ioctl (0, VT_RELDISP, /*ACK*/2);
    assert (!error);
  } else {
    DEBUG (("switch acknowledgement ignored by kernel"));
    release_request = 0;
  }
  
  /* return to raw input mode */
  raw_tty ();
}


/*
 * Get id of foreground process in the current console.  The function is
 * used for preventing the application from being run in background.  As we
 * use graphic resources, it is not possible to run in background.
 */
pid_t
tty_owner (void)
{
  return tcgetpgrp(0);
}


/*
 * Handle acquire request.  Linux kernel invokes this very function with
 * SIGUSR2 signal when the console allocated for us is activated again.
 */
static void
acquire_req(
    int sig, int code, struct sigcontext *ctx)
{
  (void) sig; (void) code; (void) ctx; 
  release_request = 0;
}


/*
 * Handle release request.  The function is invoked as a result of the user
 * wanting to switch consoles.  However, we cannot allow the consoles to be
 * switched straight away, as we may be in the middle of display cycle.
 * Hence, remember the request for later times.
 */
static void
release_req(
    int sig, int code, struct sigcontext *ctx)
{
  (void) sig; (void) code; (void) ctx; 
  release_request = 1;
}


/*
 * Initialize tty driver but remain in cooked io mode.  To switch to raw
 * input mode, use function raw_tty().
 */
void
init_tty (void)
{
  int error;
  assert (!initialized);

  /* reset variables */
  raw_mode = 0;
  memset (key, 0, sizeof (key));
  
  /* get current keyboard mode of linux console */
  error = ioctl (0, KDGKBMODE, &omode);
  if (!error) {
    struct vt_mode sw;
    struct sigaction handler;
    
    /* prepare to de-initialize tty automatically at program exit */
    error = atexit (restore_tty);
    assert (!error);
    
    /* get original input flags */
    error = tcgetattr (0, &old);
    if (error) {
      /* cannot get flags */
      abort_msg("cannot read console flags");
    }

    /* get console switch type */
    error = ioctl (0, VT_GETMODE, &oswitch);    
    if (error) {
      abort_msg("cannot read console mode");
    }    
    
    /* prepare flags of raw input mode */
    raw = old;
    raw.c_iflag &= ~(
          ISTRIP                               /* no strip eight bit */
        | IGNBRK                               /* no ignore break */
        | BRKINT                               /* no generate SIGINT */
        | INLCR                                /* no translate nl to cr */
        | ICRNL                                /* no translate cr to nl */
        | IGNCR                                /* no ignore carriage ret */
        | IXON                                 /* no flow control (output)*/
        | IXOFF);                              /* no flow control (input) */
    raw.c_iflag |=
          IGNPAR;                              /* do ignore parity errors */
    raw.c_cflag &= ~(
          PARENB                               /* no parity bit generate */
        | CSIZE);                              /* character size mask */
    raw.c_cflag |=
        CS8;                                   /* do 8 bit input */
    raw.c_lflag &= ~(
          ICANON                               /* no pseudo chars */
        | ISIG                                 /* no generate signals */
        | IEXTEN                               /* no interpret codes */
        | ECHO                                 /* no echo */
        | ECHONL);                             /* no echo newline */

    /* bind signals for switching consoles */
    memset (&handler, 0, sizeof (handler));
    handler.sa_sigaction = (sigaction_handler*) release_req;
    sigaction (SIGUSR1, &handler, NULL);
    handler.sa_sigaction = (sigaction_handler*) acquire_req;
    sigaction (SIGUSR2, &handler, NULL);  

    /* specify signal to raise when switching consoles */
    sw = oswitch;
    sw.mode = VT_PROCESS;
    sw.relsig = SIGUSR1;
    sw.acqsig = SIGUSR2;
    error = ioctl (0, VT_SETMODE, &sw);
    assert (!error);
    
    /* module is now initialized */
    raw_enable = 1;
    initialized = 1;
  } else {
    /* keyboard input disabled, not a linux console */
    raw_enable = 0;
    initialized = 1;  
  }
}


/*
 * De-initialize tty driver and restore original input mode.  Beware: the
 * keyboard may be left at unusable state if you do not call this very
 * function at program exit.  You should therefore catch common signals
 * to ensure that the tty driver gets de-initialized properly despite of
 * programming errors.
 */
void
done_tty (void)
{
  /* restore switch mode */
  if (initialized  &&  raw_enable) {
    int error;
    error = ioctl (0, VT_SETMODE, &oswitch);
    assert (!error);
  }
  
  /* restore input mode and flags */
  restore_tty ();  
  initialized = 0;
}

#endif /*PS2DEV*/
